/*
    api- and os-independed helper-functions for the saa7146 device driver
    
    Copyright (C) 1998-2001 Michael Hunold <michael@mihu.de>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "saa7146.h"

#define my_min(type,x,y) \
	({ type __x = (x), __y = (y); __x < __y ? __x: __y; })
#define my_max(type,x,y) \
	({ type __x = (x), __y = (y); __x > __y ? __x: __y; })

/* helper function */
void my_wait(struct saa7146* saa, long ms)
{
	unsigned long timeout = jiffies + ((ms+9)/10);
	while(time_before(jiffies, timeout)) {
		schedule();
	}
}

/* most of the follwing functions take a struct saa7146* as
   the first argument and assume that it has at least the following
   data fields:
   
   struct saa7146 {
	struct saa7146_modes_constants	modes_constants;
	settings according to currently selected video mode
	(PAL, NTSC)
*/ 

/* helper functions for the calculation of the horizontal- and vertical
   scaling registers, clip-format-register etc ...
   these functions take pointers to the (most-likely read-out
   original-values) and manipulate them according to the requested
   changes.
*/

/* hps_coeff used for CXY and CXUV; scale 1/1 -> scale 1/64 */
struct {
	u16 hps_coeff;
	u16 weight_sum;
} hps_h_coeff_tab [] = { 
	{0x00,   2}, {0x02,   4}, {0x00,   4}, {0x06,   8}, {0x02,   8},
	{0x08,   8}, {0x00,   8}, {0x1E,  16}, {0x0E,   8}, {0x26,   8},
	{0x06,   8}, {0x42,   8}, {0x02,   8}, {0x80,   8}, {0x00,   8},
	{0xFE,  16}, {0xFE,   8}, {0x7E,   8}, {0x7E,   8}, {0x3E,   8},
	{0x3E,   8}, {0x1E,   8}, {0x1E,   8}, {0x0E,   8}, {0x0E,   8},
	{0x06,   8}, {0x06,   8}, {0x02,   8}, {0x02,   8}, {0x00,   8},
	{0x00,   8}, {0xFE,  16}, {0xFE,   8}, {0xFE,   8}, {0xFE,   8},
	{0xFE,   8}, {0xFE,   8}, {0xFE,   8}, {0xFE,   8}, {0xFE,   8},
	{0xFE,   8}, {0xFE,   8}, {0xFE,   8}, {0xFE,   8}, {0xFE,   8},
	{0xFE,   8}, {0xFE,   8}, {0xFE,   8}, {0xFE,   8}, {0x7E,   8},
	{0x7E,   8}, {0x3E,   8}, {0x3E,   8}, {0x1E,   8}, {0x1E,   8},
	{0x0E,   8}, {0x0E,   8}, {0x06,   8}, {0x06,   8}, {0x02,   8},
	{0x02,   8}, {0x00,   8}, {0x00,   8}, {0xFE,  16}
};

/* table of attenuation values for horizontal scaling */
u8 h_attenuation[] = { 1, 2, 4, 8, 2, 4, 8, 16, 0};

/* calculate horizontal scale registers */
int calculate_h_scale_registers(struct saa7146* saa,
	int in_x, int out_x, int flip_lr,
	u32* hps_ctrl, u32* hps_v_gain, u32* hps_h_prescale, u32* hps_h_scale)
{
	/* horizontal prescaler */
	u32 dcgx = 0, xpsc = 0, xacm = 0, cxy = 0, cxuv = 0;	
	/* horizontal scaler */
	u32 xim = 0, xp = 0, xsci =0;				
	/* vertical scale & gain */
	u32 pfuv = 0;						

	/* helper variables */
	u32 h_atten = 0, i = 0;

	if ( 0 == out_x ) {
		printk("saa7146_core.o: calculate_h_scale_registers: invalid value (=0).\n");

		return -EINVAL;
	}
	
	/* mask out vanity-bit */
	*hps_ctrl &= ~MASK_29;
		
	/* calculate prescale-(xspc)-value:	[n   .. 1/2) : 1
				    		[1/2 .. 1/3) : 2
				    		[1/3 .. 1/4) : 3
						... 		*/
	if (in_x > out_x) {
		xpsc = in_x / out_x;
	}
	else {
		/* zooming */
		xpsc = 1;						
	}
	
	/* if flip_lr-bit is set, number of pixels after
	   horizontal prescaling must be < 384 */
	if ( 0 != flip_lr ) {
	
		/* set vanity bit */
		*hps_ctrl |= MASK_29;
	
		while (in_x / xpsc >= 384 )
			xpsc++;
	}
	/* if zooming is wanted, number of pixels after
	   horizontal prescaling must be < 768 */
	else {
		while ( in_x / xpsc >= 768 )
			xpsc++;
	}
	
	/* maximum prescale is 64 (p.69) */
	if ( xpsc > 64 )
		xpsc = 64;

	/* keep xacm clear*/
	xacm = 0;
	
	/* set horizontal filter parameters (CXY = CXUV) */
	cxy = hps_h_coeff_tab[( (xpsc - 1) < 63 ? (xpsc - 1) : 63 )].hps_coeff;
	cxuv = cxy;
	
	/* calculate and set horizontal fine scale (xsci) */
	
	/* bypass the horizontal scaler ? */
	if ( (in_x == out_x) && ( 1 == xpsc ) )
		xsci = 0x400;
	else	
		xsci = ( (1024 * in_x) / (out_x * xpsc) ) + xpsc;

	/* set start phase for horizontal fine scale (xp) to 0 */	
	xp = 0;
	
	/* set xim, if we bypass the horizontal scaler */
	if ( 0x400 == xsci )
		xim = 1;
	else
		xim = 0;
		
	/* if the prescaler is bypassed, enable horizontal
	   accumulation mode (xacm) and clear dcgx */
	if( 1 == xpsc ) {
		xacm = 1;
		dcgx = 0;
	}
	else {
		xacm = 0;
		/* get best match in the table of attenuations
		   for horizontal scaling */
		h_atten = hps_h_coeff_tab[( (xpsc - 1) < 63 ? (xpsc - 1) : 63 )].weight_sum;
	
		for (i = 0; h_attenuation[i] != 0; i++) {
			if (h_attenuation[i] >= h_atten)
				break;
		}
	
		dcgx = i;
	}

	/* the horizontal scaling increment controls the UV filter
	   to reduce the bandwith to improve the display quality,
	   so set it ... */
	if ( xsci == 0x400)
		pfuv = 0x00;
	else if ( xsci < 0x600)
		pfuv = 0x01;
	else if ( xsci < 0x680)
		pfuv = 0x11;
	else if ( xsci < 0x700)
		pfuv = 0x22;
	else
		pfuv = 0x33;

	
	*hps_v_gain  &= MASK_W0|MASK_B2;
	*hps_v_gain  |= (pfuv << 24);	

	*hps_h_scale 	&= ~(MASK_W1 | 0xf000);
	*hps_h_scale	|= (xim << 31) | (xp << 24) | (xsci << 12);

	*hps_h_prescale	|= (dcgx << 27) | ((xpsc-1) << 18) | (xacm << 17) | (cxy << 8) | (cxuv << 0);

	return 0;
}

struct {
	u16 hps_coeff;
	u16 weight_sum;
} hps_v_coeff_tab [] = { 
 {0x0100,   2},  {0x0102,   4},  {0x0300,   4},  {0x0106,   8},  {0x0502,   8},
 {0x0708,   8},  {0x0F00,   8},  {0x011E,  16},  {0x110E,  16},  {0x1926,  16},
 {0x3906,  16},  {0x3D42,  16},  {0x7D02,  16},  {0x7F80,  16},  {0xFF00,  16},
 {0x01FE,  32},  {0x01FE,  32},  {0x817E,  32},  {0x817E,  32},  {0xC13E,  32},
 {0xC13E,  32},  {0xE11E,  32},  {0xE11E,  32},  {0xF10E,  32},  {0xF10E,  32},
 {0xF906,  32},  {0xF906,  32},  {0xFD02,  32},  {0xFD02,  32},  {0xFF00,  32},
 {0xFF00,  32},  {0x01FE,  64},  {0x01FE,  64},  {0x01FE,  64},  {0x01FE,  64},
 {0x01FE,  64},  {0x01FE,  64},  {0x01FE,  64},  {0x01FE,  64},  {0x01FE,  64},
 {0x01FE,  64},  {0x01FE,  64},  {0x01FE,  64},  {0x01FE,  64},  {0x01FE,  64},
 {0x01FE,  64},  {0x01FE,  64},  {0x01FE,  64},  {0x01FE,  64},  {0x817E,  64},
 {0x817E,  64},  {0xC13E,  64},  {0xC13E,  64},  {0xE11E,  64},  {0xE11E,  64},
 {0xF10E,  64},  {0xF10E,  64},  {0xF906,  64},  {0xF906,  64},  {0xFD02,  64},
 {0xFD02,  64},  {0xFF00,  64},  {0xFF00,  64},  {0x01FE, 128}
};

/* table of attenuation values for vertical scaling */
u16 v_attenuation[] = { 2, 4, 8, 16, 32, 64, 128, 256, 0};

/* calculate vertical scale registers */
int calculate_v_scale_registers(struct saa7146* saa, int interlace,
	int in_y, int out_y, u32* hps_v_scale, u32* hps_v_gain)
{
	/* vertical scaling */
	u32 yacm = 0, ysci = 0, yacl = 0, ypo = 0, ype = 0;
	/* vertical scale & gain */
	u32 dcgy = 0, cya_cyb = 0;
				
	/* helper variables */
	u32 v_atten = 0, i = 0;	

	/* error, if vertical zooming */
	if ( in_y < out_y ) {
		printk("calculate_v_scale_registers: vertical zooming not allowed (in_y:%d, out_y:%d).\n",in_y,out_y);
		return -EINVAL;
	}

	/* linear phase interpolation may be used if scaling is between
	   1 and 1/2 or scaling is between 1/2 and 1/4 (if interlace is
	   set; see below) */
	if(    ((2*out_y) >= in_y) || (((4*out_y) >= in_y)
	    && interlace != 0)) {

		/* if scaling is between 1/2 and 1/4 we can only use
		   one field. we do so, if interlace is set */
		if( interlace != 0 && (out_y*4) >= in_y && (out_y*2) <= in_y) 
			out_y *= 2;

		yacm = 0;
		yacl = 0;
		cya_cyb = 0x00ff;
		
		/* calculate scaling increment */
		if ( in_y > out_y )
			ysci = ((1024 * in_y) / (out_y + 1)) - 1024;
		else
			ysci = 0;

		dcgy = 0;

		/* calculate ype and ypo */
		ype = ysci / 16;
		ypo = ype + (ysci / 64);
		
	}
	else {
		yacm = 1;	

		/* calculate scaling increment */
		ysci = (((10 * 1024 * (in_y - out_y - 1)) / in_y) + 9) / 10;

		/* calculate ype and ypo */
		ypo = ype = ((ysci + 15) / 16);

		/* the sequence length interval (yacl) has to be set according
		   to the prescale value, e.g.	[n   .. 1/2) : 0
		   				[1/2 .. 1/3) : 1
						[1/3 .. 1/4) : 2
						... */
		if ( ysci < 512) {
			yacl = 0;
		}
		else {
			yacl = ( ysci / (1024 - ysci) );
		}

		/* get filter coefficients for cya, cyb from table hps_v_coeff_tab */	
		cya_cyb = hps_v_coeff_tab[ (yacl < 63 ? yacl : 63 ) ].hps_coeff;

		/* get best match in the table of attenuations for vertical scaling */
		v_atten = hps_v_coeff_tab[ (yacl < 63 ? yacl : 63 ) ].weight_sum;

		for (i = 0; v_attenuation[i] != 0; i++) {
			if (v_attenuation[i] >= v_atten)
				break;
		}
	
		dcgy = i;
	}

	/* ypo and ype swapped in spec ? */
	*hps_v_scale	|= (yacm << 31) | (ysci << 21) | (yacl << 15) | (ypo << 8 ) | (ype << 1);

	*hps_v_gain	&= ~(MASK_W0|MASK_B2);
	*hps_v_gain	|= (dcgy << 16) | (cya_cyb << 0);

	return 0;
}

void calculate_hps_source_and_sync(struct saa7146* saa, int source, int sync, u32* hps_ctrl)
{
	*hps_ctrl &= ~(MASK_30 | MASK_31 | MASK_28);
	*hps_ctrl |= (source << 30) | (sync << 28);
}

void calculate_hxo_and_hyo(struct saa7146* saa, u32* hps_h_scale, u32* hps_ctrl)
{
	int hyo = 0, hxo = 0;
	
/*	printk("calculate_hxo_hyo_and_sources: mode:%d\n",saa->mode);
*/
	hyo = saa->modes_constants.v_offset;
	hxo = saa->modes_constants.h_offset;
				
	*hps_h_scale	&= ~(MASK_B0 | 0xf00);
	*hps_h_scale	|= (hxo <<  0);

	*hps_ctrl	&= ~(MASK_W0 | MASK_B2);
	*hps_ctrl	|= (hyo << 12);
}

void calculate_output_format_register(struct saa7146* saa, u32 palette, u32* clip_format)
{
	/* clear out the necessary bits */
	*clip_format &= 0x0000ffff;	
	/* set these bits new */
	*clip_format |=  (( ((palette&0xf00)>>8) << 30) | ((palette&0x00f) << 24) | (((palette&0x0f0)>>4) << 16));
}

void calculate_bcs_ctrl_register(struct saa7146 *saa, int brightness, int contrast, int colour, u32 *bcs_ctrl)
{
	*bcs_ctrl = ((brightness << 24) | (contrast << 16) | (colour <<  0));
}

int calculate_clipping_registers_mask(struct saa7146* saa, u32* clipping, int width, int height, int interlace, struct saa7146_video_dma* vdma2, u32* clip_format, u32* arbtr_ctrl)
{ 
	u32 clip_addr = 0, clip_pitch = 0;

	/* adjust arbitration control register */
	*arbtr_ctrl &= 0xffff00ff;
	*arbtr_ctrl |= 0x00001000;	

	clip_addr  = virt_to_bus(clipping);
	clip_pitch = ((width+31)/32)*4;

	vdma2->base_even	= clip_addr;	
	vdma2->base_page	= 0x04;	/* enable read - operation */
	vdma2->prot_addr	= clip_addr + (clip_pitch*height);

	/* convention: if scaling is between 1/2 and 1/4 we only use
	   the even lines, the odd lines get discarded (see vertical scaling) */
	if( interlace != 0 && height*4 >= saa->modes_constants.v_calc && height*2 <= saa->modes_constants.v_calc) {
		vdma2->base_odd		= vdma2->prot_addr;
		vdma2->pitch		= clip_pitch;
		vdma2->num_line_byte	= (((height)-1) << 16) | (clip_pitch-1);
	} else {
		vdma2->base_odd		= clip_addr+clip_pitch;
		vdma2->pitch		= clip_pitch*2;
		vdma2->num_line_byte	= (((height/2)-1) << 16) | (clip_pitch-1);
	}

	*clip_format &= 0xfffffff7;

	return 0;
}

/* helpfer functions for clipping */

/* helper functions to emulate rect-clipping via mask-clipping.
   note: these are extremely inefficient, but for clipping with less than 16
   windows rect-clipping should be done anyway...
*/

/* clear one pixel of the clipping memory at position (x,y) */
void set_pixel(int x, int y, int window_width, u32* mem) {

	int mem_per_row = 0;
	int offset	= 0;
	int shift 	= 0;
	int bit 	= 0;

	mem_per_row 	= window_width / 32 + ( (0 == window_width % 32) ? 0 : 1); 
	offset 		= y * mem_per_row + (x / 32);
	shift 		= 31 - (x % 32);
	bit 		= (1 << shift);
	
	mem[offset] |= bit;	
}

/* clear a box out of the clipping memory, beginning at (x,y) with "width" and "height" */
void set_box(int x, int y, int width, int height, int window_width, int window_height, u32* mem)
{
	int ty = 0;
	int tx = 0;

	/* the video_clip-struct may contain negative values to indicate that a window
	   doesn't lay completly over the video window. Thus, we correct the values */
	
	if( width < 0) {
		x += width; width = -width;
	}
	if( height < 0) {
		y += height; height = -height;
	}

	if( x < 0) {
		width += x; x = 0;
	}
	if( y < 0) {
		height += y; y = 0;
	}

	if( width <= 0 || height <= 0) {
 		printk("saa7146_core.o: set_box: sanity error!\n");
		return;
	}			

	if(x + width > window_width)
		width -= (x + width) - window_width;
	if(y + height > window_height)
		height -= (y + height) - window_height;

	/* Now, set a '1' in the memory, where no video picture should appear */
	for(ty = y; ty < y+height; ty++) {		
		for(tx = x; tx < x+width; tx++) {
			set_pixel(tx, ty, window_width, mem);
		}
	}
}

int emulate_rect_clipping(struct saa7146 *saa, u32* clipping, int clipcount, int x[], int y[], int w[], int h[], u32 w_width, u32 w_height)
{
	int i = 0;
	
	/* clear out clipping mem */
	memset(clipping, 0x0, SAA7146_CLIPPING_MEM);

	/* go through list of clipping-windows, clear out rectangular-regions in the clipping memory */
	for(i = 0; i < clipcount; i++) {
		set_box(x[i], y[i], w[i], h[i], w_width, w_height, clipping);
	}		

	return 0;
}

/* simple bubble-sort algorithm with duplicate elimination */
int sort_and_eliminate(u32* values, int* count)
{
	int low = 0, high = 0, top = 0, temp = 0;
	int cur = 0, next = 0;
	
	/* sanity checks */
	if( (0 > *count) || (NULL == values) ) {
		printk("saa7146_core.o: sort_and_eliminate: internal error #1\n");
		return -EINVAL;
	}
	
	/* bubble sort the first count items of the array values */
	for( top = *count; top > 0; top--) {
		for( low = 0, high = 1; high < top; low++, high++) {
			if( values[low] > values[high] ) {
				temp = values[low];
				values[low] = values[high];
				values[high] = temp;
			}
		}
	}

	/* remove duplicate items */
	for( cur = 0, next = 1; next < *count; next++) {
		if( values[cur] != values[next])
			values[++cur] = values[next];
	}
	
	*count = cur + 1;
	 
	return 0;
}

int calculate_clipping_registers_rect(struct saa7146 *saa, u32* clipping,
	int interlace, int clipcount, int td_flip,
	int x[], int y[], int w[], int h[],
	int width, int height, struct saa7146_video_dma* vdma2,
	u32* clip_format, u32* arbtr_ctrl)
{
	u32 line_list[32];			
	u32 pixel_list[32];
	int numdwords = 0;

	int i = 0, j = 0;
	int l = 0, r = 0, t = 0, b = 0;
	int cnt_line = 0, cnt_pixel = 0;

	/* clear out memory */
	memset(&line_list[0],  0x00, sizeof(u32)*32);
	memset(&pixel_list[0], 0x00, sizeof(u32)*32);
	memset(clipping,  0x00, SAA7146_CLIPPING_MEM);

	
	/* fill the line and pixel-lists */
	for(i = 0; i < clipcount; i++) {

//		printk("%d: x:%d, y:%d, w:%d, h:%d\n",i,x[i],y[i],w[i],h[i]);

		if( 0 != td_flip ) {
			y[i] = height - y[i] - h[i];
		}

		/* calculate values for l(eft), r(ight), t(op), b(ottom) */
		l = x[i];
		r = x[i]+w[i];
		
		t = y[i];
		b = y[i]+h[i];
		
		/* insert left/right coordinates */
		pixel_list[ 2*i   ] = my_min(int, l, width);
		pixel_list[(2*i)+1] = my_min(int, r, width);
		/* insert top/bottom coordinates */
		line_list[ 2*i   ] = my_min(int, t, height);
		line_list[(2*i)+1] = my_min(int, b, height);
	}

	/* sort and eliminate lists */
	cnt_line = cnt_pixel = 2*clipcount;
	sort_and_eliminate( &pixel_list[0], &cnt_pixel );
	sort_and_eliminate( &line_list[0], &cnt_line );

	/* calculate the number of used u32s */
	numdwords = my_max(int, (cnt_line+1), (cnt_pixel+1))*2; 
	numdwords = my_max(int, 4, numdwords);
	numdwords = my_min(int, 64, numdwords);

	/* fill up cliptable */
	for(i = 0; i < cnt_pixel; i++) {
		clipping[2*i] |= (pixel_list[i] << 16);
	}
	for(i = 0; i < cnt_line; i++) {
		clipping[(2*i)+1] |= (line_list[i] << 16);
	}

	/* fill up cliptable with the display infos */
	for(j = 0; j < clipcount; j++) {

		for(i = 0; i < cnt_pixel; i++) {

			if( x[j] < 0)
				x[j] = 0;

			if( pixel_list[i] < (x[j] + w[j])) {
			
				if ( pixel_list[i] >= x[j] ) {
					clipping[2*i] |= (1 << j);			
				}
			}
		}
		for(i = 0; i < cnt_line; i++) {

			if( y[j] < 0)
				y[j] = 0;

			if( line_list[i] < (y[j] + h[j]) ) {

				if( line_list[i] >= y[j] ) {
					clipping[(2*i)+1] |= (1 << j);			
				}
			}
		}
	}

	/* adjust arbitration control register */
	*arbtr_ctrl &= 0xffff00ff;
	*arbtr_ctrl |= 0x00001c00;	
	
	vdma2->base_even	= virt_to_bus(clipping);
	vdma2->base_odd		= virt_to_bus(clipping);
	vdma2->prot_addr	= virt_to_bus(clipping)+((sizeof(u32))*(numdwords));
	vdma2->base_page	= 0x04;
	vdma2->pitch		= 0x00;
	vdma2->num_line_byte	= (0 << 16 | (sizeof(u32))*(numdwords-1) );

	/* set clipping-mode. please note again, that for sizes below 1/2, we only use the
	   even-field. because of this, we have to specify recinterl correctly (specs, p. 97)*/
	*clip_format &= 0xfffffff7;

	if( interlace != 0 && height*4 >= saa->modes_constants.v_calc && height*2 <= saa->modes_constants.v_calc) {
		*clip_format |= 0x00000000;
	} else {
		*clip_format |= 0x00000008;
	}

	/* undo changes, so this function will work the next time, too! */
	for(i = 0; i < clipcount; i++) {
		if( 0 != td_flip ) {
			y[i] = height - (y[i] + h[i]);
		}	
	}
	
	return 0;
}

/* disable clipping */
int saa7146_set_disable_clipping(struct saa7146* saa)
{
	u32 clip_format	= saa7146_read(saa, CLIP_FORMAT_CTRL);
	
	/* mask out relevant bits (=lower word)*/
	clip_format &= MASK_W1;

	/* upload clipping-registers*/
	saa7146_write(saa, CLIP_FORMAT_CTRL,clip_format);
 	saa7146_write(saa, MC2, (MASK_05 | MASK_21));
 
 	/* disable video dma2 */
	saa7146_write(saa, MC1, (MASK_21));

	return 0;
}

int saa7146_set_clipping_rect(struct saa7146* saa, int interlace,
	int width, int height, u32* clipping_memory,
	int clipcount, int x[], int y[], int w[], int h[], int inverted, int td_flip) 
{
	struct	saa7146_video_dma	vdma2;

	u32 clip_format	= saa7146_read(saa, CLIP_FORMAT_CTRL);
	u32 arbtr_ctrl	= saa7146_read(saa, PCI_BT_V1);

	u32 type = ( inverted == 0 ? SAA7146_CLIPPING_RECT : SAA7146_CLIPPING_RECT_INVERTED );

/*
	DEB_EE(printk("saa7146_core.o: clip_windows\n"));
*/
	/* some sanity checks first */
	if ( width <= 0 || height <= 0 || clipcount < 0 ) {
		printk("saa7146_core.o: clip_windows: sanity error #1!\n");
		return -EINVAL;
	}
	
	/* check clipcount, disable clipping if clipcount == 0*/
	if( clipcount == 0 ) {
		saa7146_set_disable_clipping(saa);
		return 0;
	}

	/* some more sanity checks */
	if( NULL == x || NULL == y || NULL == w || NULL == h ) {
		printk("saa7146_core.o: clip_windows: sanity error #2!\n");
		return -EINVAL;
	}

	/* rectangle clipping can only handle 16 overlay windows; if we
	   have more, we have do emulate the whole thing with mask-clipping */
	if ( clipcount > 16 ) {
/*
			printk("saa7146_core.o: clip_windows: emulate rect clipping\n");
			emulate_rect_clipping(saa, clipping, clipcount, x,y,w,h, width, height);
			calculate_clipping_registers_mask(saa, clipping, interlace, mode, width,height,&vdma2,&clip_format,&arbtr_ctrl);
			if( SAA7146_CLIPPING_RECT == type )
				type = SAA7146_CLIPPING_MASK;
			else
				type = SAA7146_CLIPPING_MASK_INVERTED;
*/	} else {
		calculate_clipping_registers_rect(saa, clipping_memory, interlace,clipcount,td_flip,x,y,w,h,width,height,&vdma2,&clip_format,&arbtr_ctrl);
	}

	/* set clipping format */
	clip_format &= 0xffff0008;
	clip_format |= (type << 4);

	/* prepare video dma2 */
	saa7146_write(saa, BASE_EVEN2,		vdma2.base_even);
	saa7146_write(saa, BASE_ODD2,		vdma2.base_odd);
	saa7146_write(saa, PROT_ADDR2,		vdma2.prot_addr);
	saa7146_write(saa, BASE_PAGE2,		vdma2.base_page);
	saa7146_write(saa, PITCH2,		vdma2.pitch);
	saa7146_write(saa, NUM_LINE_BYTE2,	vdma2.num_line_byte);
	
	/* prepare the rest */
	saa7146_write(saa, CLIP_FORMAT_CTRL,clip_format);
	saa7146_write(saa, PCI_BT_V1, arbtr_ctrl);	

	/* upload clip_control-register, clipping-registers, enable video dma2 */
	saa7146_write(saa, MC2, (MASK_05 | MASK_21 | MASK_03 | MASK_19));
	saa7146_write(saa, MC1, (MASK_05 | MASK_21));

	return 0;
}

#ifdef axa
int saa7146_set_clip(struct saa7146* saa, int interlace, u32* clipping, int width, int height, u32* mask, int clipcount, int x[], int y[], int w[], int h[]) 
{
	struct	saa7146_video_dma	vdma2;

	u32 clip_format	= saa7146_read(saa, CLIP_FORMAT_CTRL);
	u32 arbtr_ctrl	= saa7146_read(saa, PCI_BT_V1);

	/* some sanity checks first */
	if ( width <= 0 || height <= 0 ) {
		printk("saa7146_core.o: clip_windows: sanity error #1!\n");
		return -EINVAL;
	}

	/* check if anything to do here, disable clipping if == 0 */
	if( clipcount == 0 ) {
	
		/* mask out relevant bits (=lower word)*/
		clip_format &= MASK_W1;

		/* upload clipping-registers*/
		saa7146_write(saa, CLIP_FORMAT_CTRL,clip_format);
 		saa7146_write(saa, MC2, (MASK_05 | MASK_21));
 
 		/* disable video dma2 */
		saa7146_write(saa, MC1, (MASK_21));

		return 0;
	}

	switch(type) {
		
		case SAA7146_CLIPPING_MASK_INVERTED:
		case SAA7146_CLIPPING_MASK:
		{
			/* sanity check */
			if( NULL == mask ) {
				printk("saa7146_core.o: clip_windows: sanity error #1!\n");
				return -EINVAL;
			}
			
			/* copy the clipping mask to structure */
			memmove(clipping, mask, sizeof(CLIPPING_MEM_SIZE));
			/* set clipping registers */
			calculate_clipping_registers_mask(saa,clipping,interlace,width,height,&vdma2,&clip_format,&arbtr_ctrl);

			break;
		}

		case SAA7146_CLIPPING_RECT_INVERTED:
		case SAA7146_CLIPPING_RECT:
		{
			/* see if we have anything to do */
			if ( 0 == clipcount ) {
				return 0;	
			}

			/* sanity check */
			if( NULL == x || NULL == y || NULL == w || NULL == h ) {
				printk("saa7146_core.o: clip_windows: sanity error #2!\n");
				return -EINVAL;
			}

			/* rectangle clipping can only handle 16 overlay windows; if we
			   have more, we have do emulate the whole thing with mask-clipping */
			if ( clipcount > 16 /* > 16*/ ) {
			printk("emulate\n");
				emulate_rect_clipping(saa, clipping, clipcount, x,y,w,h, width, height);
				calculate_clipping_registers_mask(saa, clipping, interlace, width,height,&vdma2,&clip_format,&arbtr_ctrl);
				if( SAA7146_CLIPPING_RECT == type )
					type = SAA7146_CLIPPING_MASK;
				else
					type = SAA7146_CLIPPING_MASK_INVERTED;
				
			}
			else {
				calculate_clipping_registers_rect(saa, clipping, interlace,clipcount,x,y,w,h,width,height,&vdma2,&clip_format,&arbtr_ctrl);
			}
			
			break;
		}

		default:
		{
			printk("saa7146_core.o: clip_windows: internal error #1!\n");
			return -EINVAL;
		}
	
	}

	/* set clipping format */
	clip_format &= 0xffff0008;
	clip_format |= (type << 4);

	saa7146_write(saa, BASE_EVEN2,     vdma2.base_even);
	saa7146_write(saa, BASE_ODD2,      vdma2.base_odd);
	saa7146_write(saa, PROT_ADDR2,     vdma2.prot_addr);
	saa7146_write(saa, BASE_PAGE2,     vdma2.base_page);
	saa7146_write(saa, PITCH2,		vdma2.pitch);
	saa7146_write(saa, NUM_LINE_BYTE2,	vdma2.num_line_byte);

	saa7146_write(saa, CLIP_FORMAT_CTRL,clip_format);
	saa7146_write(saa, PCI_BT_V1, arbtr_ctrl);	

	/* upload clip_control-register, clipping-registers, enable video dma2 */
	saa7146_write(saa, MC2, (MASK_05 | MASK_21 | MASK_03 | MASK_19));
	saa7146_write(saa, MC1, (MASK_05 | MASK_21));
/*
       	printk("ARBTR_CTRL:     0x%08x\n",saa7146_read(saa, PCI_BT_V1));
       	printk("CLIP_FORMAT:    0x%08x\n",saa7146_read(saa, CLIP_FORMAT_CTRL));
        printk("BASE_ODD1:      0x%08x\n",saa7146_read(saa, BASE_ODD1));
        printk("BASE_EVEN1:     0x%08x\n",saa7146_read(saa, BASE_EVEN1));
        printk("PROT_ADDR1:     0x%08x\n",saa7146_read(saa, PROT_ADDR1));
        printk("PITCH1:         0x%08x\n",saa7146_read(saa, PITCH1));
        printk("BASE_PAGE1:     0x%08x\n",saa7146_read(saa, BASE_PAGE1));
        printk("NUM_LINE_BYTE1: 0x%08x\n",saa7146_read(saa, NUM_LINE_BYTE1));
        printk("BASE_ODD2:      0x%08x\n",saa7146_read(saa, BASE_ODD2));
        printk("BASE_EVEN2:     0x%08x\n",saa7146_read(saa, BASE_EVEN2));
        printk("PROT_ADDR2:     0x%08x\n",saa7146_read(saa, PROT_ADDR2));
        printk("PITCH2:         0x%08x\n",saa7146_read(saa, PITCH2));
        printk("BASE_PAGE2:     0x%08x\n",saa7146_read(saa, BASE_PAGE2));
        printk("NUM_LINE_BYTE2: 0x%08x\n",saa7146_read(saa, NUM_LINE_BYTE2));
*/
	return 0;

}		
#endif

/* calculate the new memory offsets for a desired position */
int saa7146_set_position(struct saa7146* saa, int interlace, int w_x, int w_y, int w_height, int b_width, int b_depth, int b_bpl, u32 base, int td_flip)
{	
	struct	saa7146_video_dma vdma1;

	if( w_y < 0 || w_height <= 0 || b_depth <= 0 || b_bpl <= 0 || base <= 0 ) {
		printk("saa7146_core.o: calculate_video_dma1_overlay: illegal values %d,%d,%d,%d,%d\n",w_y,w_height,b_depth,b_bpl,base);
		return -EINVAL;
	}	

	/* calculate memory offsets for picture, look if we shall top-down-flip */
	vdma1.pitch	= 2*b_bpl;
	if ( 0 == td_flip ) {
		vdma1.base_even = (u32)base + (w_y * (vdma1.pitch/2)) + (w_x * (b_depth / 8)); 
		vdma1.base_odd  = vdma1.base_even + (vdma1.pitch / 2);
		vdma1.prot_addr = vdma1.base_odd + (w_height * (vdma1.pitch / 2));
	}
	else {
		vdma1.base_even = (u32)base + ((w_y+w_height) * (vdma1.pitch/2)) + (w_x * (b_depth / 8)); 
		vdma1.base_odd  = vdma1.base_even - (vdma1.pitch / 2);
		vdma1.prot_addr = vdma1.base_odd - (w_height * (vdma1.pitch / 2));
	}
	
//	printk("w_x: %d, w_y:%d, w_h:%d, b_w:%d\n",w_x,w_y,w_height,b_width);
//	printk("e: %d, o:%d, prot:%d\n",vdma1.base_even-(u32)base,vdma1.base_odd-(u32)base,vdma1.prot_addr-(u32)base);
	
	/* if scaling is between 1/2 and 1/4 and interlace != 0 we only use
	   the even lines, the odd lines get discarded (see vertical scaling) */
	if( interlace != 0 ) {
		if( 	   (w_height*4 >= saa->modes_constants.v_calc)
			&& (w_height*2 <= saa->modes_constants.v_calc)) {
	
			vdma1.base_odd = vdma1.prot_addr;
			vdma1.pitch /= 2;
		}
	}

	if ( 0 != td_flip ) {
		vdma1.pitch *= -1;
	}
		
	vdma1.base_page = 0;
	vdma1.num_line_byte = (saa->modes_constants.v_field<<16)+saa->modes_constants.h_pixels;

	saa7146_write(saa, BASE_EVEN1,		vdma1.base_even);
	saa7146_write(saa, BASE_ODD1,		vdma1.base_odd);
	saa7146_write(saa, PROT_ADDR1,     	vdma1.prot_addr);
	saa7146_write(saa, BASE_PAGE1,     	vdma1.base_page);
	saa7146_write(saa, PITCH1,		vdma1.pitch);
	saa7146_write(saa, NUM_LINE_BYTE1,	vdma1.num_line_byte);

	/* update the video dma 1 registers */
      	saa7146_write(saa, MC2, (MASK_02 | MASK_18));
	
	return 0;

}

int saa7146_set_window(struct saa7146* saa, int interlace, int width, int height, int flip_lr, int source, int sync)
{
	u32 hps_v_scale = 0, hps_v_gain  = 0, hps_ctrl = 0, hps_h_prescale = 0, hps_h_scale = 0;

	/* set vertical scale */
	hps_v_scale = 0; /* all bits get set by the function-call */
	hps_v_gain  = 0; /* fixme: saa7146_read(saa, HPS_V_GAIN);*/ 
	calculate_v_scale_registers(saa, interlace, saa->modes_constants.v_calc, height, &hps_v_scale, &hps_v_gain);

	/* set horizontal scale */
	hps_ctrl 	= 0;
	hps_h_prescale	= 0; /* all bits get set in the function */
	hps_h_scale	= 0;
	calculate_h_scale_registers(saa, saa->modes_constants.h_calc, width, flip_lr, &hps_ctrl, &hps_v_gain, &hps_h_prescale, &hps_h_scale);

	/* set hyo and hxo */
	calculate_hxo_and_hyo(saa, &hps_h_scale, &hps_ctrl);
	calculate_hps_source_and_sync(saa, source, sync, &hps_ctrl);
	
	/* write out new register contents */
	saa7146_write(saa, HPS_V_SCALE,	hps_v_scale);
	saa7146_write(saa, HPS_V_GAIN,	hps_v_gain);
	saa7146_write(saa, HPS_CTRL,	hps_ctrl);
	saa7146_write(saa, HPS_H_PRESCALE,hps_h_prescale);
	saa7146_write(saa, HPS_H_SCALE,	hps_h_scale);
	
	/* upload shadow-ram registers */
      	saa7146_write(saa, MC2, (MASK_05 | MASK_06 | MASK_21 | MASK_22) );

	return 0;

}

void saa7146_set_output_format(struct saa7146* saa, u32 palette)
{
	u32	clip_format = saa7146_read(saa, CLIP_FORMAT_CTRL);
	
	/* call helper function */
	calculate_output_format_register(saa,palette,&clip_format);

	/* update the hps registers */
	saa7146_write(saa, CLIP_FORMAT_CTRL, clip_format);
      	saa7146_write(saa, MC2, (MASK_05 | MASK_21));
}

void saa7146_set_picture_prop(struct saa7146 *saa, int brightness, int contrast, int colour)
{
	u32	bcs_ctrl = 0;
	
	calculate_bcs_ctrl_register(saa, brightness, contrast, colour, &bcs_ctrl);
	saa7146_write(saa, BCS_CTRL, bcs_ctrl);
	
	/* update the bcs register */
      	saa7146_write(saa, MC2, (MASK_06 | MASK_22));
}

/* enable(1) / disable(0) video */
void saa7146_set_overlay(struct saa7146* saa, int v)
{
/*
	DEB_EE(printk("saa7146_core.o: saa7146_set_overlay; m:%d\n",v));
*/	
	/* disable ? */
	if(v==0) {
		/* disable video dma1 */
      		saa7146_write(saa, MC1, MASK_22);
	}
	/* or enable ? */
	else {
		/* fixme: enable video */
      		saa7146_write(saa, MC1, (MASK_06 | MASK_22));

	}
}		

/* select input-source */
void saa7146_set_hps_source_and_sync(struct saa7146* saa, int source, int sync)
{
	u32 hps_ctrl = 0;

	/* read old state */
	hps_ctrl = saa7146_read(saa, HPS_CTRL);

	hps_ctrl &= ~( MASK_31 | MASK_30 | MASK_28 );
	hps_ctrl |= (source << 30) | (sync << 28);

	/* write back & upload register */
	saa7146_write(saa, HPS_CTRL, hps_ctrl);
	saa7146_write(saa, MC2, (MASK_05 | MASK_21));
	
	saa->current_hps_source = source;
	saa->current_hps_sync = sync;
} 

int calculate_video_dma_grab_packed(struct saa7146* saa, int width, int height, int depth, int interlace, int flip_td, u32 phys_page_table, struct saa7146_video_dma* vdma1) 
{
	vdma1->pitch		= (width*depth*2)/8;
	vdma1->num_line_byte	= ((saa->modes_constants.v_field<<16) + saa->modes_constants.h_pixels);
	vdma1->base_page	= phys_page_table | ME1;

	if( 0 != flip_td ) {
		vdma1->prot_addr	= 0;
		vdma1->base_even	= (vdma1->pitch/2)*height;
		vdma1->base_odd		= vdma1->base_even - (vdma1->pitch/2);
	} else {
		vdma1->base_even	= 0;
		vdma1->base_odd		= vdma1->base_even + (vdma1->pitch/2);
		vdma1->prot_addr	= (vdma1->pitch/2)*height;
	}

	/* convention: if scaling is between 1/2 and 1/4 we only use
	   the even lines, the odd lines get discarded (see vertical scaling) */
	if( interlace != 0 && height*4 >= saa->modes_constants.v_calc && height*2 <= saa->modes_constants.v_calc) {
		vdma1->base_odd = vdma1->prot_addr;
		vdma1->pitch /= 2;
	}

	if( 0 != flip_td ) {
		vdma1->pitch *= -1;
	}	

	return 0;
}

int calculate_video_dma_grab_planar(struct saa7146* saa, int width, int height, int depth, int interlace, int flip_td,
	u32 phys_page_table1, struct saa7146_video_dma* vdma1,
	u32 phys_page_table2, struct saa7146_video_dma* vdma2,
	u32 phys_page_table3, struct saa7146_video_dma* vdma3 ) 
{
	printk("\ncalculate_video_dma_grab_planar() called.\n");

	vdma1->pitch		= width*2;
	vdma1->num_line_byte	= ((saa->modes_constants.v_field<<16) + saa->modes_constants.h_pixels);
	vdma1->base_page	= phys_page_table1 | ME1;

	/* convention: if scaling is between 1/2 and 1/4 we only use
	   the even lines, the odd lines get discarded (see vertical scaling) */
	if( interlace != 0 && height*4 >= saa->modes_constants.v_calc && height*2 <= saa->modes_constants.v_calc) {
		vdma1->base_odd = vdma1->prot_addr;
		vdma1->pitch /= 2;
	}

	if( 0 != flip_td ) {
		vdma1->pitch *= -1;
	}	

	return 0;
}


/* this function runs through the i2c-messages and prepares the data to be
   sent through the saa7146. have a look at the specifications p. 122 ff 
   to understand this. it returns the number of u32s to send, or -1
   in case of an error. */
int saa7146_i2c_msg_prepare(struct i2c_msg m[], int num, u32 *op)
{
	int h1, h2;
	int i, j, addr;
	int mem = 0, op_count = 0;

	/* first determine size of needed memory */
	for(i = 0; i < num; i++) {
		mem += m[i].len + 1;
	}

	/* worst case: we need one u32 for three bytes to be send
	   plus one extra byte to address the device */
	mem = 1 + ((mem-1) / 3);

	/* we assume that op points to a memory of at least SAA7146_I2C_MEM bytes
	   size. if we exceed this limit... */
	if ( (4*mem) > SAA7146_I2C_MEM ) {
		DEB_I2C(("saa7146_core.o: cannot prepare i2c-message..\n"));
		return -1;
	}

	/* be careful: clear out the i2c-mem first */
	memset(op,0,sizeof(u32)*mem);

	/* loop through all messages */
	for(i = 0; i < num; i++) {

		/* insert the address of the i2c-slave.
		   note: we get 7-bit-i2c-addresses, so we have to perform a translation */
		addr = (m[i].addr*2) + ( (0 != (m[i].flags & I2C_M_RD)) ? 1 : 0); 
		h1 = op_count/3; h2 = op_count%3;
		op[h1] |= (	    (u8)addr << ((3-h2)*8));
		op[h1] |= (SAA7146_I2C_START << ((3-h2)*2));
		op_count++;

		/* loop through all bytes of message i */
		for(j = 0; j < m[i].len; j++) {
		
			/* insert the data bytes */
			h1 = op_count/3; h2 = op_count%3;
			op[h1] |= ( (u32)((u8)m[i].buf[j]) << ((3-h2)*8));
			op[h1] |= (       SAA7146_I2C_CONT << ((3-h2)*2));
			op_count++;
		}
		
	}			

	/* have a look at the last byte inserted:
	  if it was: ...CONT change it to ...STOP */
	h1 = (op_count-1)/3; h2 = (op_count-1)%3;
	if ( SAA7146_I2C_CONT == (0x3 & (op[h1] >> ((3-h2)*2))) ) {
//		printk("stop inserted!\n");
		op[h1] &= ~(0x2 << ((3-h2)*2));
		op[h1] |= (SAA7146_I2C_STOP << ((3-h2)*2));
	} else {
//		printk("nothin!\n");
	}

	/* return the number of u32s to send */	
	return mem;
}

/* this functions loops through all i2c-messages. normally, it should determine
   which bytes were read through the adapter and write them back to the corresponding
   i2c-message. but instead, we simply write back all bytes.
   
   fixme: this could be improved. */
int saa7146_i2c_msg_cleanup(struct i2c_msg m[], int num, u32 *op)
{
	int i, j;
	int op_count = 0;

	/* loop through all messages */
	for(i = 0; i < num; i++) {

		op_count++;

		/* loop throgh all bytes of message i */
		for(j = 0; j < m[i].len; j++) {
			/* write back all bytes that could have been read */
			m[i].buf[j] = (op[op_count/3] >> ((3-(op_count%3))*8));
			op_count++;
		}
	}
	
	return 0;
}

/* this functions returns the status-register of our i2c-device */
u32 saa7146_i2c_status(struct saa7146* saa) 
{
	u32 iicsta = saa7146_read(saa, I2C_STATUS);
	DEB_I2C(("saa7146_core.o: saa7146_i2c_status: 0x%08x\n",iicsta));
	return iicsta;
}

/* this functions resets the i2c-device and returns 0 if everything was fine, otherwise -1 */
int saa7146_i2c_reset(struct saa7146* saa) 
{
	/* get current status */
	u32 status = saa7146_i2c_status(saa);
	
	/* clear registers for sure */
	saa7146_write(saa, I2C_STATUS, saa->i2c_bitrate);
	saa7146_write(saa, I2C_TRANSFER, 0);

	/* check if any operation is still in progress */
	if ( 0 != ( status & SAA7146_I2C_BUSY) ) {

		/* yes, kill ongoing operation */
		DEB_I2C(("saa7146_core.o: i2c_reset: busy_state detected.\n"));

		/* set "ABORT-OPERATION"-bit (bit 7)*/
		saa7146_write(saa, I2C_STATUS, (saa->i2c_bitrate | MASK_07));
		saa7146_write(saa, MC2, (MASK_00 | MASK_16));
		my_wait(saa,SAA7146_I2C_DELAY);

		/* clear all error-bits pending; this is needed because p.123, note 1 */
		saa7146_write(saa, I2C_STATUS, saa->i2c_bitrate);
		saa7146_write(saa, MC2, (MASK_00 | MASK_16));
		my_wait(saa,SAA7146_I2C_DELAY);
 	}

	/* check if any error is (still) present. (this can be necessary because p.123, note 1) */
	status = saa7146_i2c_status(saa);

	if ( saa->i2c_bitrate != status ) {

		DEB_I2C(("saa7146_core.o: i2c_reset: error_state detected. (0x%08x)\n",status));

		/* Repeat the abort operation. This seems to be necessary
		   after serious protocol errors caused by e.g. the SAA7740 */
		saa7146_write(saa, I2C_STATUS, (saa->i2c_bitrate | MASK_07));
		saa7146_write(saa, MC2, (MASK_00 | MASK_16));
		my_wait(saa,SAA7146_I2C_DELAY);

		/* clear all error-bits pending */
		saa7146_write(saa, I2C_STATUS, saa->i2c_bitrate);
		saa7146_write(saa, MC2, (MASK_00 | MASK_16));
		my_wait(saa,SAA7146_I2C_DELAY);

		/* the data sheet says it might be necessary to clear the status
		   twice after an abort */
		saa7146_write(saa, I2C_STATUS, saa->i2c_bitrate);
		saa7146_write(saa, MC2, (MASK_00 | MASK_16));
		my_wait(saa,SAA7146_I2C_DELAY);
     	}

	/* if any error is still present, a fatal error has occured ... */
	status = saa7146_i2c_status(saa);
	if ( saa->i2c_bitrate != status ) {
		DEB_I2C(("saa7146_core.o: i2c_reset: i2c_reset: fatal error, status:0x%08x\n",status));
		return -1;
	}

	return 0;
}

/* this functions writes out the data-byte 'dword' to the i2c-device.
   it returns 0 if ok, -1 if the transfer failed, -2 if the transfer
   failed badly (e.g. address error) */
int saa7146_i2c_writeout(struct saa7146* saa, u32* dword)
{
	int i = 0;
	u32 status = 0, mc2 = 0;

	DEB_I2C(("saa7146_core.o: writeout: 0x%08x (before).\n",*dword));
	
	/* write out i2c-command */
	saa7146_write(saa, I2C_STATUS,	 saa->i2c_bitrate);
	saa7146_write(saa, I2C_TRANSFER, *dword);
	saa7146_write(saa, MC2, (MASK_00 | MASK_16));

	/* do not poll for i2c-status before upload is complete */
	for (i = 5; i > 0; i--) {
		mc2 = (saa7146_read(saa, MC2) & 0x1);
		if( 0 != mc2 )
			break;
		my_wait(saa,SAA7146_I2C_DELAY);
	}
	if (0 == i) {
		DEB_I2C(("saa7146_core.o: saa7146_i2c_writeout: timeout error. #1\n"));
		return -1;
	}

	/* wait until busy flag becomes inactive or we time out */
	for (i = 5; i > 0; i--) {
		status = saa7146_i2c_status(saa);
		/* check busy flag */
		if ( 0 == (status & SAA7146_I2C_BUSY))
			break;
		/* check error flag */
		if ( 0 != (status & SAA7146_I2C_ERR)) {
			break;
		}
		my_wait(saa,SAA7146_I2C_DELAY);
	}
	if (0 == i) {
		DEB_I2C(("saa7146_core.o: saa7146_i2c_writeout: timeout error. #2\n"));
		return -1;
	}

	/* give a detailed status report */
	if ( 0 != (status & SAA7146_I2C_ERR)) {

		if( 0 != (status & SAA7146_I2C_SPERR) ) {
			DEB_I2C(("saa7146_core.o: saa7146_i2c_writeout: error due to invalid start/stop condition.\n"));
		}
		if( 0 != (status & SAA7146_I2C_DTERR) ) {
			DEB_I2C(("saa7146_core.o: saa7146_i2c_writeout: error in data transmission.\n"));
		}
		if( 0 != (status & SAA7146_I2C_DRERR) ) {
			DEB_I2C(("saa7146_core.o: saa7146_i2c_writeout: error when receiving data.\n"));
		}
		if( 0 != (status & SAA7146_I2C_AL) ) {
			DEB_I2C(("saa7146_core.o: saa7146_i2c_writeout: error because arbitration lost.\n"));
		}

		/* we handle address-errors here */
		if( 0 != (status & SAA7146_I2C_APERR) ) {
			DEB_I2C(("saa7146_core.o: saa7146_i2c_writeout: error in address phase.\n"));
			return -2;
		}
		return -1;
	}

	/* read back data, just in case we were reading ... */
	*dword = saa7146_read(saa, I2C_TRANSFER);

	DEB_I2C(("saa7146_core.o: saa7146_i2c_writeout: 0x%08x (after).\n",*dword));
	
	saa7146_i2c_status(saa);
//	printk("saa7146_i2c_writeout() exit.\n");

	return 0;
}

int saa7146_i2c_transfer(struct saa7146* saa, struct i2c_msg msgs[], int num,  int retries)
{
	int i = 0, result = 0, count = 0;
	u32* buffer = saa->i2c_mem;

/*
	if(saa->dummy == 1) {
		int i = 0, j = 0;
		printk("num msgs: %d\n",num);
		for(i = 0; i < num; i++) {
		printk("%d: addr: 0x%02x, len: %d: ==> ",i,msgs[i].addr,msgs[i].len);
		for(j = 0; j<msgs[i].len;j++) {
			printk("0x%02x ",(u8)msgs[i].buf[j]);	
		}
		printk("\n");
		return -1;
		}
	}
*/
	/* prepare the message(s), get number of u32s to transfer */
	count = saa7146_i2c_msg_prepare(msgs, num, buffer);
	if ( 0 > count ) {
		DEB_I2C(("saa7146_core.o: saa7146_xfer: could not prepare i2c-message.\n"));
		return -1;
	}
	
	/* loop through number of retries ... */
	for(; retries >= 0; retries--) {
	
		/* reset the i2c-device if necessary */
		result = saa7146_i2c_reset(saa);
		if ( 0 > result ) {
			DEB_I2C(("saa7146_core.o: saa7146_i2c_transfer: could not reset i2c-device.\n"));
			return -1;
		}	

		/* write out the u32s one after another */
		for(i = 0; i < count; i++) {

//			printk("0x%08x\n",saa7146_read(saa, 0xfc ));
	
			result = saa7146_i2c_writeout( saa, &buffer[i] );
			if ( 0 != result) {
				/* if address-error occured, dont retry */
				if ( -2 == result ) {
					return -1;
				}
				DEB_I2C(("saa7146_core.o: saa7146_i2c_transfer: error while sending message(s). starting again.\n"));
				break;
			}
		}
		
		/* see if an error occured & the last retry failed */
		if( (0 != result) && (0 == retries) ) {
			DEB_I2C(("saa7146_core.o: saa7146_xfer: could not transfer i2c-message(s).\n"));
			return -1;
		}
		
		if( 0 == result )
			break;
	}
	
	/* if any things had to be read, get the results */
	if ( 0 != saa7146_i2c_msg_cleanup(msgs, num, buffer)) {
		DEB_I2C(("saa7146_core.o: saa7146_xfer: could not cleanup i2c-message.\n"));
		return -1;
	}

	saa7146_i2c_status(saa);
//	printk("saa7146_i2c_transfer() exit.\n");
	return 0;
}

/* write "data" to the gpio-pin "pin" */
void	gpio_set(struct saa7146* saa, u8 pin, u8 data)
{
	u32 value = 0;

	/* sanity check */
	if(pin > 3)
		return;

	/* read old register contents */
	value = saa7146_read(saa, GPIO_CTRL );
	
	value &= ~(0xff << (8*pin));
	value |= (data << (8*pin));

	saa7146_write(saa, GPIO_CTRL, value);
}

#ifdef axa
/* functions for accessing the debi-port. note: we currently dont support page-table-
   transfers. */

#define MY_DEBI_TIMEOUT_MS 5

int	debi_transfer(struct saa7146* saa, struct saa7146_debi_transfer* dt)
{
	u32	debi_config = 0, debi_command = 0, debi_page = 0, debi_ad = 0;
	u32	timeout = MY_DEBI_TIMEOUT_MS;

	/* sanity checks */
	if(dt->direction > 1 || dt->timeout > 15 || dt->swap > 3 || dt->slave16 > 2 || dt->intel > 1 || dt->increment > 1 || dt->tien > 1 )
		return -EINVAL;

	debi_page	= 0;
	/* keep bits 31,30,28 clear */
	debi_config	= (dt->timeout << 22) | (dt->swap << 20) | (dt->slave16 << 19) | (dt->increment << 18) | (dt->intel << 17) | (dt->tien << 16);
	debi_command	= (dt->num_bytes << 17) | (dt->direction << 16) | (dt->address << 0);
	debi_ad		= dt->mem;

	saa7146_write(saa, DEBI_PAGE,	debi_page);	
	saa7146_write(saa, DEBI_CONFIG,	debi_config);	
	saa7146_write(saa, DEBI_COMMAND,	debi_command);	
	saa7146_write(saa, DEBI_AD,	debi_ad);	

	/* upload debi-registers */
	saa7146_write(saa, MC2, (MASK_01|MASK_17));

	/* wait for DEBI upload to complete */
	while (! (saa7146_read(saa, MC2) & 0x2));

	while( --timeout ) {
		/* check, if DEBI still active */
		u32 psr = saa7146_read(saa, PSR);
		if (0 !=  (psr & SPCI_DEBI_S)) {
			/* check, if error occured */
/*			if ( 0 != (saa7146_read(saa, SSR) & (MASK_23|MASK_22))) { */
			if ( 0 != (saa7146_read(saa, SSR) & (MASK_22))) {
				/* clear error status and indicate error */
				saa7146_write(saa, ISR, SPCI_DEBI_E);
				return -1;
			}
		 }
		else {
			/* Clear status bit */
			saa7146_write(saa, ISR, SPCI_DEBI_S);
			break;
		}
		/* I dont know how we should actually wait for the debi to have finished.
		   we simply wait 1ms here and then check in a loop for max. MY_DEBI_TIMEOUT_MS */		
		mdelay(1);
	}

	/* check for timeout */
	if( 0 == timeout ) {
		return -1;
	}

	/* read back data if we did immediate read-transfer */
	if(dt->num_bytes <= 4 && dt->direction == 1) {
		dt->mem = saa7146_read(saa, DEBI_AD);
		switch(dt->num_bytes) {
			case 1:
			 dt->mem &= 0x000000ff;
			break;
			case 2:
			 dt->mem &= 0x0000ffff;
			break;
			case 3:
			 dt->mem &= 0x00ffffff;
			break;
		}		
	}
	
	return 0;
}
#endif

