/*
    saa7146_v4l2.c - v4l2 core-driver for saa7146-based videohardware
    
    Copyright (C) 1998-2001 Michael Hunold <michael@mihu.de>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <linux/module.h>	/* for module-version */
#include <linux/delay.h>	/* for delay-stuff */
#include <linux/slab.h>		/* for kmalloc/kfree */
#include <linux/pci.h>		/* for pci-config-stuff, vendor ids etc. */
#include <linux/wrapper.h>	/* for mem_map_reserve */
#include <linux/init.h>		/* for "__init" */
#include <linux/interrupt.h>	/* for IMMEDIATE_BH */

#include <linux/i2c.h>
#include <linux/i2c-id.h>

#include "saa7146_v4l2.h"

/* fixme: this belongs to i2c-id.h */
#define I2C_HW_S_SAA7146 1

/* different debug options */
#define DEB_S(x)	if (0!=(debug&0x01)) printk x	/* simple debug messages */
#define DEB_D(x)	if (0!=(debug&0x02)) printk x	/* more detailed debug messages */
#define DEB_EE(x)	if (0!=(debug&0x04)) printk x	/* print enter and exit of functions */
#define DEB_I2C(x)	if (0!=(debug&0x08)) printk x	/* i2c debug messages */
#define DEB_VBI(x)	if (0!=(debug&0x10)) printk x	/* vbi debug messages */
#define DEB_INT(x)	if (0!=(debug&0x20)) printk x	/* interrupt debug messages */
#define DEB_CAP(x)	if (0!=(debug&0x40)) printk x	/* capture debug messages */

/* module parameters */
int no_interlace = 1;
int debug = 0;
int memory = 32;

#ifdef MODULE
MODULE_PARM(memory,"i");
MODULE_PARM_DESC(memory, "maximum memory usage for capture buffers (default: 32Mb)");

MODULE_PARM(debug,"i");
MODULE_PARM_DESC(debug, "debug level (default: 0)");

MODULE_PARM(no_interlace,"i");
MODULE_PARM_DESC(no_interlace, "try to use only one field if picture size is small enough (default:1)");
#endif


/* global variables */
int saa7146_v4l2_num = 0;

struct capture_device saa7146_v4l2_devices[MAX_SAA7146_BOARDS];

#define saa7146 capture_device
#include "saa7146.c"

int xxxflags = ME1;

/*****************************************************************************/

struct saa7146_modes_constants saa7146_standards[3] = {
	{ V_OFFSET_PAL,		V_FIELD_PAL,	V_ACTIVE_LINES_PAL,
	  H_OFFSET_PAL,		H_PIXELS_PAL,	H_PIXELS_PAL+1,
	  V_ACTIVE_LINES_PAL,	768 },	/* PAL values */
	{ V_OFFSET_NTSC,	V_FIELD_NTSC,	V_ACTIVE_LINES_NTSC,
	  H_OFFSET_NTSC,	H_PIXELS_NTSC,  H_PIXELS_NTSC+1,
	  V_ACTIVE_LINES_NTSC,	640 },	/* NTSC values */
	{ V_OFFSET_SECAM,	V_FIELD_SECAM,	V_ACTIVE_LINES_SECAM,
	  H_OFFSET_SECAM,	H_PIXELS_SECAM,	H_PIXELS_SECAM+1,
	  V_ACTIVE_LINES_SECAM,	768 }	/* SECAM values */
};

#define SAA7146_PAL	0
#define SAA7146_NTSC	1
#define SAA7146_SECAM	2

/*****************************************************************************/
/* i2c-adapter helper functions                                              */

/* these are just for completeness */
int saa7146_i2c_reg(struct i2c_client *client)
{
	return 0;
}

int saa7146_i2c_unreg(struct i2c_client *client)
{
	return 0;
}

void saa7146_i2c_inc_use(struct i2c_adapter *adap)
{
#ifdef MODULE
	MOD_INC_USE_COUNT;
#endif
}

void saa7146_i2c_dec_use(struct i2c_adapter *adap)
{
#ifdef MODULE
	MOD_DEC_USE_COUNT;
#endif
}

int saa7146_i2c_ioctl(struct i2c_adapter *adapter, unsigned int cmd, unsigned long arg)
{
	DEB_I2C(("saa7146_v4l2.o: saa7146_i2c_ioctl for adapter %s called.\n", adapter->name));
	return 0;
}

u32 saa7146_i2c_func(struct i2c_adapter *adapter)
{
	DEB_I2C(("saa7146_v4l2.o: saa7146_i2c_func for adapter %s called.\n", adapter->name));

	return	  I2C_FUNC_I2C
		| I2C_FUNC_SMBUS_QUICK
		| I2C_FUNC_SMBUS_READ_BYTE	| I2C_FUNC_SMBUS_WRITE_BYTE
		| I2C_FUNC_SMBUS_READ_BYTE_DATA | I2C_FUNC_SMBUS_WRITE_BYTE_DATA;
}

void i2c_status(struct capture_device* saa) 
{
	u32 iicsta = saa7146_read(saa, I2C_STATUS);
//	printk("==> saa7146_i2c_status: 0x%08x\n",iicsta);
	iicsta = saa7146_read(saa, I2C_TRANSFER);
//	printk("==> saa7146_i2c_transf: 0x%08x\n",iicsta);
}

/* utility functions */
int saa7146_i2c_xfer(struct i2c_adapter* adapter, struct i2c_msg msg[], int num)
{
	struct capture_device* dev = (struct capture_device*)adapter->data;
					
	DEB_I2C(("saa7146_v4l2.o: saa7146_i2c_xfer for adapter %s called.\n", adapter->name));

	/* use helper function to transfer data */
	if( 0 != saa7146_i2c_transfer(dev, msg, num, adapter->retries)) {
		return -1;
	}

	/* another bug in revision 0: the i2c-registers get uploaded randomly by other
	   uploads, so we better clear them out before continueing */	
	if( 0 == dev->revision ) {
		i2c_status(dev);
		saa7146_write(dev, I2C_STATUS,   dev->i2c_bitrate);
		saa7146_write(dev, I2C_TRANSFER, 0);
		saa7146_write(dev, MC2, (MASK_00 | MASK_16));
		mdelay(5);
		i2c_status(dev);
	}

	/* return the number of delivered messages */
	return num;	
}

/* exported algorithm data */
struct i2c_algorithm saa7146_algo = {
	"saa7146 i2c-algorithm",
	I2C_ALGO_SAA7146,
	saa7146_i2c_xfer,
	NULL,
	NULL,					/* slave_xmit		*/
	NULL,					/* slave_recv		*/
	saa7146_i2c_ioctl,			/* ioctl		*/
	saa7146_i2c_func,			/* functionality	*/
};

/*********************************************************************************/
/* v4l2-structures                                                               */

/* fixme: the saa7146 supports much, much more ... any volunteers? */

/* format descriptions for capture and preview */
struct v4l2_fmtdesc cappvfmt[] = 
{
	{  0, {"RGB-8 (3-3-2)"},	V4L2_PIX_FMT_RGB332,	0, 8, {0, 0}, },
	{  1, {"RGB-16 (5-6-5)"},	V4L2_PIX_FMT_RGB565,	0, 16, {0, 0}, },
	{  2, {"RGB-24 (B-G-R)"},	V4L2_PIX_FMT_BGR24,	0, 24, {0, 0}, },
	{  3, {"RGB-32 (B-G-R)"},	V4L2_PIX_FMT_BGR32,	0, 32, {0, 0}, },
	{  4, {"Greyscale-8"},		V4L2_PIX_FMT_GREY,	V4L2_FMT_CS_601YUV, 8, {0, 0}, },
	{  5, {"YUV 4:2:2 (U-Y-V-Y)"},	V4L2_PIX_FMT_UYVY,	V4L2_FMT_CS_601YUV, 16, {0, 0}, },
	/* V4L2_PIX_FMT_YUYV must always be the last format in the list. the saa7146 (without "a")
	   is not capable of on-the-fly "byte swapping" (see below), so it does not support this 
	   format... */
	{  6, {"YUV 4:2:2 (Y-U-Y-V)"},	V4L2_PIX_FMT_YUYV,	V4L2_FMT_CS_601YUV, 16, {0, 0}, },
};

/* this is only true for saa7146a. for saa7146 (without "a") this is NUM_CAPPVFMT-1 */
int NUM_CAPPVFMT = sizeof(cappvfmt)/sizeof(struct v4l2_fmtdesc);

struct v4l2_queryctrl capture_control[] = {
 { V4L2_CID_BRIGHTNESS, "Brightness", 		0, 255, 1, 128, V4L2_CTRL_TYPE_INTEGER},
 { V4L2_CID_CONTRAST,   "Contrast",   		0, 127, 1, 64,  V4L2_CTRL_TYPE_INTEGER},
 { V4L2_CID_SATURATION, "Saturation",	 	0, 127, 1, 64,  V4L2_CTRL_TYPE_INTEGER},
 { V4L2_CID_HFLIP, 	"Horizontal flip", 	0, 1,	1, 0,   V4L2_CTRL_TYPE_BOOLEAN},
 { V4L2_CID_VFLIP, 	"Vertical flip", 	0, 1,	1, 0,   V4L2_CTRL_TYPE_BOOLEAN},
};

int MAXCONTROLS = sizeof(capture_control)/sizeof(struct v4l2_queryctrl);

#define VCTRL_BRIGHTNESS	0
#define VCTRL_CONTRAST		1
#define VCTRL_SATURATION	2
#define VCTRL_HFLIP		3
#define VCTRL_VFLIP		4

void dump_registers(struct capture_device* dev)
{
	int i = 0;
	
	printk("dumping register @ %li jiffies:\n",jiffies);
	for(i = 0; i <= 0x148; i+=4) {
		printk("0x%03x: 0x%08x\n",i,saa7146_read(dev,i));
	}
}

/********************************************************************************/
/* helper-functions (with minor changes) taken from Bill Dirks sample driver(s) */

int find_vctrl(int id)
{
        int i;

	if (id == V4L2_CID_PRIVATE_BASE ||
	    id <  V4L2_CID_BASE ||
            id >  V4L2_CID_LASTP1)
		return -EDOM;

        for (i = MAXCONTROLS - 1; i >= 0; i--)
                if (capture_control[i].id == id)
                        break;

	if (i < 0)
		i = -EINVAL;

        return i;
}

int bm_build_scatter_list_helper(struct capture_device *dev, u8* mem_ptr, long length, u32** page_table)
{
	int i, n;
	u32* dma_list_ptr = NULL;

	if( NULL == mem_ptr ) {
		printk("saa7146_v4l2.o: bm_build_scatter_list(): fatal error: mem_ptr == NULL\n");
		return -1;
	}
	if( 0 == length ) {
		printk("saa7146_v4l2.o: bm_build_scatter_list(): fatal error: length == NULL\n");
		return -1;
	}
	if( NULL != *page_table ) {
		printk("saa7146_v4l2.o: bm_build_scatter_list(): fatal error: page_table != NULL\n");
		return -1;
	}

	*page_table = (u32*)get_free_page(GFP_KERNEL);
	if( NULL == *page_table ) {
		printk("saa7146_v4l2.o: bm_build_scatter_list: cannot allocate page table buffer.\n");
		return -1;
	}

	dma_list_ptr = *page_table;

	/*  Simple algorithm will just map the buffer contiguously by pages */
	/*  Note: capture_buffer is vmalloc()ed, so it's page-aligned */

	n = length / PAGE_SIZE; /* number of pages */

	for (i = 0; i < n; ++i) {
		dma_list_ptr[i] = v4l2_vmalloc_to_bus(mem_ptr);
		mem_ptr += PAGE_SIZE;
	}

	/* hack: fill page-table with the last value for safety reasons */
	for(;i<1024;i++) {
		dma_list_ptr[i] = dma_list_ptr[n-1];
	}

	return 0;
}

int bm_build_scatter_list(struct capture_device *dev, struct stream_buffer* buf)
{
	return(bm_build_scatter_list_helper(dev, buf->vaddress, buf->vidbuf.length, &buf->page_table));
}


/* A little integer square root routine */
int isqrt(unsigned int q)
{
	int		i;
	int		r;
	unsigned int	b2	= 0x40000000;
	unsigned int	t;

	for (i = 16, r = 0; i > 0 && q; --i)
	{
		t = ((unsigned int)r << i) + b2;
		if (t <= q)
		{
			q -= t;
			r |= (1 << (i - 1));
		}
		b2 >>= 2;
	}
	return r;
}

/********************************************************************************/
/* capturing helper-functions                                                   */

int unallocate_single_buffer(struct capture_device *dev, struct stream_buffer* buf)
{
	DEB_EE(("saa7146_v4l2.o: unallocate_single_buffer called: %p\n",buf));

	if (buf->vaddress != NULL) {
		vfree(buf->vaddress);
		buf->vaddress = NULL;
	}

	if (NULL != buf->page_table) {
		free_page((unsigned long)buf->page_table);
		buf->page_table = NULL;
	}
			
	buf->vidbuf.flags = 0;

	return 0;
}

int allocate_single_buffer(struct capture_device *dev, struct stream_buffer* buf)
{
	DEB_EE(("saa7146_v4l2.o: allocate_single_buffer called: %p\n",buf));

	if (buf->vaddress != NULL || NULL != buf->page_table) {
		printk("saa7146_v4l2.o: allocate_single_buffer: buffer %p is not free.\n",buf);
		return -1;
	}
		
	buf->vaddress = vmalloc(buf->vidbuf.length);
	if (buf->vaddress == NULL) {
		DEB_S(("saa7146_v4l2.o: allocate_capture_buffer: cannot allocate capture buffer.\n"));
		return -ENOMEM;
	}
	memset(buf->vaddress,0x00,buf->vidbuf.length);

	/* build up scatter data list */	
	if (0 != bm_build_scatter_list(dev,buf)) {
		DEB_S(("saa7146_v4l2.o: allocate_capture_buffer: cannot not build scatter list.\n"));
		return -ENOMEM;
	}
	
	return 0;
}

int build_read_capture_buffer(struct capture_device *dev)
{
	struct stream_buffer *sbuf = &dev->read_buffer;

	DEB_EE(("saa7146_v4l2.o: build_read_capture_buffer() called.\n"));

	/* free read buffer if possible */
	if( 0 != unallocate_single_buffer(dev,sbuf)) {
		return -1;
	}

	memset(sbuf, 0x0, sizeof(struct stream_buffer));

	sbuf->requested = 1;
	sbuf->vidbuf.type = V4L2_BUF_TYPE_CAPTURE;
	sbuf->vidbuf.length = (dev->clientfmt.sizeimage + PAGE_SIZE - 1) & ~(PAGE_SIZE - 1);

	if( 0 != allocate_single_buffer(dev,sbuf)) {
		return -1;
	}
	
	return 0;
}

int allocate_v4l_buffers(struct capture_device *dev)
{
	struct stream_buffer* buf = &dev->stream_buf[0];
	int i = 0;

	DEB_EE(("saa7146_v4l2.o: allocate_v4l_buffers called.\n"));

	buf->vaddress = vmalloc(dev->stream_buffers_requested * buf->vidbuf.length);
	if (buf->vaddress == NULL) {
		DEB_S(("saa7146_v4l2.o: allocate_v4l_buffers: cannot allocate capture buffer.\n"));
		return -ENOMEM;
	}
//	printk("total length: %d, adr:%p\n", dev->stream_buffers_requested * buf->vidbuf.length,buf->vaddress);
	
	if (0 != bm_build_scatter_list(dev,buf)) {
		DEB_S(("saa7146_v4l2.o: allocate_v4l_buffers: cannot not build scatter list #1.\n"));
		return -ENOMEM;
	}
	buf->vidbuf.flags |= V4L2_BUF_FLAG_MAPPED;

	for (i = 1; i < dev->stream_buffers_requested; ++i) {
		buf = &dev->stream_buf[i];
		buf->vaddress = dev->stream_buf[0].vaddress + buf->vidbuf.length;
//		printk("%d: %p (%d) \n",i,buf->vaddress, buf->vidbuf.length);
		if (0 != bm_build_scatter_list(dev,buf)) {
			// fixme: fix up here first
			DEB_S(("saa7146_v4l2.o: allocate_v4l_buffers: cannot not build scatter list #2.\n"));
			return -ENOMEM;
		}
		buf->vidbuf.flags |= V4L2_BUF_FLAG_MAPPED;
	}

	return 0;
}

int unallocate_v4l_buffers(struct capture_device *dev)
{
	struct stream_buffer* buf = &dev->stream_buf[0];
	int i = 0;

	DEB_EE(("saa7146_v4l2.o: unallocate_v4l_buffers called.\n"));

	if (buf->vaddress != NULL) {
		vfree(buf->vaddress);
		buf->vaddress = NULL;
	}
	if (NULL != buf->page_table) {
		free_page((unsigned long)buf->page_table);
		buf->page_table = NULL;
	}
	buf->vidbuf.flags &= ~V4L2_BUF_FLAG_MAPPED;

	for (i = 1; i < dev->stream_buffers_requested; i++) {
		buf = &dev->stream_buf[i];
		if (NULL != buf->page_table) {
			free_page((unsigned long)buf->page_table);
			buf->page_table = NULL;
		}
		buf->vaddress = NULL;
		buf->vidbuf.flags &= ~V4L2_BUF_FLAG_MAPPED;
	}

	return 0;
}


/*  Stop the music! */
void capture_abort(struct capture_device *dev)
{
	u32 flags = 0;

	DEB_EE(("saa7146_v4l2.o: capture_abort called.\n"));

	save_flags(flags);
	cli();

	if( 0 != dev->revision ) {
		/* revision != 0 capturing: turn off rps, video-dma1 */
		saa7146_write(dev, MC1, MASK_28|MASK_22);	
	} else {
		/* disable FIDA/B irqs */
		dev->irq_mask &= ~(MASK_07|MASK_08);	
		saa7146_write(dev, IER, dev->irq_mask);

		/* shut down video-dma1 */
		saa7146_write(dev, MC1, MASK_22);
	}

 	/* clear out any rps-signals pending */
	saa7146_write(dev, MC2, 0xf8000000);		

	restore_flags(flags);
}

int program_capture_engine(struct capture_device *dev)
{
	int count = 0, begin = 0, i = 0;
	
	DEB_EE(("saa7146_v4l2.o: program_capture_engine called.\n"));

	if( 0 != dev->revision ) {

		capture_abort(dev);

		/* erase rps0-program-memory for sure */
		for(count = 0; count < 1024; count++) {
			dev->rps0[ count ] = CMD_STOP;
		}

		/* write beginning of rps-program */
		count = 0;

		/* wait for o_fid_a/b / e_fid_a/b toggle */
		dev->rps0[ count++ ] = CMD_PAUSE | ( dev->current_hps_sync == SAA7146_HPS_SYNC_PORT_A ? MASK_12 : MASK_14);	/* wait for o_fid_a/b */
		dev->rps0[ count++ ] = CMD_PAUSE | ( dev->current_hps_sync == SAA7146_HPS_SYNC_PORT_A ? MASK_11 : MASK_13);	/* wait for e_fid_a/b */

		begin = count;
		for(i = 0; i < 2; i++) {			
			/* check signal x, jump if set */
			dev->rps0[ count++ ] = CMD_JUMP  | (1 << (21+i));		
			dev->rps0[ count++ ] = virt_to_bus(&dev->rps0[40*(i+1)]	);
		}

		dev->rps0[ count++ ] = CMD_PAUSE | ( dev->current_hps_sync == SAA7146_HPS_SYNC_PORT_A ? MASK_12 : MASK_14);	/* wait for o_fid_a/b */
		dev->rps0[ count++ ] = CMD_PAUSE | ( dev->current_hps_sync == SAA7146_HPS_SYNC_PORT_A ? MASK_11 : MASK_13);	/* wait for e_fid_a/b */

		/* generate interrupt */
		dev->rps0[ count++ ] = CMD_INTERRUPT;					
		/* jump to the "beginning" */
		dev->rps0[ count++ ] = CMD_JUMP;					
		dev->rps0[ count++ ] = virt_to_bus(&dev->rps0[begin]);
	
		for(i = 0; i < 2; i++) {		
		
			count = 40*(i+1);

			dev->rps0[ count++ ] = CMD_WR_REG | (1 << 8) | BASE_PAGE1/4;
			dev->stream_db_adr[i] = &dev->rps0[count];
			dev->rps0[ count++ ] = 0x0;

			/* upload video-dma1 */
			dev->rps0[ count++ ] = CMD_UPLOAD | MASK_02;			

			/* turn on video-dma1 */
			dev->rps0[ count++ ] = CMD_WR_REG_MASK | (MC1/4);		
			dev->rps0[ count++ ] = MASK_06 | MASK_22;	    		/* => mask */
			dev->rps0[ count++ ] = MASK_06 | MASK_22;			/* => values */

			/* wait for o_fid_a/b / e_fid_a/b toggle */
			dev->rps0[ count++ ] = CMD_PAUSE | ( dev->current_hps_sync == SAA7146_HPS_SYNC_PORT_A ? MASK_12 : MASK_14);	/* wait for o_fid_a/b */
			dev->rps0[ count++ ] = CMD_PAUSE | ( dev->current_hps_sync == SAA7146_HPS_SYNC_PORT_A ? MASK_11 : MASK_13);	/* wait for e_fid_a/b */

			/* turn off video-dma1 */
			dev->rps0[ count++ ] = CMD_WR_REG_MASK | (MC1/4);
			dev->rps0[ count++ ] = MASK_22 | MASK_06;	    		/* => mask */
			dev->rps0[ count++ ] = MASK_22;					/* => values */

			/* Write MC2: unset rps-signal */
			dev->rps0[ count++ ] = CMD_WR_REG | (1 << 8) | (MC2/4); 	
			dev->rps0[ count++ ] = (1 << (27+i));

			/* generate interrupt */
			dev->rps0[ count++ ] = CMD_INTERRUPT;				

			/* jump to the "beginning" */
			dev->rps0[ count++ ] = CMD_JUMP;
			dev->rps0[ count++ ] = virt_to_bus(&dev->rps0[begin]);
		}
	}

	return 0;
}

/* this function gets called with interrupts disabled, so we can
   safely fiddle with the queues and stuff here */
   
void try_to_queue_capture_buffers(struct capture_device *dev)
{
	u32 mc2 = 0;
	
	/* check if we can queue one or two buffers */
	mc2 = saa7146_read(dev, MC2);

	DEB_CAP(("saa7146_v4l2.o: try_to_queue_capture_buffers called.\n"));

	if( 0 == (mc2 & MASK_11)) {
		/* try to queue buffer on double buffer #1 */
		struct stream_buffer* newbuf = v4l2_q_del_head(&dev->stream_q_capture);
		if( 0 == newbuf ) {
			/* todo queue is empty, we can stop here */
			DEB_CAP(("saa7146_v4l2.o: try_to_queue_capture_buffers: stream_q_capture is empty.\n"));
			goto exit;
		}
		DEB_CAP(("saa7146_v4l2.o: try_to_queue_capture_buffers: queueing into buffer 0 (%p).\n",newbuf));
		
		dev->stream_db_cur[0] = newbuf;
		/* fixme: don't erase the swap bit */				
		*(dev->stream_db_adr[0]) = (virt_to_bus(newbuf->page_table) | xxxflags);
		
		/* set bit in MC2, so the RPS starts capturing */
		saa7146_write(dev, MC2, MASK_11|MASK_27);
	}
		
	if( 0 == (mc2 & MASK_12)) {
		/* try to queue buffer on double buffer #2 */
		struct stream_buffer* newbuf = v4l2_q_del_head(&dev->stream_q_capture);
		if( 0 == newbuf ) {
			/* todo queue is empty, we can stop here */
			DEB_CAP(("saa7146_v4l2.o: try_to_queue_capture_buffers. stream_q_capture is empty.\n"));
			goto exit;
		}
		DEB_CAP(("saa7146_v4l2.o: try_to_queue_capture_buffers: queueing into buffer 1.\n"));

		dev->stream_db_cur[1] = newbuf;

		/* fixme: don't erase the swap bit */				
		*(dev->stream_db_adr[1]) = (virt_to_bus(newbuf->page_table) | xxxflags);
	
		/* set bit in MC2, so the RPS starts capturing */
		saa7146_write(dev, MC2, MASK_12|MASK_28);
	}

exit:	
	return;			
}

int capture_queuebuffer(struct capture_device *dev, struct stream_buffer *buf)
{
	u32 flags = 0;

	DEB_CAP(("saa7146_v4l2.o: capture_queuebuffer called: %p.\n",buf));

	if ((buf->vidbuf.flags & V4L2_BUF_FLAG_QUEUED)) {
		DEB_S(("saa7146_v4l2.o: capture_queuebuffer: buffer %p is already queued\n",buf));
		return -EINVAL;
	}

	save_flags(flags);
	cli();

	buf->vidbuf.flags &= ~V4L2_BUF_FLAG_DONE;
	buf->vidbuf.flags |= V4L2_BUF_FLAG_QUEUED;
	v4l2_q_add_tail(&dev->stream_q_capture, &buf->qnode);
	
	/* note: for revision 0 capturing nothing has to be do here, since
 	   buffer handling is completely done within the irq-handler */
	if( 0 != dev->revision ) {
		try_to_queue_capture_buffers(dev);
	}
	
	restore_flags(flags);

	return 0;
}

struct stream_buffer* capture_dequeuebuffer(struct capture_device *dev)
{
	struct stream_buffer *sbuf = v4l2_q_del_head(&dev->stream_q_done);

	if(sbuf == NULL) {
		DEB_S(("saa7146_v4l2.o: capture_dequeuebuffer: nothing on done queue.\n"));
		return NULL;
	}

	sbuf->vidbuf.flags &= ~V4L2_BUF_FLAG_QUEUED;
	DEB_CAP(("saa7146_v4l2.o: capture_dequeuebuffer: dequeued %p @ %li\n",sbuf,jiffies));

	return sbuf;
}

int capture_streamoff(struct capture_device *dev, u32 type)
{
	struct stream_buffer *buf;

	DEB_EE(("saa7146_v4l2.o: capture_streamoff called.\n"));

	if (type != V4L2_BUF_TYPE_CAPTURE) {
		DEB_S(("saa7146_v4l2.o: capture_streamoff: wrong buffer type.\n"));
		return -EINVAL;
	}
	
	capture_abort(dev);

	DEB_EE(("saa7146_v4l2.o: flushing queues...\n"));
	
	/*  Move any leftover DONE buffers to the free pool */
	while ((buf = v4l2_q_del_head(&dev->stream_q_done))) {
		buf->vidbuf.flags &= ~V4L2_BUF_FLAG_QUEUED;
		DEB_EE(("saa7146_v4l2.o: flushed %p from done queue\n",buf));
	}

	/*  Move any leftover TODO buffers to the free pool */
	while ((buf = v4l2_q_del_head(&dev->stream_q_capture))) {
		buf->vidbuf.flags &= ~V4L2_BUF_FLAG_QUEUED;
		DEB_EE(("saa7146_v4l2.o: flushed %p from todo queue\n",buf));
	}
	
	buf = dev->stream_db_cur[0];
	if( NULL != buf ) {
		DEB_EE(("saa7146_v4l2.o: flushed %p from db #0\n",buf));
		buf->vidbuf.flags &= ~V4L2_BUF_FLAG_QUEUED;
		dev->stream_db_cur[0] = NULL;
	}
	buf = dev->stream_db_cur[1];
	if( NULL != buf ) {
		DEB_EE(("saa7146_v4l2.o: flushed %p from db #1\n",buf));
		buf->vidbuf.flags &= ~V4L2_BUF_FLAG_QUEUED;
		dev->stream_db_cur[1] = NULL;
	}
		
	DEB_EE(("saa7146_v4l2.o: capture_streamoff() done.\n"));

	dev->streaming = 0;
	
	return 0;
}

void write_out_dma(struct saa7146* dev, int which, struct saa7146_video_dma* vdma) 
{
	int where = 0;
	
	if( which < 1 || which > 3) {
		return;
	}
	
	/* calculate starting address */
	where  = (which-1)*0x18;

	saa7146_write(dev, where, 	vdma->base_odd);
	saa7146_write(dev, where+0x04, 	vdma->base_even);
	saa7146_write(dev, where+0x08, 	vdma->prot_addr);
	saa7146_write(dev, where+0x0c, 	vdma->pitch);
	saa7146_write(dev, where+0x10, 	vdma->base_page);
	saa7146_write(dev, where+0x14,	vdma->num_line_byte);
}

int setup_grabbing(struct capture_device *dev)
{
	struct	saa7146_video_dma vdma1;

	DEB_EE(("saa7146_v4l2.o: setup_grabbing() called.\n"));

	saa7146_set_overlay(dev,0);
	saa7146_set_disable_clipping(dev);
	saa7146_set_window(dev, no_interlace, dev->clientfmt.width, dev->clientfmt.height, dev->hflip, dev->current_hps_source, dev->current_hps_sync);
	saa7146_set_output_format(dev, dev->stream_format);
/*
	saa7146_set_picture_prop(dev, 0x80, 0x40, 0x40);
*/
	calculate_video_dma_grab_packed(dev, dev->clientfmt.width, dev->clientfmt.height, dev->clientfmt.depth, no_interlace, dev->vflip, 0x0, &vdma1);
	/* note: BASE_PAGE1 gets set in irq-handler */
	write_out_dma(dev, 1, &vdma1);

/*
	printk("w:%d, h:%d, d:%d\n", dev->clientfmt.width,dev->clientfmt.height,dev->clientfmt.depth);

	printk("hps_v_scale:         0x%08x\n", saa7146_read(dev, HPS_V_SCALE));
	printk("hps_v_gain:          0x%08x\n", saa7146_read(dev, HPS_V_GAIN));
	printk("hps_ctrl:            0x%08x\n", saa7146_read(dev, HPS_CTRL));
	printk("hps_h_prescale:      0x%08x\n", saa7146_read(dev, HPS_H_PRESCALE));
	printk("hps_h_scale:         0x%08x\n", saa7146_read(dev, HPS_H_SCALE));
	printk("clip_format:         0x%08x\n", saa7146_read(dev, CLIP_FORMAT_CTRL));

	printk("vdma1.base_odd:      0x%08x\n", vdma1.base_odd);
	printk("vdma1.base_even:     0x%08x\n", vdma1.base_even);
	printk("vdma1.prot_addr:     0x%08x\n", vdma1.prot_addr);
	printk("vdma1.pitch:         0x%08x\n", vdma1.pitch);
	printk("vdma1.base_page:     0x%08x\n", vdma1.base_page);
	printk("vdma1.num_line_byte: 0x%08x\n", vdma1.num_line_byte);
*/

	if( 0 != dev->revision ) {

		/* enable rps0 irqs */
		dev->irq_mask |= MASK_27;	
		saa7146_write(dev, IER, dev->irq_mask);

		/* write the address of the rps-program */
		saa7146_write(dev, RPS_ADDR0, virt_to_bus(&dev->rps0[ 0]));
		/* turn on rps */
		saa7146_write(dev, MC1, (MASK_12 | MASK_28));	
	} else {		
		u32 dd1_init = saa7146_read(dev, DD1_INIT);
		
		/* set interrupt condition according to selected port, enable irq */
		if( 0 == dev->current_hps_sync ) {
			dev->irq_mask |= MASK_08;
			dd1_init |= MASK_22;
		} else {
			dev->irq_mask |= MASK_07;
			dd1_init |= MASK_06;
		}
		saa7146_write(dev, DD1_INIT, dd1_init);	
		saa7146_write(dev, MC2, (MASK_09 | MASK_25 | MASK_10 | MASK_26));

		saa7146_write(dev, IER, dev->irq_mask);
	}
	return 0;
}

int capture_streamon(struct capture_device *dev, u32 type)
{
	DEB_EE(("saa7146_v4l2.o: capture_streamon called.\n"));

	if (type != V4L2_BUF_TYPE_CAPTURE) {
		DEB_S(("saa7146_v4l2.o: capture_streamon: wrong buffer type.\n"));
		return -EINVAL;
	}

	if (0 != dev->streaming) {
		DEB_S(("saa7146_v4l2.o: capture_streamon: already streaming.\n"));
		/* fixme: anything to do here ? */
		return 0;
	}
	
	dev->streaming = 1;

	/*  -2 is a magic number that triggers start-of-stream logic in */
	/*    capture_interrupt()  */
	dev->stream_last_frame = -2;

	/* clear out performace statistics */
	memset(&dev->perf,0x0,sizeof(struct v4l2_performance));

	/* adjust video-dma1 according to requested programming */
	if( 0 != setup_grabbing(dev)) {
		return -1;
	}

	/* invalidate variable, because we have just invalidated the data */
	dev->overlay_data_valid = 0;
	
	return 0;
}

int capture_new_format(struct capture_device* dev) 
{
	int max_image_size = 0;
	int max_height = 0;
	int max_width = 0;
	int max_pixels = 0;
	int t = 0;
		
	DEB_EE(("saa7146_v4l2.o: capture_new_format called: format \"%4.4s\", w:%d, h:%d\n", (char*)&dev->clientfmt.pixelformat, dev->clientfmt.width, dev->clientfmt.height));

	/* sanity check */
	if( dev->clientfmt.width == 0 || dev->clientfmt.height == 0) {
		printk("saa7146_v4l2.o: capture_new_format(): illegal value for w (%d) or h (%d)\n",dev->clientfmt.width,dev->clientfmt.height);
		return -1;
	}
	
	/* fixme: what is this good for anyway? */
	dev->clientfmt.flags = 0;

	dev->stream_flags = ME1;
	switch (dev->clientfmt.pixelformat)
	{
	case V4L2_PIX_FMT_RGB332:
		dev->stream_format = RGB08_COMPOSED;
		dev->clientfmt.depth = 8;
		break;
	case V4L2_PIX_FMT_RGB555:
	case V4L2_PIX_FMT_RGB565:
		dev->stream_format = RGB16_COMPOSED;
		dev->clientfmt.depth = 16;
		break;
	case V4L2_PIX_FMT_BGR24:
		dev->stream_format = RGB24_COMPOSED;
		dev->clientfmt.depth = 24;
		break;
	case V4L2_PIX_FMT_BGR32:
		dev->stream_format = RGB32_COMPOSED;
		dev->clientfmt.depth = 32;
		break;
	case V4L2_PIX_FMT_GREY:
		dev->stream_format = Y8;
		dev->clientfmt.depth = 8;
		break;
	case V4L2_PIX_FMT_YUYV:
		dev->stream_format = YUV422_COMPOSED;
		dev->stream_flags |= SAA7146_TWO_BYTE_SWAP;
		dev->clientfmt.depth = 16;
		break;
	case V4L2_PIX_FMT_UYVY:
		dev->stream_format = YUV422_COMPOSED;
		dev->clientfmt.depth = 16;
		break;
	default:
		printk("saa7146_v4l2.o: unknown format %4.4s\n", (char *)&dev->clientfmt.pixelformat);
		return -1;
	}

#define MAX_IMAGE_SIZE 0x1b0000
	max_image_size = MAX_IMAGE_SIZE;

	/* first see, if buffers are already mapped. if so, we should not
	   exceed the mapped memory size */
	if ( 0 != dev->stream_buffers_mapped ) {
		DEB_S(("saa7146_v4l2.o: capture_new_format: warning: buffers already mapped. you might not get what you want!\n"));
		max_image_size = dev->stream_buf[0].vidbuf.length;
	}
		
//	printk("format %4.4s\n", (char *)&dev->clientfmt.pixelformat);

	max_width = dev->capability.maxwidth; 
	max_height = dev->capability.maxheight;
	max_pixels = max_image_size / (dev->clientfmt.depth / 8);

	t = isqrt((max_pixels * dev->clientfmt.width) / dev->clientfmt.height);
	if (t < max_width)
		max_width = t;
	t = isqrt((max_pixels * dev->clientfmt.height) / dev->clientfmt.width);
	if (t < max_height)
		max_height = t;

	if (dev->clientfmt.width > max_width)
		dev->clientfmt.width = max_width;
	if (dev->clientfmt.height > max_height)
		dev->clientfmt.height = max_height;

	if (dev->clientfmt.width < dev->capability.minwidth)
		dev->clientfmt.width = dev->capability.minwidth;
	if (dev->clientfmt.height < dev->capability.minheight)
		dev->clientfmt.height = dev->capability.minheight;
	if (dev->clientfmt.width > dev->capability.maxwidth)
		dev->clientfmt.width = dev->capability.maxwidth;
	if (dev->clientfmt.height > dev->capability.maxheight)
		dev->clientfmt.height = dev->capability.maxheight;

	dev->clientfmt.sizeimage = (dev->clientfmt.width
		* dev->clientfmt.height
		* dev->clientfmt.depth)
		/ 8;

	DEB_S(("saa7146_v4l2.o: capture_new_format: granted: w:%d, h:%d, size:%d\n",dev->clientfmt.width,dev->clientfmt.height,dev->clientfmt.sizeimage));
	
	return 0;
}

/*********************************************************************************/
/* saa7146-stuff that is not that generic                                        */

int saa7146_retrieve_overlay_data(struct capture_device* dev, struct v4l2_window *w)
{		
	memset(w,0x0,sizeof(struct v4l2_window));

	if( 0 != dev->overlay_data_available) {
		w->x = dev->overlay_x;
		w->y = dev->overlay_y;
		w->width = dev->overlay_width;
		w->height = dev->overlay_height;
		return 0;
	}

	return -1;
}

int saa7146_store_overlay_data(struct capture_device* dev, struct v4l2_window *w)
{		
	/* check if we are exceeding the width & height of current norm */
	if( dev->modes_constants.v_max_out < w->height || dev->modes_constants.h_max_out < w->width) {
		return -1;
	}

	if( 0 != dev->clip_data ) {
		kfree(dev->clip_data);
		dev->clip_data = NULL;
	}
	dev->clip_count = 0;
		
	/* look for clipping informations */
	if( NULL != w->clips && 0 != w->clipcount ) {
			
		int 	i = 0;
		int	*x = NULL, *y = NULL, *wi = NULL, *he = NULL;
		struct	v4l2_clip* vc = NULL;
			
		DEB_EE(("saa7146_v4l2.o: saa7146_store_overlay_data, doing rectangle clipping.\n"));

		/* prepare clipping rectangles */
		x = (int*)kmalloc(sizeof(int)*w->clipcount*4,GFP_KERNEL);
		if( NULL == x ) {
			DEB_S(("saa7146_v4l2.o: saa7146_store_overlay_data error, out of kernel-memory.\n"));
			return -ENOMEM;
		}
		y = x+(1*w->clipcount);
		wi = x+(2*w->clipcount);
		he = x+(3*w->clipcount);

		/* transform clipping-windows, copy clips to kernel-space */
		vc = vmalloc(sizeof(struct v4l2_clip)*(w->clipcount));
		if( NULL == vc	) {
			DEB_S(("saa7146_v4l2.o: saa7146_store_overlay_data error, no memory #2.\n"));
			return -ENOMEM;
		}
		if(copy_from_user(vc,w->clips,sizeof(struct v4l2_clip)*w->clipcount)) {
			DEB_S(("saa7146_v4l2.o: saa7146_store_overlay_data error, could not copy from user-space.\n"));
			return -EFAULT;
		}

		/* copy the clip-list to the arrays
		   note: the video_clip-struct may contain negative values to indicate that a window
			 doesn't lay completly over the video window. Thus, we correct the values right here */
		for(i = 0; i < w->clipcount; i++) {

//			printk("vor  %d: %d,%d,%d,%d\n", i,vc[i].width, vc[i].height, vc[i].x, vc[i].y);

			if( vc[i].width < 0) {
				vc[i].x += vc[i].width; vc[i].width = -vc[i].width;
			}
			if( vc[i].height < 0) {
				vc[i].y += vc[i].height; vc[i].height = -vc[i].height;
			}
			if( vc[i].x < 0) {
				vc[i].width += vc[i].x; vc[i].x = 0;
			}
			if( vc[i].y < 0) {
				vc[i].height += vc[i].y; vc[i].y = 0;
			}	

//			printk("nach %d: %d,%d,%d,%d\n", i,vc[i].width, vc[i].height, vc[i].x, vc[i].y);

			if(vc[i].width <= 0 || vc[i].height <= 0) {
				vfree(vc);
				DEB_S(("saa7146_v4l2.o: saa7146_store_overlay_data error, width and height out of range.\n"));
				return -EINVAL;
			}
			x[i]	= vc[i].x;
			y[i]	= vc[i].y;
			wi[i]	= vc[i].width;
			he[i]	= vc[i].height;
		}		

		/* free memory used for temporary clips */
		vfree(vc);
		
		dev->clip_data = x;
		dev->clip_count = w->clipcount;
			
	} else if ( NULL != w->bitmap ) {
		DEB_EE(("saa7146_v4l2.o: saa7146_store_overlay_data, doing mask clipping.\n"));
	} else {
		DEB_EE(("saa7146_v4l2.o: saa7146_store_overlay_data, no clipping informations found.\n"));
	}

	dev->overlay_width = w->width;
	dev->overlay_height = w->height;
	dev->overlay_x = w->x;
	dev->overlay_y = w->y;
	
	dev->overlay_data_available = 1;

	return 0;
}

int saa7146_use_overlay_data(struct capture_device* dev) 
{	
	DEB_EE(("saa7146_v4l2.o: saa7146_use_overlay_data.\n"));

	/* check if we have informations at all */
	if( 0 == dev->overlay_data_available ) {
		DEB_S(("saa7146_v4l2.o: saa7146_use_overlay_data: ouch, no overlay data available.\n"));
        	return -1;
	}

	/* check if the user has called S_FBUF before and
	   supplied the color depth of the display... */

	if( -1 == dev->overlay_palette ) {
		printk("saa7146_v4l2.o: saa7146_use_overlay_data: no information about color depth available. forgot to call S_FBUF (%d)?\n",dev->overlay_palette);
		return -1;
	}

	if( dev->overlay_width > dev->capability.maxwidth ) {
		dev->overlay_width = dev->capability.maxwidth;
	}
	if( dev->overlay_height > dev->capability.maxheight ) {
		dev->overlay_height = dev->capability.maxheight;
	}


	/* while fiddling with the saa7146, disable overlay video */
	if( 0 != dev->overlay ) {
		saa7146_set_overlay(dev, 0);
	}

	if( 0 == dev->clip_count ) {
		saa7146_set_disable_clipping(dev);
	} else {
		saa7146_set_clipping_rect(dev, no_interlace, dev->overlay_width, dev->overlay_height, dev->clipping, dev->clip_count, &dev->clip_data[0], &dev->clip_data[0+(1*dev->clip_count)], &dev->clip_data[0+(2*dev->clip_count)], &dev->clip_data[0+(3*dev->clip_count)], 0, dev->vflip);
	}

	saa7146_set_window(dev, no_interlace, dev->overlay_width, dev->overlay_height, dev->hflip, dev->current_hps_source, dev->current_hps_sync);
	saa7146_set_position(dev, no_interlace, dev->overlay_x, dev->overlay_y, dev->overlay_height, dev->fb.fmt.width, dev->fb.fmt.depth, dev->fb.fmt.bytesperline, (u32)dev->fb.base[0], dev->vflip);
	saa7146_set_output_format(dev, dev->overlay_palette);

	if( 0 != dev->overlay ) {
		saa7146_set_overlay(dev, 1);
	}

	dev->overlay_data_valid = 1;
        return 0;
}

int preview(struct capture_device *dev, int i)
{
	if ( 0 != dev->streaming ) {
		DEB_EE(("saa7146_v4l2.o: VIDIOC_PREVIEW: illegal while streaming.\n"));
		return -EPERM;
	}
	
        /* enable overlay */	
	if( 0 != i ) {
		/* check if overlay_data is valid, if not, gracefully recall informations first */
		if( 0 == dev->overlay_data_valid) {
			if( 0 != saa7146_use_overlay_data(dev)) {
               	       	 	printk("saa7146_v4l2.o: VIDIOC_PREVIEW: cannot enable overlay. call S_WIN first!\n");
                	               return -EINVAL;
			}
		}
		dev->overlay = 1;
		saa7146_set_overlay(dev, 1);
	} else {
		dev->overlay = 0;
		saa7146_set_overlay(dev, 0);
        }
	
	return 0;
}


/*********************************************************************************/
/* mmap-functions 1/2                                                              */

struct capture_device* capture_device_from_file(struct file *file)
{
	return (struct capture_device *)v4l2_device_from_file(file);
}

int mmap_unrequest_buffers(struct capture_device *dev)
{
	int	i = 0;

	DEB_EE(("saa7146_v4l2.o: mmap_unrequest_buffers() called.\n"));

	if (dev->stream_buffers_requested == 0 ) {
		DEB_EE(("saa7146_v4l2.o: mmap_unrequest_buffers(): no buffers previously requested.\n"));
		return -EINVAL;
	}

	if (dev->stream_buffers_mapped != 0 ) {
		printk("saa7146_v4l2.o: mmap_unrequest_buffers(): error: buffers are mapped!\n");
		return -EINVAL;
	}

	for (i = 0; i < MAX_SAA7146_CAPTURE_BUFFERS; ++i) {
		dev->stream_buf[i].requested = 0;		
	}
	
	dev->stream_buffers_requested = 0;

	return 0;
}


int mmap_request_buffers(struct capture_device *dev, struct v4l2_requestbuffers *req)
{
	int	i;
	u32	buflen;
	u32	type;

	DEB_EE(("saa7146_v4l2.o: mmap_request_buffers called.\n"));

	/* can't make requests if buffers are mapped */
	if( 0 != dev->stream_buffers_mapped ) {
		printk("saa7146_v4l2.o: mmap_request_buffers(): error: buffers are already mapped!\n");
		return -EINVAL;
	}
	
	/* check maximum / minimun number of buffers */
	if (req->count < 1)
		req->count = 1;
	if (req->count > MAX_SAA7146_CAPTURE_BUFFERS) {
		DEB_D(("saa7146_v4l2.o: maximum number of capture buffers is %d (requested: %d).\n", MAX_SAA7146_CAPTURE_BUFFERS, req->count));
		req->count = MAX_SAA7146_CAPTURE_BUFFERS;
	}
	
	type = V4L2_BUF_TYPE_CAPTURE;

	dev->stream_contig_map = 0;
	/* check if we neep continuous buffers for v4l backward compatibility */
	if (req->type & V4L2_BUF_REQ_CONTIG) {
		dev->stream_contig_map = 1;
		req->type = type | V4L2_BUF_REQ_CONTIG;
		/* note: _REQ_CONTIG is only used in v4l2_requestbuffers */
	}

	/*  The buffer length needs to be a multiple of the page size  */
	buflen = (dev->clientfmt.sizeimage + PAGE_SIZE - 1)
		& ~(PAGE_SIZE - 1);

	/* check if we exceed the maximum memory usage */
	if( (buflen*req->count) > (memory * 1048576) ) {
		DEB_D(("saa7146_v4l2.o: maximum memory usage is %dMb, you requested %d buffers a %d bytes.\n",memory,req->count,buflen));
		req->count = (memory * 1048576) / buflen;
		DEB_D(("saa7146_v4l2.o:  limiting number of buffers to %d.\n",req->count));
	}
	
	/*  Now initialize the buffer structures. Don't allocate the */
	/*  buffers until they're mapped. */
	for (i = 0; i < req->count; ++i)
	{
		dev->stream_buf[i].requested = 1;
		dev->stream_buf[i].vidbuf.index = i;
		dev->stream_buf[i].vidbuf.type = type;
		dev->stream_buf[i].vidbuf.offset = i<<PAGE_SHIFT;/* anything unique */
		dev->stream_buf[i].vidbuf.length = buflen;
		dev->stream_buf[i].vidbuf.bytesused = 0;
		dev->stream_buf[i].vidbuf.flags = 0;
		dev->stream_buf[i].vidbuf.timestamp = 0;
		dev->stream_buf[i].vidbuf.sequence = 0;
		memset(&dev->stream_buf[i].vidbuf.timecode, 0, sizeof(struct v4l2_timecode));
	}
	
	for (i = req->count; i < MAX_SAA7146_CAPTURE_BUFFERS; ++i)
		dev->stream_buf[i].requested = 0;

	dev->stream_buffers_requested = req->count;

	DEB_D(("saa7146_v4l2.o: mmap_request_buffers: granting %d buffers.\n",req->count));

	return 0;
}

/*********************************************************************************/
/* mmap-functions 2/2                                                            */

struct stream_buffer *mmap_stream_buffer_from_offset(struct capture_device *dev, struct vm_area_struct *vma)
{
	int	i;
		
/*
	DEB_EE(("saa7146_v4l2.o: mmap_stream_buffer_from_offset called.\n"));
*/

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
	i = vma->vm_offset >> PAGE_SHIFT;;
#else
	i = vma->vm_pgoff;
#endif
	if( i < 0 || i > dev->stream_buffers_requested)
		return NULL;

	return &dev->stream_buf[i];
}

void mmap_vma_open(struct vm_area_struct *vma)
{
	struct capture_device *dev = capture_device_from_file(vma->vm_file);

	DEB_EE(("saa7146_v4l2.o: mmap_vma_open called.\n"));

	if (dev == NULL)
		return;
}

void mmap_vma_close(struct vm_area_struct *vma)
{
	struct capture_device *dev = capture_device_from_file(vma->vm_file);
	struct stream_buffer *buf = mmap_stream_buffer_from_offset(dev, vma);
	
	DEB_EE(("saa7146_v4l2.o: mmap_vma_close called.\n"));

	if( 0 != dev->streaming) {
		DEB_D(("saa7146_v4l2.o: Warning - munmap() called while streaming.\n"));
		capture_streamoff(dev, buf->vidbuf.type);
	}

	if( 0 != dev->stream_contig_map) {
		/* v4l compatiblity */
		unallocate_v4l_buffers(dev);
	} else {
		unallocate_single_buffer(dev,buf);
		buf->vidbuf.flags &= ~V4L2_BUF_FLAG_MAPPED;
	}
	
	dev->stream_buffers_mapped--;
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
static unsigned long mmap_vma_nopage(struct vm_area_struct *vma, unsigned long address, int write)
#else
struct page *mmap_vma_nopage(struct vm_area_struct *vma, unsigned long address, int write)
#endif
{
	struct capture_device	*dev;
	struct stream_buffer	*buf;
	unsigned long		offset_into_buffer;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
	unsigned long page;
#else
	struct page *page;
#endif

	int n = 1;

	dev = capture_device_from_file(vma->vm_file);
	if(dev == NULL) {
	 	return 0;
	}
	
	offset_into_buffer = address - vma->vm_start;

	if(0 != dev->stream_contig_map) {
		buf = &dev->stream_buf[0];
		n = dev->stream_buffers_requested;
	} else {
		buf = mmap_stream_buffer_from_offset(dev, vma);
		if (buf == NULL) {
			printk("saa7146_v4l2.o: mmap_vma_nopage: fatal: NULL pointer buffer argument.\n.");
			return 0;
		}
	}
			
	if (offset_into_buffer >= buf->vidbuf.length*n ) {
		DEB_S(("saa7146_v4l2.o: mmap_vma_nopage: Attempt to read past end of mmap() buffer\n."));
		return 0;
	}
	page = v4l2_vmalloc_to_page(buf->vaddress + offset_into_buffer);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
	if (page == 0) {
		DEB_S(("saa7146_v4l2.o: mmap_vma_nopage - page == NULL"));
		return 0;
	}
	atomic_inc(&mem_map[MAP_NR(page)].count);
#else
	if (page == NULL) {
		DEB_S(("saa7146_v4l2.o: mmap_vma_nopage - page == NULL"));
		return 0;
	}
	atomic_inc(&page->count);
#endif

	return page;
}

static struct vm_operations_struct capture_vma_operations =
{
	open:	mmap_vma_open,
	close:	mmap_vma_close,
	nopage:	mmap_vma_nopage
};

/*********************************************************************************/
/* irq-handler functions                                                         */
   
void interrupt_bh(void* v)
{
	struct capture_device *dev = (struct capture_device*)v;

	u32 isr = dev->irq_left;
	u32 isr_ext = dev->irq_left & dev->ext->irq_mask;
	dev->irq_left = 0;
			
	DEB_INT(("%s: bh-irq-call @ %li: isr:0x%08x\n",dev->name,jiffies,isr));

	/* rps1 interrupt ? ==> revision != 0 vbi capture */
	if ( 0 != (isr & MASK_28)) {
        
		DEB_INT(("%s: rps1 vbi interrupt, jiffies:%li.\n",dev->name,jiffies));	

		isr &= ~(MASK_28);

		/* write counter to last word memory */
		dev->vbi_mem[dev->vbi_current][((dev->vbi_pixel_to_capture*dev->vbi_depth*32)-4)/4] = dev->vbi_counter;
		dev->vbi_counter++;
		
                if( dev->vbi_counter == (dev->vbi_last_read-2)) {
			DEB_INT(("%s: vbi-frame lost (%li/%li).\n",dev->name,dev->vbi_counter,dev->vbi_last_read));	
        	}
		dev->vbi_current = !dev->vbi_current;	

		/* notify any pending process */
		wake_up(&dev->vbi_wq);
	}

	/* rps0 interrupt ? ==> revision != 0 capturing */
	if ( 0 != (isr & MASK_27)) {

		struct stream_buffer *buf;
		stamp_t	timestamp_rough;
		int captured = 0;

		u32 mc2 = 0;
				
		DEB_CAP(("%s: bh: rps0-interrupt, stream/read capture.\n",dev->name));

		isr &= ~MASK_27;
	
		/* read out rps-signals */
		mc2 = saa7146_read(dev, MC2);

		/* first frame of the stream */
		if(dev->stream_last_frame == -2 ) {
			v4l2_masterclock_gettime(&dev->stream_begin);
//			v4l2_timestamp_correct(&dev->stream_begin, 100); // dev->capture.timeperframe);
			dev->stream_last_frame = -1;
			
		}

		/*  Compute current stream time  */
		v4l2_masterclock_gettime(&timestamp_rough);
		/*  We do computations on a relative timestamp (since beginning  */
		/*  of stream), then add 'dev->stream_begin' back in later       */
		timestamp_rough -= dev->stream_begin;

		/* check if we captured to buffer 0 */
		buf = dev->stream_db_cur[0];
		if( NULL != buf && (0 == (mc2 & MASK_11))) {
			DEB_CAP(("%s: bh: grab to buffer 0\n",dev->name));

			buf->vidbuf.flags |= V4L2_BUF_FLAG_DONE;
			buf->vidbuf.timestamp = timestamp_rough;
//			v4l2_timestamp_correct(&buf->vidbuf.timestamp, 100); // dev->capture.timeperframe);

			buf->vidbuf.timestamp += dev->stream_begin;
			v4l2_q_add_tail(&dev->stream_q_done, &buf->qnode);

			captured++;

			dev->perf.frames++;
			dev->perf.bytesout += buf->vidbuf.length;
			
			dev->stream_db_cur[0] = NULL;		
		}
		
		buf = dev->stream_db_cur[1];
		/* check if we captured to buffer 1 */
		if( NULL != buf && (0 == (mc2 & MASK_12))) {
			DEB_CAP(("%s: bh: grab to buffer 1\n",dev->name));

			buf->vidbuf.flags |= V4L2_BUF_FLAG_DONE;
			buf->vidbuf.timestamp = timestamp_rough;
//			v4l2_timestamp_correct(&buf->vidbuf.timestamp, 100); // dev->capture.timeperframe);

			buf->vidbuf.timestamp += dev->stream_begin;
			v4l2_q_add_tail(&dev->stream_q_done, &buf->qnode);

			captured++;

			dev->perf.frames++;
			dev->perf.bytesout += buf->vidbuf.length;

			dev->stream_db_cur[1] = NULL;		
		}

		try_to_queue_capture_buffers(dev);
		dev->stream_last_frame++;
			
		/* check if we missed a frame */
		if( 0 == captured ) {
			DEB_CAP(("%s: frame missed (total: %d).\n",dev->name,dev->perf.framesdropped));
			dev->perf.framesdropped++;
		} else {
			/* notify any pending process */
			wake_up(&dev->capture_wq);
		}
	}

	/* FIDA/FIDB interrupt && revision == 0 => revision 0 capturing */
	if ( dev->revision == 0 && (0 != (isr & (MASK_07|MASK_08)))) {

		struct stream_buffer *buf = NULL;
		stamp_t	timestamp_rough;

		isr &= ~(MASK_07|MASK_08);

		DEB_CAP(("%s: bh: FIDA/FIDB-interrupt.\n",dev->name));
		
		/* first frame of the stream  */
		if(dev->stream_last_frame == -2) {

			v4l2_masterclock_gettime(&dev->stream_begin);
//			v4l2_timestamp_correct(&dev->stream_begin, 100); // dev->capture.timeperframe);
			dev->stream_last_frame = -1;

			/* look if any buffers are queued */
			buf = v4l2_q_del_head(&dev->stream_q_capture);
			if( 0 == buf ) {
				/* nope, nothing to be done here */
				dev->stream_db_cur[0] = NULL;
				DEB_CAP(("%s: bh: no buffer queued for first grab.\n",dev->name));
			} else {
				dev->stream_db_cur[0] = buf;
				dev->stream_rev0_data = saa7146_read(dev, PROT_ADDR1);

				/* set new page table + upload */		
				saa7146_write(dev, BASE_PAGE1, virt_to_bus(buf->page_table) | xxxflags);
				saa7146_write(dev, MC2, MASK_02|MASK_18);		

				/* start video-dma1 */
				saa7146_write(dev, MC1, MASK_06|MASK_22);		

				DEB_CAP(("%s: bh: setting up first capture.\n",dev->name));
			}
/*
			printk("MC1: 0x%08x\n",saa7146_read(dev,MC1));
			printk("MC2: 0x%08x\n",saa7146_read(dev,MC2));
			printk("BASE_ODD1:      0x%08x\n", saa7146_read(dev, BASE_ODD1));
			printk("BASE_EVEN1:     0x%08x\n", saa7146_read(dev, BASE_EVEN1));
			printk("PROT_ADDR1:     0x%08x\n", saa7146_read(dev, PROT_ADDR1));
			printk("PITCH1:         0x%08x\n", saa7146_read(dev, PITCH1));
			printk("BASE_PAGE1:     0x%08x\n", saa7146_read(dev, BASE_PAGE1));
			printk("NUM_LINE_BYTE1: 0x%08x\n", saa7146_read(dev, NUM_LINE_BYTE1));
*/
		} else {

			/* did we capture to a buffer? */
			if( NULL == dev->stream_db_cur[0] ) {
				/* we did not set up a capture, so this frame is lost */
				DEB_CAP(("%s: frame missed (total: %d).\n",dev->name,dev->perf.framesdropped));
				dev->perf.framesdropped++;
			} else {
				DEB_CAP(("%s: frame captured.\n",dev->name));
				buf = dev->stream_db_cur[0];
				/* ok, buf now points to the buffer we have just captured to */
							
				buf->vidbuf.flags |= V4L2_BUF_FLAG_DONE;

				/*  Compute current stream time  */
				v4l2_masterclock_gettime(&timestamp_rough);
				/*  We do computations on a relative timestamp (since beginning  */
				/*  of stream), then add 'dev->stream_begin' back in later       */

				timestamp_rough -= dev->stream_begin;	
				buf->vidbuf.timestamp = timestamp_rough;
//				v4l2_timestamp_correct(&buf->vidbuf.timestamp, 100); // dev->capture.timeperframe);
				buf->vidbuf.timestamp += dev->stream_begin;

				dev->stream_last_frame++;

				/* performance statistics */
				dev->perf.frames++;
				dev->perf.bytesout += buf->vidbuf.length;

				/* add this buffer to our done queue */
				v4l2_q_add_tail(&dev->stream_q_done, &buf->qnode);

				/* notify any pending process */
				wake_up(&dev->capture_wq);
			}
		
			/* reprogram the capture engine, if there is another buffer we can capture to */
			buf = v4l2_q_del_head(&dev->stream_q_capture);
			if( 0 == buf ) {
				/* we cannot set up a new capture, so stop video-dma1 */
				dev->stream_db_cur[0] = NULL;

				saa7146_write(dev, PROT_ADDR1, 	0x0);
				saa7146_write(dev, MC2, MASK_02|MASK_18);		
/*
				saa7146_write(dev, MC1, MASK_22);		
*/
				DEB_CAP(("%s: capture not reprogrammed!.\n",dev->name));
			} else {
				dev->stream_db_cur[0] = buf;

				/* set new page table */		
				saa7146_write(dev, PROT_ADDR1, dev->stream_rev0_data);
				saa7146_write(dev, BASE_PAGE1, virt_to_bus(buf->page_table) | xxxflags);
				saa7146_write(dev, MC2, MASK_02|MASK_18);		

				/* start video-dma1 just in case it is not running already */
				saa7146_write(dev, MC1, MASK_06|MASK_22);			
			}
		}
	}

	/* check if extension handles any other interrupt */
	if( 0 != isr_ext ) {
		if( NULL != dev->ext->irq_func) {
			dev->ext->irq_func(dev,&isr);
		}
	}
	
	/* isr might get modified within the extension irq-function */
	
	if( 0 != (isr|isr_ext) ) {
		DEB_INT(("%s: warning: unhandled interrupt detected: isr:0x%08x, isr_ext:0x%08x\n",dev->name,isr,isr_ext));
	}
}

void interrupt_hw(int irq, void *dev_id, struct pt_regs *regs)
{
	struct capture_device *dev = (struct capture_device*)dev_id;
	u32 isr = 0;
		
	/* read out the interrupt status register */
	isr = saa7146_read(dev, ISR);
	
	/* is this our interrupt? */
	if ( 0 == isr ) {
		/* nope, some other device */
		return;
	}
	dev->irq_left = isr;

	/* schedule our bottom-half */	
	dev->bh_handler.routine = interrupt_bh;
	dev->bh_handler.data = dev;
	queue_task(&dev->bh_handler, &tq_immediate);
	mark_bh(IMMEDIATE_BH);

 	/* clear interrupts */
	saa7146_write(dev, ISR, isr);
}

/********************************************************************************/
/* interrupt helper-functions                                                   */

void release_irq_if_possible(struct capture_device *dev)
{
	DEB_EE(("saa7146_v4l2.o: release_irq_if_possible called.\n"));

	if (dev->open_count == 0) {

		/* shut down all video dma transfers */
		saa7146_write(dev, MC1, 0x00ff0000);
		
		/* disable all irqs, release irq-routine */
		dev->irq_mask = 0;
		saa7146_write(dev, IER, dev->irq_mask);
		saa7146_write(dev, ISR, 0xffffffff);

		free_irq(dev->pci_device->irq, (void *)dev);
		
		DEB_EE(("saa7146_v4l2.o: irq released.\n"));
	} else {
		DEB_EE(("saa7146_v4l2.o: cannot release irq, still in use.\n"));
	}
}

int request_irq_if_necessary(struct capture_device *dev)
{
	int result = 0;
	
	/* request irq, if device is opened for the first time */
	if (dev->open_count == 1) {

		/* request an interrupt for the saa7146 */
		result = request_irq(	dev->pci_device->irq,
					interrupt_hw, SA_SHIRQ | SA_INTERRUPT,
					dev->name,
					(void *)dev);			
		switch(result) {
	
			case -EINVAL:
			{
				printk("saa7146_v4l2.o: Bad irq number or handler\n");
				return -EINVAL;
			}
			case -EBUSY:
			{
				printk("saa7146_v4l2.o: IRQ %d busy, change your PnP config in BIOS\n",dev->pci_device->irq);
				return -EBUSY;
			}
			case 0:
			{
				DEB_EE(("saa7146_v4l2.o: irq requested.\n"));
				break;
			}
			default:
			{
				DEB_EE(("saa7146_v4l2.o: request_irq_if_necessary called.\n"));
				return result;
			}
		}
	} else {
		DEB_EE(("saa7146_v4l2.o: irq already requested.\n"));
	}
	
	return 0;
}

/*********************************************************************************/
/* vbi functions                                                  */

void reprogram_vbi_data_capture(struct capture_device* dev)
{
	int count = 0, i = 0;
	int pitch = dev->vbi_pixel_to_capture*dev->vbi_depth;
			
	for(i = 0; i < 2; i++) {

		/* reprogram video dma3 */
		dev->rps1[ count++ ] = CMD_WR_REG | (6 << 8) | BASE_ODD3/4;
		dev->rps1[ count++ ] = virt_to_bus(dev->vbi_mem[i])+(pitch*dev->vbi_count);
		dev->rps1[ count++ ] = virt_to_bus(dev->vbi_mem[i]);
		dev->rps1[ count++ ] = (virt_to_bus(dev->vbi_mem[i])+(pitch*dev->vbi_count*2))-1;
		dev->rps1[ count++ ] = pitch;
		dev->rps1[ count++ ] = 0x0;
		dev->rps1[ count++ ] = (dev->vbi_count<<16)|((dev->vbi_pixel_to_capture*2)<<0);

		/* upload video-dma3 */
		dev->rps1[ count++ ] = CMD_UPLOAD | MASK_04;			

		/* turn on video-dma3 */
		dev->rps1[ count++ ] = CMD_WR_REG_MASK | (MC1/4);		
		dev->rps1[ count++ ] = MASK_04 | MASK_20;	    		/* => mask */
		dev->rps1[ count++ ] = MASK_04 | MASK_20;			/* => values */
		
		/* wait for o_fid_a/b / e_fid_a/b toggle */
		dev->rps1[ count++ ] = CMD_PAUSE | ( dev->current_hps_sync == SAA7146_HPS_SYNC_PORT_A ? MASK_12 : MASK_14);	/* wait for o_fid_a/b */
		dev->rps1[ count++ ] = CMD_PAUSE | ( dev->current_hps_sync == SAA7146_HPS_SYNC_PORT_A ? MASK_11 : MASK_13);	/* wait for e_fid_a/b */

		/* wait for video dma 3 done */
		dev->rps1[ count++ ] = CMD_PAUSE | MASK_04;

		/* turn off video-dma3 */
		dev->rps1[ count++ ] = CMD_WR_REG_MASK | (MC1/4);		
		dev->rps1[ count++ ] = MASK_04 | MASK_20;	    		/* => mask */
		dev->rps1[ count++ ] = MASK_20;					/* => values */

		/* generate interrupt */
		dev->rps1[ count++ ] = CMD_INTERRUPT;				
	}

	/* jump to the "beginning" */
	dev->rps1[ count++ ] = CMD_JUMP;					
	dev->rps1[ count++ ] = virt_to_bus(&dev->rps1[0]);

/*
	printk("BASE_ODD3:      0x%08x\n", saa7146_read(dev, BASE_ODD3));
	printk("BASE_EVEN3:     0x%08x\n", saa7146_read(dev, BASE_EVEN3));
	printk("PROT_ADDR3:     0x%08x\n", saa7146_read(dev, PROT_ADDR3));
	printk("PITCH3:         0x%08x\n", saa7146_read(dev, PITCH3));
	printk("BASE_PAGE3:     0x%08x\n", saa7146_read(dev, BASE_PAGE3));
	printk("NUM_LINE_BYTE3: 0x%08x\n", saa7146_read(dev, NUM_LINE_BYTE3));
	printk("BRS_CTRL:       0x%08x\n", saa7146_read(dev, BRS_CTRL));
*/

	/* enable rps1 irqs */
	dev->irq_mask |= MASK_28;	
	saa7146_write(dev, IER, dev->irq_mask);

	/* start rps1 */
	saa7146_write(dev, RPS_ADDR1, virt_to_bus(&dev->rps1[ 0]));
	saa7146_write(dev, MC1, (MASK_13 | MASK_29));	
}

int	vbi_open(struct v4l2_device *v, int flags, void **idptr)
{
	struct capture_device *dev = (struct capture_device *)(v-1);
	int result = 0, index = 0, i = 0;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
	struct wait_queue wait = {current, NULL};
#else
	DECLARE_WAITQUEUE(wait, current);
#endif
	int pitch = 0;
    
	DEB_EE(("saa7146_v4l2.o: vbi_open called.\n"));

	/* check for another vbi-open */
	if( 0 != dev->vbi_in_use ) {
		return -EBUSY;
	}
	MOD_INC_USE_COUNT;
	dev->vbi_in_use++;
	dev->open_count++;

	if( 0 != (result = request_irq_if_necessary(dev))) {

		dev->open_count--;
		dev->vbi_in_use--;

		MOD_DEC_USE_COUNT;

		return result;
	}
	
	*idptr = dev;

	dev->vbi_pixel_to_capture = 720;
	dev->vbi_depth = 1;	
	
	if( 0 != dev->ext ) {
		if( NULL != dev->ext->vbi_bypass ) {
			if( 0 == dev->ext->vbi_bypass(dev)) {
				dev->vbi_depth = 2;
			}
		}
	}

	/* reset counter values */
	dev->vbi_current = 0;
	dev->vbi_counter = 0;
	dev->vbi_last_read = -1;
	
	/* once again, a bug in the saa7146: the brs acquisition
	   is buggy and especially the BXO-counter does not work
	   as specified. there is this workaround, but please
	   don't let me explain it. ;-) */

	pitch = dev->vbi_pixel_to_capture*dev->vbi_depth;
	/* setup some basic programming, just for the workaround */
	saa7146_write(dev, BASE_EVEN3, 	 virt_to_bus(dev->vbi_mem[0]));
	saa7146_write(dev, BASE_ODD3,	 virt_to_bus(dev->vbi_mem[0])+(pitch*dev->vbi_count));
	saa7146_write(dev, PROT_ADDR3,	(virt_to_bus(dev->vbi_mem[0])+(pitch*dev->vbi_count*2))-1);
	saa7146_write(dev, PITCH3, 	pitch);
	saa7146_write(dev, BASE_PAGE3,	0x0);
	saa7146_write(dev, NUM_LINE_BYTE3, (dev->vbi_count<<16)|((dev->vbi_pixel_to_capture*2)<<0));
	saa7146_write(dev, MC2, MASK_04|MASK_20);
	
	/* we have to do the workaround two times to be sure that
	   everything is ok */
	for(i = 0; i < 2; i++) {

		saa7146_write(dev, NUM_LINE_BYTE3, (1<<16)|(2<<0));
		saa7146_write(dev, MC2, MASK_04|MASK_20);
	
		index = 0;

		/* load brs-control register */
		dev->rps1[index++] = CMD_WR_REG | (1 << 8) | (BRS_CTRL/4);
		/* BXO = 1h, BRS to outbound */
		dev->rps1[index++]=0xc000008c;   
		/* wait for vbi_a */
		dev->rps1[index++] = CMD_PAUSE | MASK_10;
		/* upload brs */
		dev->rps1[index++] = CMD_UPLOAD | MASK_08;
		/* load brs-control register */
		dev->rps1[index++] = CMD_WR_REG | (1 << 8) | (BRS_CTRL/4);
		/* BYO = 1, BXO = NQBIL (=1728 for PAL, for NTSC this is 858*2) - NumByte3 (=1440) = 288 */
		dev->rps1[index++] = ((1728-(dev->vbi_pixel_to_capture*2)) << 7) | MASK_19;
		/* wait for brs_done */
		dev->rps1[index++] = CMD_PAUSE | MASK_08;
		/* upload brs */
		dev->rps1[index++] = CMD_UPLOAD | MASK_08;
		/* load video-dma3 NumLines3 and NumBytes3 */
		dev->rps1[index++] = CMD_WR_REG | (1 << 8) | (NUM_LINE_BYTE3/4);
		/* dev->vbi_count*2 lines, 720 pixel (= 1440 Bytes) */
		dev->rps1[index++]= ((dev->vbi_count) << 16) | (dev->vbi_pixel_to_capture*2);
		/* load brs-control register */
		dev->rps1[index++] = CMD_WR_REG | (1 << 8) | (BRS_CTRL/4);
		/* Set BRS right: note: this is an experimental value for BXO (=> PAL!) */
		dev->rps1[index++] = (540 << 7) | (dev->vbi_start << 19);    
		/* wait for brs_done */
		dev->rps1[index++] = CMD_PAUSE | MASK_08;
		/* upload brs and video-dma3*/
		dev->rps1[index++] = CMD_UPLOAD | MASK_08 | MASK_04;
		/* load mc2 register: enable dma3 */
		dev->rps1[index++] = CMD_WR_REG | (1 << 8) | (MC1/4);
		dev->rps1[index++] = MASK_20 | MASK_04;
		/* generate interrupt */
		dev->rps1[index++] = CMD_INTERRUPT;
		/* stop rps1 */
		dev->rps1[index++] = CMD_STOP;
	
		/* enable rps1 irqs */
		dev->irq_mask |= MASK_28;	
		saa7146_write(dev, IER, dev->irq_mask);

		/* prepare to wait to be woken up by the irq-handler */
		add_wait_queue(&dev->vbi_wq, &wait);
		current->state = TASK_INTERRUPTIBLE;

		/* start rps1 to enable workaround */
		saa7146_write(dev, RPS_ADDR1, virt_to_bus(&dev->rps1[ 0]));
		saa7146_write(dev, MC1, (MASK_13 | MASK_29));	
		
		schedule();

		DEB_VBI(("saa7146_v4l2.o: vbi_open: brs workaround %d/1.\n",i));
	
		remove_wait_queue(&dev->vbi_wq, &wait);
		current->state = TASK_RUNNING;

		/* disable rps1 irqs */
		dev->irq_mask &= ~MASK_28;	
		saa7146_write(dev, IER, dev->irq_mask);

		/* stop video-dma3 */
		saa7146_write(dev, MC1, MASK_20);

		if(signal_pending(current)) {
		
			DEB_EE(("saa7146_v4l2.o: vbi_open aborted.\n"));

			/* stop rps1 for sure */
			saa7146_write(dev, MC1, MASK_29);

			dev->open_count--;
			release_irq_if_possible(dev);

			dev->vbi_in_use--;
			MOD_DEC_USE_COUNT;

			return -EINTR;
		}
	}

	/* reset some values */
	dev->vbi_current = 0;
	dev->vbi_counter = 0;

	/* this code is here for vbi capturing with revision 0 boards. i hope someone gets 
	   this to work in the future ... */
	if( dev->revision == 0 ) {

		int pitch = dev->vbi_pixel_to_capture*dev->vbi_depth;

		saa7146_write(dev, BASE_ODD3,	virt_to_bus(dev->vbi_mem[0])+(pitch*dev->vbi_count));
		saa7146_write(dev, BASE_EVEN3,	virt_to_bus(dev->vbi_mem[0]));
		saa7146_write(dev, PROT_ADDR3,	(virt_to_bus(dev->vbi_mem[0])+(pitch*dev->vbi_count*2))-1);
		saa7146_write(dev, PITCH3,	pitch);
		saa7146_write(dev, BASE_PAGE3,	0x0);
		saa7146_write(dev, NUM_LINE_BYTE3,	(dev->vbi_count<<16)|((dev->vbi_pixel_to_capture*2)<<0));
		saa7146_write(dev, MC2, MASK_04|MASK_20);

		saa7146_write(dev, BRS_CTRL, (0 << 7) | (dev->vbi_start << 19));
		saa7146_write(dev, MC2, MASK_08|MASK_24);

		saa7146_write(dev, MC1, MASK_04|MASK_20);

		mdelay(5);

		printk("BASE_ODD3:      0x%08x\n", saa7146_read(dev, BASE_ODD3));
		printk("BASE_EVEN3:     0x%08x\n", saa7146_read(dev, BASE_EVEN3));
		printk("PROT_ADDR3:     0x%08x\n", saa7146_read(dev, PROT_ADDR3));
		printk("PITCH3:         0x%08x\n", saa7146_read(dev, PITCH3));
		printk("BASE_PAGE3:     0x%08x\n", saa7146_read(dev, BASE_PAGE3));
		printk("NUM_LINE_BYTE3: 0x%08x\n", saa7146_read(dev, NUM_LINE_BYTE3));
		printk("BRS_CTRL:       0x%08x\n", saa7146_read(dev, BRS_CTRL));

	} else {
		/* now comes the real programming */
		reprogram_vbi_data_capture(dev);
	}

	/* nothing to start here, the process gets triggered
	   by the irq routine */

	dev->ext->vbi_open(dev);
	        
#ifdef __MY_VBI_DEBUG__
	/* disable rps1 irqs */
	dev->irq_mask &= ~MASK_28;	
	saa7146_write(dev, IER, dev->irq_mask);

	/* stop rps1 for sure */
	saa7146_write(dev, MC1, MASK_29);

	/* disable FIDA/B irqs */
	dev->irq_mask &= ~MASK_07|MASK_08;	
	saa7146_write(dev, IER, dev->irq_mask);

	/* overlay display -- start */
	saa7146_write(dev, BASE_ODD3,  (u32)dev->fb.base[0]);
	saa7146_write(dev, BASE_EVEN3,   (u32)(dev->fb.base[0]+(80*4096)));
	saa7146_write(dev, PROT_ADDR3, ((u32)(dev->fb.base[0]+4096*289)-1));
	saa7146_write(dev, PITCH3, (4096/2));
	saa7146_write(dev, BASE_PAGE3, 0);
	saa7146_write(dev, NUM_LINE_BYTE3, (64<<16)|((dev->vbi_pixel_to_capture*2)<<0)); // set above!
	saa7146_write(dev, MC2, MASK_04|MASK_20);

	saa7146_write(dev, BRS_CTRL, ((2-dev->vbi_depth))|(0<<19)); // |(6<<19)|(2<<7));
	saa7146_write(dev, MC2, MASK_08|MASK_24);

	printk("BASE_ODD3:      0x%08x\n", saa7146_read(dev, BASE_ODD3));
	printk("BASE_EVEN3:     0x%08x\n", saa7146_read(dev, BASE_EVEN3));
	printk("PROT_ADDR3:     0x%08x\n", saa7146_read(dev, PROT_ADDR3));
	printk("PITCH3:         0x%08x\n", saa7146_read(dev, PITCH3));
	printk("BASE_PAGE3:     0x%08x\n", saa7146_read(dev, BASE_PAGE3));
	printk("NUM_LINE_BYTE3: 0x%08x\n", saa7146_read(dev, NUM_LINE_BYTE3));
	printk("BRS_CTRL:       0x%08x\n", saa7146_read(dev, BRS_CTRL));

	saa7146_write(dev, MC1, MASK_04|MASK_20);
	/* overlay display -- stop */
#endif
	return 0;
}

void	vbi_close(void *id)
{
	struct capture_device *dev = (struct capture_device *)id;
	
	DEB_EE(("saa7146_v4l2.o: vbi_close called.\n"));

	if( 0 == dev->revision ) {
	} else {
		/* disable rps1 irqs */
		dev->irq_mask &= ~MASK_28;	
		saa7146_write(dev, IER, dev->irq_mask);

		/* stop rps1 for sure */
		saa7146_write(dev, MC1, MASK_29);
	}

	/* stop video dma3  */
	saa7146_write(dev, MC1, MASK_20);

	dev->open_count--;
	release_irq_if_possible(dev);

	dev->vbi_in_use--;

	dev->ext->vbi_close(dev);

	MOD_DEC_USE_COUNT;
}

int	vbi_ioctl(void *id, unsigned int cmd, void *arg) 
{
	struct capture_device *dev = (struct capture_device *)id;

	switch(cmd)
	{
	case VIDIOC_G_STD:
	{
		struct v4l2_standard *std = arg;
	
		/* fixme: simply return the standard ? */
		memcpy(std, &dev->standard, sizeof(struct v4l2_standard));
		DEB_VBI(("saa7146_v4l2.o: VIDIOC_G_STD (vbi) '%s'.\n",std->name));

		break;
	}
	case VIDIOC_QUERYCAP:
	{
		struct v4l2_capability *c = arg;

		/* clear out everything */
		memset(c, 0x0, sizeof(struct v4l2_capability));
		
		DEB_VBI(("saa7146_v4l2.o: VIDIOC_QUERYCAP (vbi) called.\n"));
		
		strcpy(c->name, dev->vbi.name);
		c->type = V4L2_TYPE_VBI;
		c->flags = V4L2_FLAG_READ | V4L2_FLAG_SELECT | V4L2_FLAG_DATA_SERVICE;

		break;
	}
	case VIDIOC_G_FMT:
	{
		struct v4l2_format *fmt = arg;
		struct v4l2_vbi_format *vbi_fmt = &fmt->fmt.vbi;
		int id = 0;

		DEB_VBI(("saa7146_v4l2.o: VIDIOC_G_FMT called.\n"));

		/* we have to distinguish between different 
		   standards in some things (PAL/NTSC/SECAM) */
		id = v4l2_video_std_confirm(&dev->standard);

		fmt->type = V4L2_BUF_TYPE_VBI;

		vbi_fmt->sampling_rate		= 13500000*dev->vbi_depth;
		vbi_fmt->offset			= 248; /* todo */ 
		vbi_fmt->samples_per_line	= dev->vbi_depth * dev->vbi_pixel_to_capture;
		vbi_fmt->sample_format		= V4L2_VBI_SF_UBYTE;
		vbi_fmt->start[0] 		= dev->vbi_start;		

		switch( id ) {
			case V4L2_STD_NTSC: {
				vbi_fmt->start[1] = dev->vbi_start+262;
				break;
			}
			case V4L2_STD_PAL: {
				vbi_fmt->start[1] = dev->vbi_start+312;
				break;
			}
			case V4L2_STD_SECAM : {
				/* fixme */
				return -EINVAL;
			}
			default: { /* no known standard, should not happen */
				printk("saa7146_v4l2.o: VIDIOC_G_FMT (vbi) no known standard.\n");
				return -EINVAL;
			}
		}

		vbi_fmt->count[0]	= dev->vbi_count;		
		vbi_fmt->count[1]	= dev->vbi_count;		
		vbi_fmt->flags		= 0;

		return 0;
	}

	case VIDIOC_S_FMT:
	{
		struct v4l2_format *fmt = arg;
		struct v4l2_vbi_format *vbi_fmt = &fmt->fmt.vbi;
		int id = 0, lines_per_field = 0;

		DEB_VBI(("saa7146_v4l2.o: VIDIOC_S_FMT called.\n"));

		/* we must have the same count on both fields */
		if( vbi_fmt->count[0] != (vbi_fmt->count[1]) ) {
			vbi_fmt->count[1] = vbi_fmt->count[0];
		}

		/* we have to distinguish between different 
		   standards in some things (PAL/NTSC/SECAM) */
		id = v4l2_video_std_confirm(&dev->standard);

		vbi_fmt->sampling_rate		= 13500000*dev->vbi_depth;
		vbi_fmt->offset			= 318; /* todo */ 
		vbi_fmt->samples_per_line	= dev->vbi_depth * dev->vbi_pixel_to_capture;
		vbi_fmt->sample_format		= V4L2_VBI_SF_UBYTE;

		switch( id ) {
			case V4L2_STD_NTSC: {
				lines_per_field = 262;
				break;
			}
			case V4L2_STD_PAL: {
				lines_per_field = 312;
				break;
			}
			case V4L2_STD_SECAM : {
				lines_per_field = 262;
				break;
			}
			default: { /* no known standard */
				printk("saa7146_v4l2.o: VIDIOC_S_FMT (vbi) no known standard.\n");
				return -EINVAL;
			}
		}
			
		if( vbi_fmt->count[0] > lines_per_field ) {
			vbi_fmt->count[0] = lines_per_field;
			vbi_fmt->count[1] = vbi_fmt->count[0];
		}
			
		/* we must begin with the same start on both fields */
		if( vbi_fmt->start[0] != (vbi_fmt->start[1]-lines_per_field) ) {
			vbi_fmt->start[1] = vbi_fmt->start[0]+lines_per_field;
		}

		if( (vbi_fmt->start[0] + vbi_fmt->count[0]) > lines_per_field ) {
			vbi_fmt->count[0] = lines_per_field-vbi_fmt->start[0];
			vbi_fmt->count[1] = vbi_fmt->count[0];
		}
	
		/* check if we have enough memory reserverd */
		if( 12 != ((dev->vbi_count * dev->vbi_depth * dev->vbi_pixel_to_capture)/PAGE_SIZE)+1 ) {
			/* fixme: we should handle this more gracefully */
			return -ENOMEM;		
		}
			
		/* reset counter values */
		dev->vbi_current = 0;
		dev->vbi_counter = 0;
		break;
	}
	default:
		return -ENOIOCTLCMD;
	}
	
	return 0;
}

int	vbi_poll(void *id, struct file *file, poll_table *table)
{
	struct capture_device *dev = (struct capture_device *)id;
	unsigned int mask = 0;
	
	DEB_EE(("saa7146_v4l2.o: vbi_poll() called @ jiffies:%li.\n",jiffies));

	if( dev->revision == 0) {
	} else {

		poll_wait(file, &dev->vbi_wq, table);

		/* check the sequence counter of the last buffer read out
		   and the sequence counter of the current buffer.
		   if they are not equal, we a new vbi frame is available */
		if( dev->vbi_last_read != dev->vbi_counter ) {
			mask |= (POLLIN|POLLRDNORM);
		}
	}

	DEB_EE(("saa7146_v4l2.o: vbi_poll() called @ jiffies:%li.\n",jiffies));
	return mask;
}

long	vbi_read(void *id, char *buf, unsigned long count, int noblock)
{
	struct capture_device *dev = (struct capture_device *)id;

	DEB_EE(("saa7146_v4l2.o: vbi_read called, jiffies:%li.\n",jiffies));

	if(count != dev->vbi_depth*dev->vbi_pixel_to_capture*(2*dev->vbi_count) ) {
		printk("saa7146_v4l2.o: vbi_read, illegal count %lu. (depth:%d,ptc:%d,count:%d => %d bytes)\n", count,dev->vbi_depth,dev->vbi_pixel_to_capture,(2*dev->vbi_count),dev->vbi_depth*dev->vbi_pixel_to_capture*(2*dev->vbi_count));
		return -EINVAL;
	}

	/* this code is here for vbi capturing with revision 0 boards. i hope someone gets 
	   this to work in the future ... */
	if( dev->revision == 0) {
		DEB_VBI(("saa7146_v4l2.o: vbi_read called, noblock:%d, current:%d, counter:%lu.\n",noblock,!dev->vbi_current,dev->vbi_counter));
		copy_to_user(buf,dev->vbi_mem[0],sizeof(char)*count);

		printk("BASE_ODD3:      0x%08x\n", saa7146_read(dev, BASE_ODD3));
		printk("BASE_EVEN3:     0x%08x\n", saa7146_read(dev, BASE_EVEN3));
		printk("PROT_ADDR3:     0x%08x\n", saa7146_read(dev, PROT_ADDR3));
		printk("PITCH3:         0x%08x\n", saa7146_read(dev, PITCH3));
		printk("BASE_PAGE3:     0x%08x\n", saa7146_read(dev, BASE_PAGE3));
		printk("NUM_LINE_BYTE3: 0x%08x\n", saa7146_read(dev, NUM_LINE_BYTE3));
		printk("BRS_CTRL:       0x%08x\n", saa7146_read(dev, BRS_CTRL));

	} else {

#ifdef axa
		/* some apps don't use the 'poll'-call to sync with the
		   vbi-capture process. we check for the sequence counter
		   of the last buffer read out and the sequence counter
		   of the current buffer. if they are equal, we have to
		   wait for the next frame */
		if( dev->vbi_last_read == dev->vbi_counter ) {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
			struct wait_queue wait = {current, NULL};
#else
			DECLARE_WAITQUEUE(wait, current);
#endif

			DEB_VBI(("saa7146_v4l2.o: vbi_read. waiting for next frame...\n"));
	
			add_wait_queue(&dev->vbi_wq, &wait);
			current->state = TASK_INTERRUPTIBLE;
	
			schedule();
	
			if(signal_pending(current)) {
				remove_wait_queue(&dev->vbi_wq, &wait);
				current->state = TASK_RUNNING;
				DEB_VBI(("saa7146_v4l2.o: vbi_read aborted.\n"));
				return -EINTR;
			}
	
			remove_wait_queue(&dev->vbi_wq, &wait);
			current->state = TASK_RUNNING;
		}
#endif
		dev->vbi_last_read = dev->vbi_counter;

		DEB_VBI(("saa7146_v4l2.o: vbi_read called, noblock:%d, current:%d, counter:%lu.\n",noblock,!dev->vbi_current,dev->vbi_counter));
		copy_to_user(buf,dev->vbi_mem[!dev->vbi_current],sizeof(char)*count);
	}
		
		
	return count;

}
int	vbi_init_done(struct v4l2_device *v)
{
	struct capture_device *dev = (struct capture_device *)(v-1);

	DEB_EE(("saa7146_v4l2.o: vbi_init_done called.\n"));

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
	dev->vbi_wq = NULL;
#else
	init_waitqueue_head(&dev->vbi_wq);
#endif

	/* start with some sane initial values for VBI */
	dev->vbi_start = 5;
	dev->vbi_count = 16;

	return 0;
}

/*********************************************************************************/
/* v4l2-functions                                                                */

int v4l2_open(struct v4l2_device *v, int flags, void **idptr)
{
	struct capture_device *dev = (struct capture_device *)v;
	int i = 0, cap = 0, result = 0;

	DEB_EE(("saa7146_v4l2.o: v4l2_open() called.\n"));

	/* check if an extension is registered */
	if( NULL == dev->ext ) {
		DEB_S(("saa7146_v4l2.o: v4l2_open: no driver extension registered.\n"));
		return -ENODEV;
	}

	MOD_INC_USE_COUNT;
	dev->open_count++;

	/* check if anyone is capturing */
	for(i = 0; i < MAX_SAA7146_OPENS; i++) {
		if (0 == dev->open_data[i].noncapturing && 0 != dev->open_data[i].isopen) {
			cap = 1;
			break;
		}
	}

	/* search for an open "slot" */
	for(i = 0; i < MAX_SAA7146_OPENS; i++) {
		if (0 == dev->open_data[i].isopen)
			break;
	}
	
	/* check if we have an open "slot" */
	if(MAX_SAA7146_OPENS == i) {
		DEB_S(("saa7146_v4l2.o: v4l2_open: no more opens on this device.\n"));
		dev->open_count--;
		MOD_DEC_USE_COUNT;
		return -EBUSY;
	}

	memset(&dev->open_data[i], 0, sizeof(struct device_open));
	
	if(0 != (flags & O_NONCAP)) {
		dev->open_data[i].noncapturing = 1;
	} else {
		if(0 != cap) {
			DEB_S(("saa7146_v4l2.o: v4l2_open: no more capturing opens on this device.\n"));
			dev->open_count--;
			MOD_DEC_USE_COUNT;
			return -EBUSY;
		}

		if( 0 != build_read_capture_buffer(dev)) {
			printk("saa7146_v4l2.o: v4l2_open: allocating read capture buffer failed.\n");
			dev->open_count--;
			MOD_DEC_USE_COUNT;
			return -EBUSY;
		}

		dev->open_data[i].noncapturing = 0;
		memset(&dev->perf,0x0,sizeof(struct v4l2_performance));

		if( 0 != (result = request_irq_if_necessary(dev))) {
			dev->open_count--;
			MOD_DEC_USE_COUNT;
			return result;
		}
		/*  Keep track of whether there is a capturing open  */
		++dev->capturing_opens;
	}
	
	dev->ext->v4l2_open(dev);
	
	dev->open_data[i].isopen = 1;
	dev->open_data[i].dev = dev;
	*idptr = &dev->open_data[i];

	if( 0 == (flags & O_NONCAP)) {
	}
		
	return 0;
}

void v4l2_close(void *id)
{
	struct device_open *o = (struct device_open *)id;
	struct capture_device *dev = o->dev;
	
	int i = 0;

	DEB_EE(("saa7146_v4l2.o: v4l2_close() called.\n"));

        dev->open_count--;

	o->isopen = 0;

	if (0 != o->noncapturing) {
	} else {
	
		--dev->capturing_opens;

		capture_streamoff(dev, V4L2_BUF_TYPE_CAPTURE);

		/*  Stop capturing and free all resources used for capture. */
		if( 0 != dev->stream_contig_map ) {
			DEB_EE(("saa7146_v4l2.o: releasing v4l1 capture buffers...\n"));
			unallocate_v4l_buffers(dev);
		} else {
			DEB_EE(("saa7146_v4l2.o: releasing v4l2 capture buffers...\n"));
			for (i = 0; i < dev->stream_buffers_requested; ++i) {
				unallocate_single_buffer(dev,&dev->stream_buf[i]);
			}
		}
		DEB_EE(("saa7146_v4l2.o: releasing read buffer...\n"));
		unallocate_single_buffer(dev,&dev->read_buffer);

		release_irq_if_possible(dev);
	}

	dev->ext->v4l2_close(dev);
	MOD_DEC_USE_COUNT;
}

long v4l2_write(void *id, const char *buf, unsigned long count, int noblock)
{
	/* we don't handle write */
	return -EINVAL;
}


int saa7146_s_std(struct device_open *o, struct v4l2_standard *std)
{
	struct capture_device *dev = o->dev;

	int	id  = 0, i = 0;

	if ((o->noncapturing && dev->capturing_opens)) // || dev->stream_buffers_mapped)
		return -EPERM;

	if ( 0 != dev->streaming ) {
		DEB_EE(("saa7146_v4l2.o: saa7146_s_std illegal while streaming.\n"));
		return -EPERM;
	}

	id = v4l2_video_std_confirm(std);

	if( id != V4L2_STD_PAL && id != V4L2_STD_NTSC && id != V4L2_STD_SECAM ) {
		return -EINVAL;
	}

	memcpy(&dev->standard, std, sizeof(struct v4l2_standard));

	/* switch standard here */
	switch( id ) {
		case V4L2_STD_NTSC: {
			i = SAA7146_NTSC;
			break;
		}
		case V4L2_STD_SECAM: {
			i = SAA7146_SECAM;
			break;
		}
		default: /* case V4L2_STD_PAL:*/ {
			i = SAA7146_PAL;
			break;
		}
	}
	memcpy(&dev->modes_constants, &saa7146_standards[i], sizeof(struct saa7146_modes_constants));

	dev->capability.maxwidth = dev->modes_constants.h_max_out; 
	dev->capability.maxheight = dev->modes_constants.v_max_out; 

	saa7146_use_overlay_data(dev);

	DEB_EE(("saa7146_v4l2.o: saa7146_s_std - '%s'.\n",std->name));

	return 0;
}

int saa7146_enumstd(struct capture_device* dev, struct v4l2_enumstd *e)
{
	int id = 0;

	DEB_EE(("saa7146_v4l2.o: saa7146_enumstd %d.\n",e->index));
		
	switch( e->index ) {
		case 0: {
			id = V4L2_STD_PAL;
			break;
		}
		case 1: {
			id = V4L2_STD_NTSC;
			break;
		}
		case 2: {
			id = V4L2_STD_SECAM;
			break;
		}
		default: {
			return -EINVAL;
		}
	}						

	v4l2_video_std_construct(&e->std, id, 0);
	e->inputs = (u32)-1;	/* we assume: all inputs support PAL/NTSC/SECAM */
	e->outputs = 0;		/* we don't have any outputs */

	return 0;
}

/*  The arguments are already copied into kernel memory, so don't use
    copy_from_user() or copy_to_user() on arg.  */
int v4l2_ioctl(void *id, unsigned int cmd, void *arg) 
{
	struct device_open *o = (struct device_open *)id;
	struct capture_device *dev = o->dev;
	
	int result = 0, ee = 0;

	int flags = 0;

	/* check if extension handles the command */
	for(ee = 0; dev->ext->ioctls[ee].flags != 0; ee++) {
		if( cmd == dev->ext->ioctls[ee].cmd )
			break;
	}
	flags = dev->ext->ioctls[ee].flags;
	
	if( 0 != (flags & SAA7146_EXCLUSIVE) ) {
//		DEB_D(("saa7146_v4l2.o: extension handles ioctl exclusive.\n"));
		result = dev->ext->ioctl(o, cmd, arg);
		return result;
	}

	if( 0 != (flags & SAA7146_BEFORE) ) {
//		DEB_D(("saa7146_v4l2.o: extension handles ioctl before.\n"));
		result = dev->ext->ioctl(o, cmd, arg);
		if( -EAGAIN != result ) {
			return result;
		}
	}

	switch(cmd)
	{
	case VIDIOC_G_FBUF:
	{
		DEB_EE(("saa7146_v4l2.o: VIDIOC_G_FBUF.\n"));

		memcpy(arg, &dev->fb, sizeof(struct v4l2_framebuffer));
		
		break;
	}
	case VIDIOC_S_FBUF:
	{
		struct v4l2_framebuffer *fb = arg;

		DEB_EE(("saa7146_v4l2.o: VIDIOC_S_FBUF.\n"));
		
		if(!capable(CAP_SYS_ADMIN))
			return -EPERM;

		dev->fb.flags = fb->flags;
		dev->fb.base[0] = fb->base[0];
		dev->fb.base[1] = fb->base[1];
		dev->fb.base[2] = fb->base[2];

//		printk("0x%08x, 0x%08x, 0x%08x\n",(int)fb->base[0],(int)fb->base[1],(int)fb->base[2]);

		memcpy(&dev->fb.fmt, &fb->fmt, sizeof(struct v4l2_pix_format));
		if( 0 == (fb->fmt.flags & V4L2_FMT_FLAG_BYTESPERLINE)) {
			fb->fmt.bytesperline = (fb->fmt.width*fb->fmt.depth)>>3;
		}

		/* determine the correct overlay format */
		switch (dev->fb.fmt.pixelformat)
		{
		case V4L2_PIX_FMT_RGB332:
			dev->overlay_palette = RGB08_COMPOSED;
			break;
		case V4L2_PIX_FMT_RGB555:
		case V4L2_PIX_FMT_RGB565:
			dev->overlay_palette = RGB16_COMPOSED;
			break;
		case V4L2_PIX_FMT_BGR24:
			dev->overlay_palette = RGB24_COMPOSED;
			break;
		case V4L2_PIX_FMT_BGR32:
			dev->overlay_palette = RGB32_COMPOSED;
			break;
		case V4L2_PIX_FMT_GREY:
			dev->overlay_palette = Y8;
			break;
		case V4L2_PIX_FMT_YUYV:
			dev->overlay_palette = YUV422_COMPOSED;
			/* fixme: add two byte swap here for revision != 0 overlay! */
			break;
		case V4L2_PIX_FMT_UYVY:
			dev->overlay_palette = YUV422_COMPOSED;
			break;
		default:
			printk("saa7146_v4l2.o: VIDIOC_S_FBUF unknown format %4.4s\n", (char *)&dev->fb.fmt.pixelformat);
			return -EINVAL;
		}

		dev->overlay_data_valid = 0;

		break;
	}
	case VIDIOC_ENUM_PIXFMT:
	{
		struct v4l2_fmtdesc *f = arg;

		DEB_EE(("saa7146_v4l2.o: VIDIOC_ENUM_PIXFMT: %d.\n", f->index));

		if( 0 == dev->revision ) {
			if( f->index < 0 || f->index >= NUM_CAPPVFMT-1 )
				return -EINVAL;
		} else {
			if( f->index < 0 || f->index >= NUM_CAPPVFMT )
				return -EINVAL;
		}

		*f = cappvfmt[f->index];

		break;
	}
	case VIDIOC_ENUM_FBUFFMT:
	{
		struct v4l2_fmtdesc *f = arg;

		DEB_EE(("saa7146_v4l2.o: VIDIOC_ENUM_FBUFFMT: %d.\n",f->index));

		if( 0 == dev->revision ) {
			if ( f->index < 0 || f->index >= NUM_CAPPVFMT )
				return -EINVAL;
		} else {
			if ( f->index < 0 || f->index >= NUM_CAPPVFMT-1 )
				return -EINVAL;
		}
		
		*f = cappvfmt[f->index];

		break;
	}
	case VIDIOC_PREVIEW:
	{	
		int i = *(int*)arg;

		DEB_EE(("saa7146_v4l2.o: VIDIOC_PREVIEW: %d.\n",i));

		preview(dev,i);
/*
	printk("w:%d, h:%d, x:%d, y:%d\n",dev->overlay_width, dev->overlay_height, dev->overlay_x, dev->overlay_y);	
	printk("MC1: 0x%08x\n",saa7146_read(dev,MC1));
	printk("MC2: 0x%08x\n",saa7146_read(dev,MC2));
	printk("BASE_ODD1:      0x%08x\n", saa7146_read(dev, BASE_ODD1));
	printk("BASE_EVEN1:     0x%08x\n", saa7146_read(dev, BASE_EVEN1));
	printk("PROT_ADDR1:     0x%08x\n", saa7146_read(dev, PROT_ADDR1));
	printk("PITCH1:         0x%08x\n", saa7146_read(dev, PITCH1));
	printk("BASE_PAGE1:     0x%08x\n", saa7146_read(dev, BASE_PAGE1));
	printk("NUM_LINE_BYTE1: 0x%08x\n", saa7146_read(dev, NUM_LINE_BYTE1));
*/

//		dump_registers(dev);		
		break;
	}
	case VIDIOC_QUERYCAP:
	{
		struct v4l2_capability *c = arg;

		DEB_EE(("saa7146_v4l2.o: VIDIOC_QUERYCAP.\n"));

		memcpy(c, &dev->capability, sizeof(struct v4l2_capability));

		break;
	}
	case VIDIOC_G_FMT:
	{
		struct v4l2_format *fmt = arg;
		if (fmt->type != V4L2_BUF_TYPE_CAPTURE) {
			DEB_EE(("saa7146_v4l2.o: VIDIOC_G_FMT: wrong buffer type %d\n",fmt->type));
			return -EINVAL;
		}

		fmt->fmt.pix = dev->clientfmt;
		DEB_EE(("saa7146_v4l2.o: VIDIOC_G_FMT => \"%4.4s\".\n",(char*)&dev->clientfmt.pixelformat));

		break;
	}
	case VIDIOC_S_FMT:
	{
		struct v4l2_format *fmt = arg;

		DEB_EE(("saa7146_v4l2.o: VIDIOC_S_FMT.\n"));

		if (o->noncapturing) {
			DEB_EE(("saa7146_v4l2.o: VIDIOC_S_FMT: illegal in non-capturing open.\n"));
			return -EPERM;
		}
		
		if ( 0 != dev->streaming ) {
			DEB_EE(("saa7146_v4l2.o: VIDIOC_S_FMT: illegal while streaming.\n"));
			return -EPERM;
		}
		
		dev->clientfmt = fmt->fmt.pix;
		if (0 != capture_new_format(dev))
			return -EINVAL;
		fmt->fmt.pix = dev->clientfmt;

		build_read_capture_buffer(dev);

		break;
	}
	case VIDIOC_G_COMP:
	{
		DEB_EE(("saa7146_v4l2.o: VIDIOC_G_COMP.\n"));
		return -ENOIOCTLCMD;
	}
	case VIDIOC_S_COMP:
	{
		DEB_EE(("saa7146_v4l2.o: VIDIOC_S_COMP.\n"));
		return -ENOIOCTLCMD;
	}
	case VIDIOC_REQBUFS:
	{
		struct v4l2_requestbuffers *req = arg;
				
		if ( 0 != o->noncapturing ) {
			DEB_S(("saa7146_v4l2.o: VIDIOC_REQBUFS illegal in non-capturing open.\n"));
			return -EPERM;
		}
		if ( 0 != dev->stream_buffers_mapped ) {
			DEB_S(("saa7146_v4l2.o: VIDIOC_REQBUFS: can't request buffers if buffers are already mapped.\n"));
			return -EPERM;
		}
		if ( 0 != dev->streaming ) {
			DEB_EE(("saa7146_v4l2.o: VIDIOC_REQBUFS: illegal while streaming.\n"));
			return -EPERM;
		}

		DEB_EE(("saa7146_v4l2.o: VIDIOC_REQBUFS: %d.\n",req->count));

		mmap_unrequest_buffers(dev);
		
		if (0 != mmap_request_buffers(dev, req))
			return -EINVAL;
				
		break;
	}
	case VIDIOC_QUERYBUF:
	{
		struct v4l2_buffer *buf = arg;
		int	i;

		if (o->noncapturing) {
			DEB_EE(("saa7146_v4l2.o: VIDIOC_QUERYBUF illegal in non-capturing open.\n"));
			return -EPERM;
		}

		i = buf->index;

		if (i < 0 || i >= MAX_SAA7146_CAPTURE_BUFFERS ||
		    !dev->stream_buf[i].requested ||
		    (buf->type & V4L2_BUF_TYPE_field) != 
		     (dev->stream_buf[i].vidbuf.type & V4L2_BUF_TYPE_field)) {
			DEB_EE(("saa7146_v4l2.o: VIDIOC_QUERYBUF bad parameter (%d?).\n",i));
			return -EINVAL;
		}
		*buf = dev->stream_buf[i].vidbuf;
		DEB_EE(("saa7146_v4l2.o: VIDIOC_QUERYBUF: buffer %d, length %d.\n",i,buf->length));

		break;
	}
	case VIDIOC_QBUF:
	{
		struct v4l2_buffer *buf = arg;
		struct stream_buffer *sbuf = NULL;
		int i = 0;
		
		if (o->noncapturing) {
			DEB_EE(("saa7146_v4l2.o: VIDIOC_QBUF illegal in non-capturing open.\n"));
			return -EPERM;
		}
			
		if(buf->type != V4L2_BUF_TYPE_CAPTURE) {
			DEB_S(("saa7146_v4l2.o: VIDIOC_QBUF: wrong type.\n"));
			return -EINVAL;
		}

		i = buf->index;
		
		if (i < 0 || i >= MAX_SAA7146_CAPTURE_BUFFERS || 0 == dev->stream_buf[i].requested) {
			DEB_S(("saa7146_v4l2.o: VIDIOC_QBUF: buffer index %d is out of range (%d)\n", i, dev->stream_buf[i].requested));
			return -EINVAL;
		}

		sbuf = &dev->stream_buf[i];

		if (!(sbuf->vidbuf.flags & V4L2_BUF_FLAG_MAPPED)) {
			DEB_S(("saa7146_v4l2.o: VIDIOC_QBUF: this buffer (%p) is not mapped\n", buf));
			return -EINVAL;
		}


		DEB_CAP(("saa7146_v4l2.o: VIDIOC_QBUF.\n"));

		if (0 != capture_queuebuffer(dev, sbuf))
			return -EINVAL;

		break;
	}

	case VIDIOC_DQBUF:
	{
		struct v4l2_buffer *buf = arg;
		struct stream_buffer* sbuf = NULL;

		if (o->noncapturing) {
			DEB_EE(("saa7146_v4l2.o: VIDIOC_DQBUF illegal in non-capturing open.\n"));
			return -EPERM;
		}

		if (buf->type != V4L2_BUF_TYPE_CAPTURE) {
			DEB_S(("saa7146_v4l2.o: capture_dequeuebuffer: wrong type.\n"));
			return -EINVAL;
		}

		DEB_CAP(("saa7146_v4l2.o: VIDIOC_DQBUF.\n"));

		if(0 == (sbuf = capture_dequeuebuffer(dev))) {
			return -EINVAL;
		}
		
		*buf = sbuf->vidbuf;

		break;
	}
	case VIDIOC_STREAMON:
	{
		u32	*type = arg;
		
		if (o->noncapturing) {
			DEB_EE(("saa7146_v4l2.o: VIDIOC_STREAMON illegal in non-capturing open.\n"));
			return -EPERM;
		}

		DEB_EE(("saa7146_v4l2.o: VIDIOC_STREAMON.\n"));

		if( 0 != capture_streamon(dev, *type)) {
			return -EINVAL;
		}

		/* fixme: overlay_on is used to determine,
                   if overlay was on before streaming was enabled. if so,
                   switching off streaming will turn on overlay, again.
                   this is for satisfying 'xawtv'... */
		dev->overlay_data_valid = 0;

		break;
	}

	case VIDIOC_STREAMOFF:
	{
		u32	*type = arg;

		if (o->noncapturing) {
			DEB_EE(("saa7146_v4l2.o: VIDIOC_STREAMOFF illegal in non-capturing open.\n"));
			return -EPERM;
		}

		if (0 == dev->streaming) {
			DEB_S(("saa7146_v4l2.o: VIDIOC_STREAMOFF: not streaming\n"));
			/* fixme: anything to do here ? */
			break;
		}

		capture_streamoff(dev, *type);

		DEB_S(("saa7146_v4l2.o: VIDIOC_STREAMOFF: captured:%d, dropped:%d\n",dev->perf.frames,dev->perf.framesdropped));
		break;
	}
	case VIDIOC_G_WIN:
	{
		struct v4l2_window *w = arg;

		if ( 0 != dev->streaming ) {
			DEB_EE(("saa7146_v4l2.o: VIDIOC_G_WIN illegal while streaming.\n"));
			return -EPERM;
		}

		DEB_EE(("saa7146_v4l2.o: VIDIOC_G_WIN.\n"));
		if( 0 != saa7146_retrieve_overlay_data(dev, w)) {
			DEB_EE(("saa7146_v4l2.o: VIDIOC_G_WIN: no overlay data available.\n"));
			return -EINVAL;
		}

		break;
	}
	case VIDIOC_S_WIN:
	{
		struct v4l2_window *w = arg;
		
		if ( 0 != dev->streaming ) {
			DEB_EE(("saa7146_v4l2.o: VIDIOC_S_WIN illegal while streaming.\n"));
			return -EPERM;
		}

		DEB_EE(("saa7146_v4l2.o: VIDIOC_S_WIN: i:%d, w:%d, h:%d, x:%d, y:%d, wid:%d, dep:%d\n", no_interlace, w->width, w->height, w->x, w->y, dev->fb.fmt.width, dev->fb.fmt.depth));

		if( 0 != saa7146_store_overlay_data(dev, w)) {
			return -EINVAL;
		}
		
		saa7146_use_overlay_data(dev);
/*
		u32 hps_ctrl = 0;
		hps_ctrl = saa7146_read(dev, HPS_CTRL);
		hps_ctrl &= MASK_B3;
		saa7146_write(dev, HPS_CTRL, hps_ctrl);
		
		hps_ctrl = saa7146_read(dev, CLIP_FORMAT_CTRL);
		hps_ctrl &= ~(MASK_30|MASK_31|MASK_24|MASK_25|MASK_26|MASK_27);
			
		saa7146_write(dev, CLIP_FORMAT_CTRL, hps_ctrl);
		saa7146_write(dev, MC2, MASK_05|MASK_21);
*/

		break;
}
	case VIDIOC_G_PERF:
	{
		DEB_EE(("saa7146_v4l2.o: VIDIOC_G_PERF.\n"));

		memcpy(arg, &dev->perf, sizeof(dev->perf));

		break;
	}
	case VIDIOC_G_PARM:
	{
		struct v4l2_streamparm *sp = arg;

		if (sp->type != V4L2_BUF_TYPE_CAPTURE) {
			DEB_EE(("saa7146_v4l2.o: VIDIOC_G_PARM: invalid capture type.\n"));
			return -EINVAL;
		}
		
		sp->parm.capture = dev->capture;

		DEB_EE(("saa7146_v4l2.o: VIDIOC_G_PARM.\n"));

		break;
	}
	case VIDIOC_S_PARM:
	{
		struct v4l2_streamparm *sp = arg;
		struct v4l2_captureparm *vp = &sp->parm.capture;

		if (vp->capturemode & ~dev->capture.capability) {
			DEB_EE(("saa7146_v4l2.o: VIDIOC_S_PARM: unsupported capture capability %08X\n", vp->capturemode));
			return -EINVAL;
		}

		/* hmm, anything to do here? */

		DEB_EE(("saa7146_v4l2.o: VIDIOC_S_PARM.\n"));

		break;
	}
	case VIDIOC_G_STD:
	{
		struct v4l2_standard *std = arg;
		
		memcpy(std, &dev->standard, sizeof(struct v4l2_standard));
		DEB_EE(("saa7146_v4l2.o: VIDIOC_G_STD '%s'.\n",std->name));

		break;
	}
	/* the saa7146 supports (used in conjunction with the saa7111a for example)
	   PAL / NTSC / SECAM. if your hardware does not (or does more)
	   -- override this function in your extension */
	case VIDIOC_ENUMSTD:
	{
		struct v4l2_enumstd *e = arg;
		DEB_EE(("saa7146_v4l2.o: VIDIOC_ENUMSTD: %d\n",e->index));
		return saa7146_enumstd(dev,e);	
	}
	case VIDIOC_S_STD:
	{
		struct v4l2_standard *std = arg;
		DEB_EE(("saa7146_v4l2.o: VIDIOC_S_STD: %s\n",std->name));
		return saa7146_s_std(o,std);	
	}
	/* the saa7146 provides some controls (brightness, contrast, saturation)
	   which gets registered here. if any extension registers other controls
	   above, it must return with a value != 0, otherwise these controls won't be
	   registered ! */
	case VIDIOC_QUERYCTRL:
	{
		struct v4l2_queryctrl	*qc = arg;
		int			i;

		DEB_EE(("saa7146_v4l2.o: VIDIOC_QUERYCTRL %d.\n",qc->id));

		i = find_vctrl(qc->id);
		if (i == -EINVAL) {
			qc->flags = V4L2_CTRL_FLAG_DISABLED;
			break;
		}
		
		if (i < 0) {
			return -EINVAL;
		}
		
		/*  V4L2 filled in category and group, preserve them */
		capture_control[i].category = qc->category;
		memcpy(capture_control[i].group, qc->group, sizeof(qc->group));
		*qc = capture_control[i];

		break;
	}
	case VIDIOC_QUERYMENU:
	{
		DEB_EE(("saa7146_v4l2.o: VIDIOC_QUERYMENU.\n"));
		return -ENOIOCTLCMD;
	}
	case VIDIOC_G_CTRL:
	{
		struct v4l2_control	*vc = arg;
		int			i = find_vctrl(vc->id);
		u32			value = 0;
		
		if (i < 0) {
			DEB_EE(("saa7146_v4l2.o: VIDIOC_G_CTRL invalid value."));
			return -EINVAL;
		}
		
		value = saa7146_read(dev, BCS_CTRL);
		
		switch(i) {
			case VCTRL_HFLIP: {
				DEB_EE(("saa7146_v4l2.o: VIDIOC_G_CTRL: VCTRL_HFLIP %d.\n",value));
				value = dev->hflip;
				break;
			}
			case VCTRL_VFLIP: {
				DEB_EE(("saa7146_v4l2.o: VIDIOC_G_CTRL: VCTRL_VFLIP %d.\n",value));
				value = dev->vflip;
				break;
			}
			case VCTRL_BRIGHTNESS: {
				value = 0xff & (value >> 24);
				DEB_EE(("saa7146_v4l2.o: VIDIOC_G_CTRL: VCTRL_BRIGHTNESS %d.\n",value));
				break;
			}
			case VCTRL_CONTRAST: {
				value = 0x7f & (value >> 16);
				DEB_EE(("saa7146_v4l2.o: VIDIOC_G_CTRL: VCTRL_CONTRAST %d.\n",value));
				break;
			}
			case VCTRL_SATURATION: {
				value = 0x7f & (value >> 0);
				DEB_EE(("saa7146_v4l2.o: VIDIOC_G_CTRL: VCTRL_SATURATION %d.\n",value));
				break;
			}
			default: {
				printk("saa7146_v4l2.o: VIDIOC_G_CTRL: unknown ctrl called.\n");
				return -EINVAL;
			}
		}	

		vc->value = value;

		break;
	}

	case VIDIOC_S_CTRL:
	{
		struct v4l2_control	*vc = arg;
		int			i = find_vctrl(vc->id);
		
		int handled = 0;
		
		if (i < 0)
			return -EINVAL;

		if(i == VCTRL_BRIGHTNESS || i == VCTRL_CONTRAST || i == VCTRL_SATURATION ) {
			u32 value = saa7146_read(dev, BCS_CTRL);
		
			switch(i) {
				case VCTRL_BRIGHTNESS: {
					value &= 0x00ffffff;
					value |= (vc->value << 24);
					DEB_EE(("saa7146_v4l2.o: VIDIOC_S_CTRL VCTRL_BRIGHTNESS %d.\n",vc->value));
					break;
				}
				case VCTRL_CONTRAST: {
					value &= 0xff00ffff;
					value |= (vc->value << 16);
					DEB_EE(("saa7146_v4l2.o: VIDIOC_S_CTRL VCTRL_CONTRAST %d.\n",vc->value));
					break;
				}
				case VCTRL_SATURATION: {
					value &= 0xffffff00;
					value |= (vc->value << 0);
					DEB_EE(("saa7146_v4l2.o: VIDIOC_S_CTRL VCTRL_SATURATION %d.\n",vc->value));
					break;
				}
			}	

			i2c_status(dev);
			saa7146_write(dev, BCS_CTRL, value);
			saa7146_write(dev, MC2, MASK_22 | MASK_06 );
			i2c_status(dev);

			break;
		}

		/* fixme: it should definitely be possible to change the controls
		   while capturing is on! */
		if ( 0 != dev->streaming ) {
			DEB_EE(("saa7146_v4l2.o: VIDIOC_S_CTRL: illegal while streaming.\n"));
			return -EPERM;
		}
		
		switch(i) {
			case VCTRL_HFLIP: {
				dev->hflip = vc->value;
				DEB_EE(("saa7146_v4l2.o: VIDIOC_S_CTRL: VCTRL_HFLIP %d.\n",dev->hflip));
				handled = 1;
				break;
			}
			case VCTRL_VFLIP: {
				dev->vflip = vc->value;
				DEB_EE(("saa7146_v4l2.o: VIDIOC_G_CTRL: VCTRL_VFLIP %d.\n",dev->vflip));
				handled = 1;
				break;
			}
			default: {
				printk("saa7146_v4l2.o: VIDIOC_S_CTRL: unknown ctrl called.\n");
				return -EINVAL;
			}
		}

		if( 0 != handled ) {
			if(0 != dev->overlay) {
				dev->overlay_data_valid = 0;
				preview(dev,1);
			}
			break;
		}
			

		break;
	}
	default:
/*
		printk("saa7146_v4l2.o: no such IOCTL.\n");
*/
		return -ENOIOCTLCMD;
	}

	if( 0 != (flags & SAA7146_AFTER) ) {
	
		DEB_D(("saa7146_v4l2.o: extension handles ioctl afterwards.\n"));
	
		result = dev->ext->ioctl(o, cmd, arg);
		if( 0 != result ) {
			return result;
		}
	}
	
	return 0;

}

int v4l2_mmap(void *id, struct vm_area_struct *vma)
{
	struct device_open	*o   = (struct device_open *)id;
	struct capture_device	*dev = o->dev;
	struct stream_buffer	*buf;

	int n = 1;

	if (o->noncapturing) {
		DEB_S(("saa7146_v4l2.o: mmap() called on non-capturing open.\n"));
		return -ENODEV;
	}
	
	/* check wheter to allocate all n buffers in one contiguous map => v4l backward compatibility */

	if( 0 != dev->stream_contig_map) {
		/* v4l compatiblity */
		DEB_S(("saa7146_v4l2.o: mmap(): v4l backward compatibility.\n"));
		buf = &dev->stream_buf[0];
		n = dev->stream_buffers_requested;
	} else {
		buf = mmap_stream_buffer_from_offset(dev, vma);
		if (buf == NULL) {
			DEB_S(("saa7146_v4l2.o: mmap(): NULL offset parameter\n."));
			return -EINVAL;/* no such buffer */
		}
	}

	DEB_EE(("saa7146_v4l2.o: v4l2_mmap called: buffer: %p, length:%li\n",buf, (long)(vma->vm_end - vma->vm_start)));

	if (buf->vidbuf.length*n != (vma->vm_end - vma->vm_start)) {
		DEB_S(("saa7146_v4l2.o: mmap(): Wrong length parameter (%d:%li)\n",buf->vidbuf.length*n,(vma->vm_end - vma->vm_start)));
		return -EINVAL;/* wrong length */
	}

	if (!buf->requested) {
		DEB_S(("saa7146_v4l2.o: mmap(): buffer is not available for mapping.\n"));
		return -EINVAL;/* not requested */
	}
	
	if (buf->vidbuf.flags & V4L2_BUF_FLAG_MAPPED) {
		DEB_S(("saa7146_v4l2.o: mmap(): buffer is already mapped.\n"));
		return -EINVAL;/* already mapped */
	}

	if( 0 != dev->stream_contig_map) {
		/* v4l compatiblity */
		unallocate_v4l_buffers(dev);
		if( 0 != allocate_v4l_buffers(dev)) {
			DEB_S(("saa7146_v4l2.o: mmap(): Could not allocate v4l mmap() buffer.\n"));
			return -ENOMEM;
		}
	} else {
		unallocate_single_buffer(dev,buf);
		if( 0 != allocate_single_buffer(dev, buf)) {
			DEB_S(("saa7146_v4l2.o: mmap(): Could not allocate v4l2 mmap() buffer.\n"));
			return -ENOMEM;
		}
		buf->vidbuf.flags |= V4L2_BUF_FLAG_MAPPED;
	}
	
	dev->stream_buffers_mapped++;

	vma->vm_ops = &capture_vma_operations;
	if (vma->vm_ops->open)
		vma->vm_ops->open(vma);

	/*  Note: vma->vm_file will be set up by V4L2  */

	return 0;
}



int v4l2_poll(void *id, struct file *file, poll_table *table)
{
	struct device_open *o = (struct device_open *)id;
	struct capture_device *dev = o->dev;
	
	DEB_CAP(("saa7146_v4l2.o: v4l2_poll called @ %li jiffies.\n",jiffies));

	if( 0 != o->noncapturing ) {
		DEB_S(("saa7146_v4l2.o: poll() illegal in non-capturing open.\n"));
		return POLLERR;
	}

	if ( 0 != dev->streaming ) {

		void	*node = v4l2_q_peek_head(&dev->stream_q_done);
		/* data is ready now */
		if (node != NULL) {
			DEB_CAP(("saa7146_v4l2.o: v4l2_poll: data is ready @ %li.\n",jiffies));
			return (POLLIN | POLLRDNORM);
		}
		
		/* no data available, anything in progress to wait for ? */
		if( dev->stream_db_cur[0] == NULL && NULL == dev->stream_db_cur[1]) {
			/* hmm, anything queued at all ? */
			node = v4l2_q_peek_head(&dev->stream_q_capture);
			if (node == NULL) {
				DEB_CAP(("saa7146_v4l2.o: v4l2_poll: nothing queued to wait for.\n"));
				return POLLERR;
			}
		}
		poll_wait(file, &dev->capture_wq, table);
		return 0;	
	}

	/*  fixme: Capture is through read() call 
		look at v4l2 sample driver to fix this*/
	return POLLERR;
}

long v4l2_read(void *id, char *buf, unsigned long count, int noblock)
{
	struct device_open *o = (struct device_open *)id;
	struct capture_device *dev = o->dev;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
	struct wait_queue wait = {current, NULL};
#else
	DECLARE_WAITQUEUE(wait, current);
#endif
	struct stream_buffer *sbuf = NULL;
	
	if(o->noncapturing){
		DEB_S(("saa7146_v4l2.o: read() illegal in non-capturing open.\n"));
		return -EPERM;
	}
	if(dev->streaming) {
		DEB_S(("saa7146_v4l2.o: Can't read() when streaming is on.\n"));
		return -EPERM;
	}

	/* fixme */
	DEB_EE(("saa7146_v4l2.o: read() called:%lu\n",count));
		
	if (0 != capture_queuebuffer(dev, &dev->read_buffer)) {
		printk("saa7146_v4l2.o: read(): cannot queue read buffer\n");
		return -EINVAL;
	}

	capture_streamon(dev, V4L2_BUF_TYPE_CAPTURE);

	DEB_EE(("saa7146_v4l2.o: read() going to sleep.\n"));

	add_wait_queue(&dev->capture_wq, &wait);
	current->state = TASK_INTERRUPTIBLE;

	schedule();
	
	remove_wait_queue(&dev->capture_wq, &wait);
	current->state = TASK_RUNNING;
		
	if(signal_pending(current)) {
		capture_streamoff(dev, V4L2_BUF_TYPE_CAPTURE);
		DEB_D(("saa7146_v4l2.o: read() aborted.\n"));
		return -EINTR;
	}
	
	if (0 == (sbuf = capture_dequeuebuffer(dev))) {
		capture_streamoff(dev, V4L2_BUF_TYPE_CAPTURE);
		printk("saa7146_v4l2.0: fatal error in v4l2_read().\n");
		return -EINVAL;
	}
	
	capture_streamoff(dev, V4L2_BUF_TYPE_CAPTURE);

	/* copy data to userspace */
	if( count > sbuf->vidbuf.length ) {
		count = sbuf->vidbuf.length;
	}
	DEB_EE(("saa7146_v4l2.o: read() copying buffer (%li bytes) (vidbuf:%d).\n",count,sbuf->vidbuf.length));

	copy_to_user(buf,sbuf->vaddress,sizeof(char)*count);

	/* check if we interrupted overlay. if so, turn it on again */
	if(1 == dev->overlay) {
		preview(dev,1);
	}

	return count;
}

/* Remaining initialization of video decoder etc. This is only
   done when the device is successfully identified and registered. */
int v4l2_init_done(struct v4l2_device *v)
{
	struct capture_device *dev = (struct capture_device *)v;

	struct v4l2_capability capability = 
		{	"",
			V4L2_TYPE_CAPTURE,
			0,		/* inputs */
			1,		/* outputs */
			0,		/* audios */
			768,
			576,
			32,
			32,
			25,		/* max. framerate */
			V4L2_FLAG_READ | V4L2_FLAG_STREAMING | V4L2_FLAG_SELECT | V4L2_FLAG_PREVIEW
		};

	memcpy(&dev->capability, &capability, sizeof(struct v4l2_capability));

	DEB_EE(("saa7146_v4l2.o: v4l2_init_done called.\n"));

	/*  Capture parameters  */
	dev->capture.capability = 0;
	dev->capture.capturemode = 0;
	dev->capture.extendedmode = 0;
	dev->capture.timeperframe = 400000; /* for pal */

	/* bring the hardware to a sane state. all settings can
	   be overriden if necessary using the ioctls */

	dev->fb.capability = V4L2_FBUF_CAP_CLIPPING; // fixme: todo: |V4L2_FBUF_CAP_BITMAP_CLIPPING;
	dev->fb.flags = V4L2_FBUF_FLAG_PRIMARY;
	dev->fb.base[0] = NULL;
	dev->fb.base[1] = NULL;
	dev->fb.base[2] = NULL;

	/* adjust arbitrition control */
	saa7146_write(dev, PCI_BT_V1, 		0xffff101f);
	
	/* set dd1 stream a & b */
      	saa7146_write(dev, DD1_STREAM_B,	0x00000000);
	saa7146_write(dev, DD1_INIT, 		0x02000200);
	saa7146_write(dev, MC2, (MASK_09 | MASK_25 | MASK_10 | MASK_26));
	
	/* set initial standard */
	v4l2_video_std_construct(&dev->standard, V4L2_STD_PAL, 0);
	memcpy(&dev->modes_constants, &saa7146_standards[SAA7146_PAL], sizeof(struct saa7146_modes_constants));

	dev->current_hps_source = SAA7146_HPS_SOURCE_PORT_A;
	dev->current_hps_sync = SAA7146_HPS_SYNC_PORT_A;
	
	v4l2_q_init(&dev->stream_q_capture);
	v4l2_q_init(&dev->stream_q_done);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
	dev->capture_wq = NULL;
#else
	init_waitqueue_head(&dev->capture_wq);
#endif

	dev->clientfmt.width = 352;
	dev->clientfmt.height = 288;
	dev->clientfmt.depth = 16;
	dev->clientfmt.pixelformat = V4L2_PIX_FMT_RGB565;
	dev->clientfmt.flags = 0;
	dev->clientfmt.bytesperline = 0;
	dev->clientfmt.sizeimage = 0; /* gets set by capture_new_format() */

	capture_new_format(dev);
	
	if( 0 != dev->revision ) {
		program_capture_engine(dev);
	}

	dev->overlay_palette = -1;
		
	/* the rest */
	saa7146_set_picture_prop(dev, 0x80, 0x40, 0x40);
        							
	return 0;
}

/*********************************************************************************/
/* extension handling functions                                                  */

int saa7146_v4l2_register_extension(struct saa7146_v4l2_extension* ext)
{
	int i = 0, j = 0;
	struct capture_device* dev = 0;

	DEB_EE(("saa7146_v4l2.o: saa7146_v4l2_register_extension called.\n"));

	for(i = 0; i < saa7146_v4l2_num; i++) {

		dev = &saa7146_v4l2_devices[i];

		/* check if already handled by extension */
		if( 0 != dev->ext ) {
			continue;
		}

		if( 0 != ext->preinit(dev) ) {
			printk("saa7146_v4l2.o: pre-init failed. skipping.\n");
			continue;
		}

		/* register new i2c-bus */
		if(i2c_add_adapter(dev->i2c_adapter) < 0) {
			printk("saa7146_v4l2.o: cannot register i2c-device. skipping.\n");
			continue;
		}	
	
		if( 0 != ext->probe(dev) ) {
			printk("saa7146_v4l2.o: probe failed. skipping device.\n");
			i2c_del_adapter(dev->i2c_adapter);
			continue;
		}

		if( 0 != ext->attach(dev) ) {
			printk("saa7146_v4l2.o: attach failed. skipping device.\n");
			i2c_del_adapter(dev->i2c_adapter);
			continue;
		}
		
		/* v4l2 initialization stuff */
		strcpy(dev->v4l2.name, ext->name);
		dev->v4l2.type		= V4L2_TYPE_CAPTURE;
		dev->v4l2.minor		= ext->unit_video[j];
		dev->v4l2.open 		= v4l2_open;
		dev->v4l2.close 	= v4l2_close;
		dev->v4l2.read 		= v4l2_read;
		dev->v4l2.write 	= v4l2_write;
		dev->v4l2.ioctl 	= v4l2_ioctl;
		dev->v4l2.mmap		= v4l2_mmap;
		dev->v4l2.poll		= v4l2_poll;
		dev->v4l2.initialize 	= v4l2_init_done;

		/* device configured successfully */
		if ( 0 != v4l2_register_device(&dev->v4l2)) {
			printk("saa7146_v4l2.o: cannot register v4l2-device. skipping.\n");

			/* in this case the extension has probably 
			   already "done something", but the v4l2-device
			   could not be registered. this is not as bad as
			   it looks - you cannot access the device.
			   we simply call the 'release' function. */
			ext->detach(dev);
			i2c_del_adapter(dev->i2c_adapter);
			continue;
		}


		/* v4l2 initialization stuff (vbi) (only for revision > 0)*/
		if( NULL != ext->unit_vbi && dev->revision != 0) {

			sprintf(dev->vbi.name, "%s", ext->name);
			dev->vbi.type		= V4L2_TYPE_VBI;
			dev->vbi.minor		= ext->unit_vbi[j];
			dev->vbi.open 		= vbi_open;
			dev->vbi.close 		= vbi_close;
			dev->vbi.read 		= vbi_read;
			dev->vbi.write 		= NULL;			/* vbi_write; */
			dev->vbi.ioctl 		= vbi_ioctl;
			dev->vbi.mmap		= NULL; 		/* vbi_mmap; */
			dev->vbi.poll		= vbi_poll;
			dev->vbi.initialize 	= vbi_init_done;

			/* device configured successfully */
			if ( 0 != v4l2_register_device(&dev->vbi)) {
				printk("saa7146_v4l2.o: cannot register vbi-device. skipping.\n");
				v4l2_unregister_device(&dev->v4l2);
				ext->detach(dev);
				i2c_del_adapter(dev->i2c_adapter);
				continue;
			}
		}

		if( 0 != ext->init_done(dev) ) {
			printk("saa7146_v4l2.o: init failed. continuing anyway.\n");
		}

		j++;

		dev->capability.inputs = ext->inputs;
		dev->capability.audios = ext->audios;
		dev->capability.flags |= ext->flags;
		strcpy(dev->capability.name, ext->name);
		dev->ext = ext;
	}
	
	return 0;
}

int saa7146_v4l2_unregister_extension(struct saa7146_v4l2_extension* ext)
{
	int i = 0;
	struct capture_device* dev = 0;

	DEB_EE(("saa7146_v4l2.o: saa7146_v4l2_unregister_extension called.\n"));

	for(i = 0; i < saa7146_v4l2_num; i++) {

		dev = &saa7146_v4l2_devices[i];

		/* check if handled by this extension */
		if( ext != dev->ext ) {
			continue;
		}

		if( 0 != ext->detach(dev) ) {
			printk("saa7146_v4l2.o: release failed. ignoring.\n");
		}

		i2c_del_adapter(dev->i2c_adapter);
		v4l2_unregister_device(&dev->v4l2);
		if( NULL != dev->ext->unit_vbi && dev->revision != 0) {
			v4l2_unregister_device(&dev->vbi);
		}
		dev->ext = NULL;
	}
	
	return 0;
}

/*********************************************************************************/
/* configuration-functions                                                       */

int config_a_device(struct pci_dev *pci, struct capture_device* dev)
{
	unsigned long adr = 0, len = 0;
	DEB_EE(("saa7146_v4l2.o: config_a_device called.\n"));

	if (pci_enable_device(pci))
		return -EIO;

	/* clear out mem for sure */
	memset(dev, 0x0, sizeof(struct capture_device));

	/* todo: check module-parameters for sanity for this device*/

	dev->pci_device = pci;

	/* get chip-revision; this is needed to enable bug-fixes */
	if( 0 > pci_read_config_dword(dev->pci_device, PCI_CLASS_REVISION, &dev->revision)) {
		printk("saa7146_v4l2.o: cannot read from pci-device!\n");
		return -ENODEV;
	}
	dev->revision &= 0xf;

	/* remap the memory from virtual to physical adress */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
	adr = pci->base_address[0];
	len = 0x1000;
#else
	adr = pci_resource_start(pci,0);
	len = pci_resource_len(pci,0);

	if (!request_mem_region(pci_resource_start(pci,0), pci_resource_len(pci,0), "saa7146_v4l2")) {
	    	printk("saa7146_v4l2.o: failed to request mem region.\n");
		return -ENODEV;
	}
#endif
	dev->mem=ioremap(adr,len);
	if ( 0 == dev->mem ) {
	    	printk("saa7146_v4l2.o: failed to remap pci-address.\n");
	        release_mem_region(adr,len);
		return -ENODEV;
	}

	/* we don't do a master reset here anymore, it screws up
	   some boards that don't have an i2c-eeprom for configuration
	   values */
/*
	saa7146_write(dev, MC1, MASK_31);
*/

//	DEB_D(("===");dump_registers(dev));
	
	/* shut down all dma transfers */
	saa7146_write(dev, MC1, 0x00ff0000);

	/* clear out any rps-signals pending */
	saa7146_write(dev, MC2, 0xf8000000);
		
	/* disable alle irqs, clear irq-mask */
	saa7146_write(dev, ISR, 0xffffffff);
	dev->irq_left = 0;
	dev->irq_mask = 0;
	saa7146_write(dev, IER, dev->irq_mask);

	/* get memory for various stuff */
	dev->rps0 = (u32*)kmalloc(SAA7146_RPS_MEM, GFP_KERNEL);
	if( NULL == dev->rps0 )
		goto kmalloc_error_1;
	memset(dev->rps0, 0x0, SAA7146_RPS_MEM);	

	dev->rps1 = (u32*)kmalloc(SAA7146_RPS_MEM, GFP_KERNEL);
	if( NULL == dev->rps1 )
		goto kmalloc_error_2;
	memset(dev->rps1, 0x0, SAA7146_RPS_MEM);	

	dev->debi = (u32*)kmalloc(SAA7146_DEBI_MEM, GFP_KERNEL);
	if( NULL == dev->debi )
		goto kmalloc_error_3;
	memset(dev->debi, 0x0, SAA7146_DEBI_MEM);	

	dev->clipping = (u32*)kmalloc(SAA7146_CLIPPING_MEM, GFP_KERNEL);
	if( NULL == dev->clipping )
		goto kmalloc_error_4;
	memset(dev->clipping, 0x0, SAA7146_CLIPPING_MEM);	

	dev->i2c_adapter = kmalloc(sizeof(struct i2c_adapter), GFP_KERNEL);
	if ( NULL == dev->i2c_adapter )
		goto kmalloc_error_5;
	memset(dev->i2c_adapter, 0x0, sizeof(struct i2c_adapter));	

	dev->vbi_mem[0] = (u32*)kmalloc(SAA7146_VBI_MEM, GFP_KERNEL);
	if( NULL == dev->vbi_mem[0] )
		goto kmalloc_error_6;
	memset(dev->vbi_mem[0], 0x00, SAA7146_VBI_MEM);	

	dev->vbi_mem[1] = (u32*)kmalloc(SAA7146_VBI_MEM, GFP_KERNEL);
	if( NULL == dev->vbi_mem[1] )
		goto kmalloc_error_7;
	memset(dev->vbi_mem[1], 0x00, SAA7146_VBI_MEM);	

	dev->i2c_mem = (u32*)kmalloc(SAA7146_I2C_MEM, GFP_KERNEL);
	if( NULL == dev->i2c_mem )
		goto kmalloc_error_8;
	memset(dev->i2c_mem, 0x00, SAA7146_I2C_MEM);	
		
	/* i2c-adapter preparations */
	dev->i2c_bitrate = SAA7146_I2C_BUS_BIT_RATE_480;
	saa7146_i2c_reset(dev);

	/* create a nice little name */
	sprintf(&dev->name[0], "saa7146(%d)",saa7146_v4l2_num);
		
/*
	printk("IICSTA: 0x%08x\n",saa7146_read(dev,0x90));
	printk("IICCTR: 0x%08x\n",saa7146_read(dev,0x8c));
	printk("MC1:    0x%08x\n",saa7146_read(dev,MC1));
	printk("MC2:    0x%08x\n",saa7146_read(dev,MC2));
*/
	memset(dev->i2c_adapter,0,sizeof(struct i2c_adapter));
	sprintf(dev->i2c_adapter->name, "saa7146(%d)",saa7146_v4l2_num);
	dev->i2c_adapter->data		= dev;
	dev->i2c_adapter->algo		= &saa7146_algo;
	dev->i2c_adapter->algo_data	= NULL;
	dev->i2c_adapter->id		= I2C_HW_S_SAA7146;
	
	dev->i2c_adapter->inc_use		= saa7146_i2c_inc_use;
	dev->i2c_adapter->dec_use		= saa7146_i2c_dec_use;
	dev->i2c_adapter->client_register	= saa7146_i2c_reg;
	dev->i2c_adapter->client_unregister	= saa7146_i2c_unreg;

	dev->i2c_adapter->timeout = SAA7146_I2C_TIMEOUT;
	dev->i2c_adapter->retries = SAA7146_I2C_RETRIES;

	/* enable i2c-port pins, video-port-pins */
	saa7146_write(dev, MC1, (MASK_08 | MASK_24 | MASK_10 | MASK_26));

	/* the rest + print status message */
    	printk("saa7146_v4l2.o: found saa7146 @ mem 0x%08x (rev. %d, irq %d).\n", (unsigned int)dev->mem, dev->revision,dev->pci_device->irq);
	dev->ext = NULL;

	/* enable bus-mastering */
 	pci_set_master(pci);
		
	/* create unique serial number */
	dev->serial = saa7146_v4l2_num;	
	return 0;
	
kmalloc_error_8:
	kfree( dev->vbi_mem[1] );
kmalloc_error_7:
	kfree( dev->vbi_mem[0] );
kmalloc_error_6:
	kfree( dev->i2c_adapter );
kmalloc_error_5:
	kfree( dev->clipping );
kmalloc_error_4:
	kfree( dev->debi );
kmalloc_error_3:
	kfree( dev->rps1 );
kmalloc_error_2:
	kfree( dev->rps0 );
kmalloc_error_1:
	printk("saa7146_v4l2.o: not enough kernel-memory.\n");

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
	/* undo any stuff already done */
	iounmap((void*)(((u32)dev->mem)&0xfffff000));
#else
	iounmap(dev->mem);
        release_mem_region(adr,len);
#endif
	return -ENOMEM;
}


void unconfig_a_device(struct capture_device* dev)
{
	DEB_EE(("saa7146_v4l2.o: unconfig_a_device called.\n"));

//	DEB_D(("===");dump_registers(dev));
	
	/* free kernel memory */	
	kfree(dev->rps0 );
	kfree(dev->rps1 );
	kfree(dev->debi );
	kfree(dev->clipping);
	kfree(dev->vbi_mem[0]);
	kfree(dev->vbi_mem[1]);
	kfree(dev->i2c_mem);

	/* unmap the memory, if necessary */
	if (0 != dev->mem) {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
		iounmap((void*)(((u32)dev->mem)&0xfffff000));
#else
		iounmap(dev->mem);
		release_mem_region(pci_resource_start(dev->pci_device,0), pci_resource_len(dev->pci_device,0));
#endif
	}

	kfree(dev->i2c_adapter);
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
int __init saa7146_v4l2_init(void)
{
	struct pci_dev *dev = NULL;

	DEB_EE(("saa7146_v4l2.o: saa7146_v4l2_init called.\n"));

	/* first check for a pci bios */
	if ( 0 == pci_present()) {
		printk("saa7146_v4l2.o: PCI-BIOS not present or not accessable!\n");
		return -1;
	}

	/* loop through all pci-devices */
        while ( 0 != (dev = pci_find_device(PCI_VENDOR_ID_PHILIPS, PCI_DEVICE_ID_PHILIPS_SAA7146,dev))) {

		/* check for maximum number of saa7146-devices */
		if ( saa7146_v4l2_num > MAX_SAA7146_BOARDS ) {
			printk("saa7146_v4l2.o: maximum number of saa7146-devices reached.\n");
			continue;
		}
		
		/* ok, found one saa7146 */		
		if( 0 == config_a_device(dev, &saa7146_v4l2_devices[saa7146_v4l2_num]) ) {
			saa7146_v4l2_num++;
		}
	}

	if( 0 != saa7146_v4l2_num ) {
		printk("saa7146_v4l2.o: %d saa7146 chipset(s) found.\n", saa7146_v4l2_num);
		return 0;
	}
	
	printk("saa7146_v4l2.o: no saa7146 chipset(s) found.\n");
	return -1;
}

#ifdef MODULE
int init_module(void) 
{
	return saa7146_v4l2_init();
}

void cleanup_module(void) 
{
	int i = 0;
	
	DEB_EE(("saa7146_v4l2.o: cleanup_module called.\n"));

	/* release all saa7146s */
	for ( i = 0; i < saa7146_v4l2_num; i++) {
		unconfig_a_device(&saa7146_v4l2_devices[i]);
	}
}
#endif

#else
int saa7146_v4l2_suspend(struct pci_dev *pdev, u32 state)
{
	DEB_EE(("saa7146_v4l2.o: saa7146_v4l2_suspend called.\n"));

	/* private data lies in pdev->driver_data, call ioctl to handle
	   suspend wish, call extension function, too! */
	
	return 0;
}

int saa7146_v4l2_resume(struct pci_dev *pdev)
{
	DEB_EE(("saa7146_v4l2.o: saa7146_v4l2_resume called.\n"));

	/* private data lies in pdev->driver_data, call ioctl to handle
	   resume wish, call extension function, too! */

	return 0;
}

int __devinit saa7146_v4l2_init_one(struct pci_dev *pdev, const struct pci_device_id *ent)
{
        int result = 0;

	DEB_EE(("saa7146_v4l2.o: saa7146_v4l2_init_one called.\n"));

	pdev->driver_data = &saa7146_v4l2_devices[saa7146_v4l2_num];
	result = config_a_device(pdev, pdev->driver_data);
	saa7146_v4l2_num++;
	
	return 0;
}

void __devexit saa7146_v4l2_remove_one(struct pci_dev *pdev)
{
	DEB_EE(("saa7146_v4l2.o: saa7146_v4l2_remove_one called.\n"));

 	unconfig_a_device(pdev->driver_data);

}

struct pci_device_id saa7146_v4l2_pci_tbl[] __devinitdata = {
	{ PCI_VENDOR_ID_PHILIPS, PCI_DEVICE_ID_PHILIPS_SAA7146,
		PCI_ANY_ID, PCI_ANY_ID, },
	{ 0,},
};

MODULE_DEVICE_TABLE(pci, saa7146_v4l2_pci_tbl);
	
struct pci_driver saa7146_v4l2_driver = {
	name:		"saa7146 v4l2",
	id_table:	saa7146_v4l2_pci_tbl,
	probe:		saa7146_v4l2_init_one,
	remove:		saa7146_v4l2_remove_one,
	suspend:	saa7146_v4l2_suspend,
	resume:		saa7146_v4l2_resume,
};


#ifdef MODULE
int __init saa7146_v4l2_init_module(void)
{
	return pci_module_init(&saa7146_v4l2_driver);
}

void __exit saa7146_v4l2_cleanup_module(void)
{
	pci_unregister_driver(&saa7146_v4l2_driver);
}

module_init(saa7146_v4l2_init_module);
module_exit(saa7146_v4l2_cleanup_module);
#endif
#endif

#ifdef MODULE
MODULE_AUTHOR("Michael Hunold <michael@mihu.de>");
MODULE_DESCRIPTION("v4l2 core-driver for saa7146-based video-hardware");

EXPORT_SYMBOL(saa7146_v4l2_register_extension);
EXPORT_SYMBOL(saa7146_v4l2_unregister_extension);

EXPORT_SYMBOL(saa7146_set_overlay);
EXPORT_SYMBOL(saa7146_set_hps_source_and_sync);
EXPORT_SYMBOL(saa7146_enumstd);
EXPORT_SYMBOL(saa7146_s_std);
#endif

#ifdef MODULE
MODULE_LICENSE("GPL");
#endif
