#ifndef __SAA7146_V4L2__
#define __SAA7146_V4L2__

#include <linux/i2c.h>

#include "videodevX/videodev.h"

#include "saa7146.h"

#define MAX_SAA7146_BOARDS		6

#define MAX_SAA7146_OPENS		5
#define MAX_SAA7146_CAPTURE_BUFFERS	64

struct capture_device;

struct saa7146_v4l2_extension_ioctls
{
	unsigned int	cmd;
	int		flags;	
};

struct saa7146_v4l2_extension
{
	char	name[64];		/* name of the device */
	int	inputs;			
	int	audios;			
	u32	flags;

	int*	unit_video;
	int*	unit_vbi;
	
	struct saa7146_v4l2_extension_ioctls* ioctls;
	
	u32	irq_mask;	/* mask to indicate, which irq-events are handled by the extension */
	void 	(*irq_func)(struct capture_device*, u32* irq_mask);
	
	/* extension functions */
	int (*preinit)(struct capture_device*);
	int (*probe)(struct capture_device*);
	int (*init_done)(struct capture_device*);

	int (*attach)(struct capture_device*);
	int (*detach)(struct capture_device*);

	void (*v4l2_open)(struct capture_device*);
	void (*v4l2_close)(struct capture_device*);

	void (*vbi_open)(struct capture_device*);
	int  (*vbi_bypass)(struct capture_device*);
	void (*vbi_close)(struct capture_device*);

	int (*ioctl)(void *id, unsigned int cmd, void *arg);
};

#define SAA7146_EXCLUSIVE	0x1
#define SAA7146_BEFORE		0x2
#define SAA7146_AFTER		0x4

int saa7146_v4l2_register_extension(struct saa7146_v4l2_extension*);
int saa7146_v4l2_unregister_extension(struct saa7146_v4l2_extension*);

#include <asm/io.h> /* for accessing devices */
#define saa7146_write(saa,adr,dat)    writel((dat),(saa->mem+(adr)))
#define saa7146_read(saa,adr)         readl(saa->mem+(adr))

struct saa7146_modes_constants
{
	int v_offset;
	int v_field;
	int v_calc;
	
	int h_offset;
	int h_pixels;
	int h_calc;
	
	int v_max_out;
	int h_max_out;
};

/*  Per-open data for handling multiple opens on one device */
struct device_open
{
	int			isopen;
	int			noncapturing;
	struct capture_device	*dev;
};

/*  Streaming data buffer  */
struct stream_buffer
{
	struct v4l2_q_node	qnode;
	struct v4l2_buffer	vidbuf;
	int			requested;
	
	u8*	vaddress;  /* vmalloc() */
	u32*	page_table;
};

struct capture_device
{
	struct v4l2_device		v4l2;			/*  Must be first */
	struct v4l2_device		vbi;

	int				serial;		

	unsigned char*			mem;			/* pointer to mapped IO memory */
	int				revision;		/* chip revision; needed for bug-workarounds*/
	
	struct v4l2_capability		capability;
	
	/* vbi stuff */
	int				vbi_in_use;
	long				vbi_last_read;
		
	int				vbi_pixel_to_capture;
	int				vbi_depth;
	int				vbi_start;
	int				vbi_count;

	long				vbi_counter;
	int				vbi_current;
	u32*				vbi_mem[2];
#if LINUX_VERSION_CODE < 0x020300
        struct wait_queue		*vbi_wq;		/* vbi interrupt queue */
#else
        wait_queue_head_t		vbi_wq;			/* vbi interrupt queue */
#endif
	
	/* pci-device & irq stuff*/
	struct pci_dev			*pci_device;
	char				name[32];
	u32				irq_mask;		/* mask which shows which irqs are handled */
	u32				irq_left;		/* shows which irqs are undhandled */
	struct tq_struct		bh_handler;		/* for bottom-half irq-handler */
	
	/* i2c-stuff */
	struct i2c_adapter		*i2c_adapter;	
	u32				i2c_bitrate;
	u32*				i2c_mem;		/* pointer to i2c memory */
	
	/* memories */
	u32*				clipping;		/* pointer to clipping memory */
	u32*				rps0;
	u32*				rps1;
	u32*				debi;
	
	struct saa7146_v4l2_extension*	ext;		/* indicates if handled by extension */
	void*				ext_priv;	/* pointer for extension private use (most likely some private data) */
		
	int			open_count;
	struct device_open	open_data[MAX_SAA7146_OPENS];
	int			capturing_opens;

	struct v4l2_performance perf;	/* performance statistics */
	struct v4l2_framebuffer	fb;	/* framebuffer settings */
	struct v4l2_standard	standard;
	struct v4l2_format 	format;

/*	Client capture format and capture modes	*/
	struct v4l2_pix_format	clientfmt;
	struct v4l2_captureparm	capture;
	int			input;

	int	vflip;
	int 	hflip;

/*	Capture state (read)	*/
	struct stream_buffer	read_buffer;

/*	Capture state (streaming)	*/
	int			streaming;

	struct stream_buffer	stream_buf[MAX_SAA7146_CAPTURE_BUFFERS];
	int			stream_buffers_requested;
	int			stream_buffers_mapped;
	int			stream_buffers_planar;
	struct v4l2_queue	stream_q_capture;
	struct v4l2_queue	stream_q_done;
	stamp_t			stream_begin;
	long			stream_last_frame;
	
	u32			stream_flags;
	u32			stream_format;
	
	int			stream_contig_map;	// for v4l backward compatibility

	u32*			stream_db_adr[2];
	struct stream_buffer*	stream_db_cur[2];	
	int			stream_db_capure_mask;
	
	u32			stream_rev0_data;

#if LINUX_VERSION_CODE < 0x020300
        struct wait_queue *capture_wq;                /* rps0 interrupt queue (=> capture) */
#else
        wait_queue_head_t capture_wq;                /* rps0 interrupt queue (=> capture) */
#endif
	
	struct	saa7146_modes_constants	modes_constants;
	
	int 	current_hps_source;
	int 	current_hps_sync;

	int	overlay;

	int	overlay_data_available;
	int 	overlay_data_valid;

	int	overlay_width;
	int	overlay_height;
	int	overlay_x;
	int	overlay_y;
	int	overlay_palette;
	
	int*	clip_data;
	int	clip_count;

};

int saa7146_s_std(struct device_open *o, struct v4l2_standard *std);
int saa7146_enumstd(struct capture_device* dev, struct v4l2_enumstd *e);

void saa7146_set_overlay(struct capture_device*, int);
void saa7146_set_hps_source_and_sync(struct capture_device*, int, int);
#endif
