/*
    dpc_v4l2.c - v4l2 driver for dpc7146 demonstration board
    
    Copyright (C) 2000 Michael Hunold <michael@mihu.de>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <linux/module.h>	/* for module-version */
#include <linux/slab.h>		/* for kmalloc/kfree */
#include <linux/init.h>		/* for "__init" */
#include <linux/delay.h>	/* for mdelay */
#include <linux/poll.h>
#include <linux/i2c.h>

#include "videodevX/videodev.h"

#include "core/saa7146_v4l2.h"

#include "i2c_modules/saa7111a.h"

/* global defines */
#define DEB1(x) if (debug>=1) x
#define DEB2(x) if (debug>=2) x

/* module parameters */
int debug = 2;

/* ignore card */
int ignore[MAX_SAA7146_BOARDS] = { 0, };

/* minor device-node to be used */
int unit_video[MAX_SAA7146_BOARDS] = { 0, 1, 2, 3, };
int unit_vbi[MAX_SAA7146_BOARDS] = { 224, 225, 226, 227, };

/* MAX_SAA7146_BOARDS gets defined in saa7146.h */

#ifdef MODULE
MODULE_PARM(debug,"i");
MODULE_PARM_DESC(debug, "debug level - 0: off; 1: normal; 2: verbose");

MODULE_PARM(ignore, "1-"__MODULE_STRING(MAX_SAA7146_BOARDS)"i");
MODULE_PARM_DESC(ignore, "exclude device from being recognized");

MODULE_PARM(unit_video, "1-"__MODULE_STRING(MAX_SAA7146_BOARDS)"i");
MODULE_PARM_DESC(unit_video, "minor device-node(s) to be used");
#endif

/* global variables */
int dpc_v4l2_num = 0;

#define DPC_INPUTS	2
struct v4l2_input dpc_inputs[DPC_INPUTS] = {
	{ 0, "Port 0",	V4L2_INPUT_TYPE_CAMERA,	V4L2_INPUT_CAP_AUDIO, 1, },
	{ 1, "Port 1",	V4L2_INPUT_TYPE_CAMERA,	V4L2_INPUT_CAP_AUDIO, 1, }
};

#define DPC_AUDIOS	0

struct saa7146_v4l2_extension_ioctls ioctls[] = {
	{ VIDIOC_G_INPUT,	SAA7146_EXCLUSIVE },
	{ VIDIOC_S_INPUT,	SAA7146_EXCLUSIVE },
	{ VIDIOC_ENUMINPUT, 	SAA7146_EXCLUSIVE },
	{ VIDIOC_S_STD,		SAA7146_AFTER },
	{ 0,			0 }
};

struct dpc_v4l2
{
	struct i2c_client*	saa7111a;
	
	int cur_input;	/* current input */
};

/* this function gets called very early in the registration-process of
   the extension. it has been reported that some devices need to enable
   the i2c-bus explicitly for example -- this can be done here. please
   note, that you cannot be sure that the device is really the hardware
   you expect, so you should do as little as possible in here, in order
   to avoid confusing the hardware. */
int	dpc_v4l2_preinit(struct capture_device* dev)
{
	DEB2(printk("dpc_v4l2.o: dpc_v4l2_pre_init called.\n"));

	/* first check if this device is to be ignored
	   by using serial identifier */
	if( 1 == ignore[ dev->serial ]) {
		printk("dpc_v4l2.o: note: device ignored due to insmod parameter.\n");	
		return -ENODEV;
	}

	return 0;
}

int	dpc_v4l2_probe(struct capture_device* dev)
{
	struct dpc_v4l2* dpc = 0;	

	int i = 0;

	DEB2(printk("dpc_v4l2.o: dpc_v4l2_probe called.\n"));

	dpc = (struct dpc_v4l2*)kmalloc(sizeof(struct dpc_v4l2), GFP_KERNEL);
	if( NULL == dpc ) {
		printk("dpc_v4l2.o: dpc_v4l2_attach: not enough kernel memory.\n");
		return -ENOMEM;
	}
	memset(dpc, 0x0, sizeof(struct dpc_v4l2));	

	/* loop through all i2c-devices on the bus and look who is there */
	for(i = 0; i < dev->i2c_adapter->client_count; i++) {
		if( I2C_SAA7111A == dev->i2c_adapter->clients[i]->addr ) 
			dpc->saa7111a = dev->i2c_adapter->clients[i];
	}

	/* check if all devices are present */
	if( 0 == dpc->saa7111a ) {
		DEB2(printk("dpc_v4l2.o: dpc_v4l2_attach failed for this device.\n"));	
		kfree(dpc);
		return -ENODEV;
	}
	
	/* all devices are present, probe was successful */	
	DEB2(printk("dpc_v4l2.o: dpc_v4l2_probe succeeded for this device.\n"));	

	/* we store the pointer in our private data field */
	(struct dpc_v4l2*)dev->ext_priv = dpc;

	return 0;
}

/* bring hardware to a sane state. this has to be done, just in case someone
   wants to capture from this device before it has been properly initialized.
   the capture engine would badly fail, because no valid signal arrives on the
   saa7146, thus leading to timeouts and stuff. */
int dpc_v4l2_init_done(struct capture_device* dev)
{
	struct dpc_v4l2* dpc = (struct dpc_v4l2*)dev->ext_priv;

	int i = 0;

	DEB2(printk("dpc_v4l2.o: dpc_v4l2_init_done called.\n"));

	/* initialize the helper ics to useful values */
	i = 0;
	if ( 0 != dpc->saa7111a->driver->command(dpc->saa7111a,SAA711X_SET_MODE, &i)) {
		return -EFAULT;
	}

	i  = 0;
/*	i  = SAA711X_USE_ANTIALIAS;
	i |= SAA711X_USE_AMPLIFIER; 
*/
	if ( 0 !=  dpc->saa7111a->driver->command(dpc->saa7111a,SAA711X_SET_FUSE, &i)) {
		return -EFAULT;
	}
		
	i = 3;	
	if ( 0 != dpc->saa7111a->driver->command(dpc->saa7111a,SAA711X_SET_UPDATE_GAIN, &i)) {
		return -EFAULT;
	}
	
	i = SAA711X_VIDEO_MODE_PAL;
	if ( 0 != dpc->saa7111a->driver->command(dpc->saa7111a,SAA711X_SET_NORM, &i)) {
		return -EFAULT;
	}
	
	i2c_smbus_write_byte_data(dpc->saa7111a, 0x00, 0x11);

	i2c_smbus_write_byte_data(dpc->saa7111a, 0x02, 0xc0);
	i2c_smbus_write_byte_data(dpc->saa7111a, 0x03, 0x30);
	i2c_smbus_write_byte_data(dpc->saa7111a, 0x04, 0x00);
	i2c_smbus_write_byte_data(dpc->saa7111a, 0x05, 0x00);
	i2c_smbus_write_byte_data(dpc->saa7111a, 0x06, 0xde);
	i2c_smbus_write_byte_data(dpc->saa7111a, 0x07, 0xad);
	i2c_smbus_write_byte_data(dpc->saa7111a, 0x08, 0xa8);
	i2c_smbus_write_byte_data(dpc->saa7111a, 0x09, 0x00);
	i2c_smbus_write_byte_data(dpc->saa7111a, 0x0a, 0x80);
	i2c_smbus_write_byte_data(dpc->saa7111a, 0x0b, 0x47);
	i2c_smbus_write_byte_data(dpc->saa7111a, 0x0c, 0x40);
	i2c_smbus_write_byte_data(dpc->saa7111a, 0x0d, 0x00);
	i2c_smbus_write_byte_data(dpc->saa7111a, 0x0e, 0x03);

	i2c_smbus_write_byte_data(dpc->saa7111a, 0x10, 0xd0);
	i2c_smbus_write_byte_data(dpc->saa7111a, 0x11, 0x1c);
	i2c_smbus_write_byte_data(dpc->saa7111a, 0x12, 0xc1);
	i2c_smbus_write_byte_data(dpc->saa7111a, 0x13, 0x30);

	i2c_smbus_write_byte_data(dpc->saa7111a, 0x1f, 0x81);
	
/*
	for(i = 0; i < 0x1f; i++) {
		printk("0x%02x: 0x%02x\n", i, i2c_smbus_read_byte_data(dpc->saa7111a,i));
	}
*/	
	return 0;
}

/* this function only gets called when the probing was successful */
int dpc_v4l2_attach(struct capture_device* dev)
{
	struct dpc_v4l2* dpc = (struct dpc_v4l2*)dev->ext_priv;
	
	DEB2(printk("dpc_v4l2.o: dpc_v4l2_attach called.\n"));

	/* checking for i2c-devices can be omitted here, because we
	   already did this in "dpc_vl42_probe" */

	dpc_v4l2_num++;
	
	i2c_inc_use_client(dpc->saa7111a);

	/* the rest */
	dpc->cur_input = 0;
	
	return 0;
}

int dpc_v4l2_detach(struct capture_device* dev)
{
	struct dpc_v4l2* dpc = (struct dpc_v4l2*)dev->ext_priv;
	
	DEB2(printk("dpc_v4l2.o: dpc_v4l2_detach called.\n"));

	i2c_dec_use_client(dpc->saa7111a);

	dpc_v4l2_num--;

	kfree(dpc);

	return 0;
}


void dpc_v4l2_open(struct capture_device* dev)
{
#ifdef MODULE
	MOD_INC_USE_COUNT;
#endif
}

void dpc_v4l2_close(struct capture_device* dev)
{
#ifdef MODULE
	MOD_DEC_USE_COUNT;
#endif
}

int dpc_v4l2_vbi_bypass(struct capture_device* dev)
{
	struct dpc_v4l2* dpc = (struct dpc_v4l2*)dev->ext_priv;
	
	int i = 1;

	/* switch bypass in saa7111a */
	if ( 0 != dpc->saa7111a->driver->command(dpc->saa7111a,SAA711X_VBI_BYPASS, &i)) {
		printk("dpc_v4l2.o: VBI_BYPASS: could not address saa7111a.\n");
		return -1;
	}			

	return 0;
}

int dpc_v4l2_ioctl(struct capture_device *dev, unsigned int cmd, void *arg) 
{
	struct dpc_v4l2* dpc = (struct dpc_v4l2*)dev->ext_priv;

	switch(cmd)
	{
	case VIDIOC_S_STD:
	{
		struct v4l2_standard	*std = arg;
		int			id  = 0, j = 0;

		id = v4l2_video_std_confirm(std);

		DEB2(printk("dpc_v4l2.o: VIDIOC_S_STD %d.\n",id));

		if( id != V4L2_STD_PAL && id != V4L2_STD_NTSC && id != V4L2_STD_SECAM ) {
			printk("dpc_v4l2.o: VIDIOC_S_STD: invalid video norm requested\n");
			return -EINVAL;
		}
		
		/* switch norm */	
		switch( id ) {
			case V4L2_STD_NTSC: {
				printk("dpc_v4l2.o: VIDIOC_S_STD: setting NTSC\n");
				j = SAA711X_VIDEO_MODE_NTSC;
				break;
			}
			case V4L2_STD_SECAM: {
				printk("dpc_v4l2.o: VIDIOC_S_STD: setting SECAM\n");
				j = SAA711X_VIDEO_MODE_SECAM;
				break;
			}
			default: /* case V4L2_STD_PAL: */ {
				printk("dpc_v4l2.o: VIDIOC_S_STD: setting PAL\n");
				j = SAA711X_VIDEO_MODE_PAL;
				break;
			}
		}
		
		if ( 0 != dpc->saa7111a->driver->command(dpc->saa7111a,SAA711X_SET_NORM, &j)) {
			printk("dpc_v4l2.o: VIDIOC_S_STD: could not address saa7111a.\n");
			return -EFAULT;
		}
		return 0;
	}	
	case VIDIOC_G_INPUT:
	{
		int *input = (int *)arg;
		*input = dpc->cur_input;

		DEB2(printk("dpc_v4l2.o: VIDIOC_G_INPUT: %d\n",*input));
		
		return 0;
	}
	case VIDIOC_S_INPUT:
	{
		int	input = *(int *)arg;

		if (input < 0 || input >= DPC_INPUTS) {
			return -EINVAL;
		}
	
		dpc->cur_input = input;

		/* fixme: switch input here, switch audio, too! */
		printk("dpc_v4l2.o: VIDIOC_S_INPUT: fixme switch input.\n");
		
		return 0;
	}
	case VIDIOC_ENUMINPUT:
	{
		struct v4l2_input *i = arg;
		
		/* sanity check to satisfy xawtv */
		if( i->index < 0 || i->index >= DPC_INPUTS) {
			return -EINVAL;
		}
		
		memcpy(i, &dpc_inputs[i->index], sizeof(struct v4l2_input));

		DEB2(printk("dpc_v4l2.o: v4l2_ioctl: VIDIOC_ENUMINPUT %d.\n",i->index));
		return 0;
	}
	default:
/*
		DEB2(printk("dpc_v4l2.o: v4l2_ioctl does not handle this ioctl.\n"));
*/
		return -ENOIOCTLCMD;
	}
	return 0;
}

struct saa7146_v4l2_extension extension = {
	"dpc7146 demonstration board",
	DPC_INPUTS,
	DPC_AUDIOS,
	0,			/* no flags */
	&unit_video[0],
	&unit_vbi[0],
	&ioctls[0],
	0,
	NULL,
	dpc_v4l2_preinit,
	dpc_v4l2_probe,
	dpc_v4l2_init_done,
	dpc_v4l2_attach,
	dpc_v4l2_detach,
	dpc_v4l2_open,
	dpc_v4l2_close,
	dpc_v4l2_open,		/* no error: nothing special to do here */
	dpc_v4l2_vbi_bypass,
	dpc_v4l2_close,		/* no error: nothing special to do here */
	dpc_v4l2_ioctl,
};

int __init dpc_v4l2_init(void)
{
	DEB2(printk("dpc_v4l2.o: dpc_v4l2_init called.\n"));
	
	if( 0 != saa7146_v4l2_register_extension(&extension)) {
		DEB1(printk("dpc_v4l2.o: failed to register extension.\n"));
		return -ENODEV;
	}

	/* dpc_v4l2_num gets increased if the probe function is called successfully */

	if( 0 == dpc_v4l2_num ) {
		printk(KERN_ERR "dpc_v4l2.o: no DPC(s) found.\n");
		return -ENODEV;
	}

	printk(KERN_ERR "dpc_v4l2.o: %d DPC(s) found.\n", dpc_v4l2_num);
	return 0;
}

#ifdef MODULE
MODULE_AUTHOR("Michael Hunold <michael@mihu.de>");
MODULE_DESCRIPTION("v4l2-driver for the dpc7146 demonstration board");

EXPORT_NO_SYMBOLS;

int init_module(void) 
{
	return dpc_v4l2_init();
}

void cleanup_module(void) 
{
	DEB2(printk("dpc_v4l2.o: cleanup_module called.\n"));
	saa7146_v4l2_unregister_extension(&extension);
}

#endif

#ifdef MODULE
MODULE_LICENSE("GPL");
#endif


