/*
    mxb_v4l2.c - v4l2 driver for Multimedia eXtension Board
    
    Copyright (C) 2000,2001 Michael Hunold <michael@mihu.de>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <linux/module.h>	/* for module-version */
#include <linux/slab.h>		/* for kmalloc/kfree */
#include <linux/init.h>		/* for "__init" */
#include <linux/delay.h>	/* for mdelay */
#include <linux/poll.h>
#include <linux/i2c.h>

#include <linux/sound.h>
#include <linux/soundcard.h>

#include "core/saa7146_v4l2.h"

#include "i2c_modules/saa7111a.h"
#include "i2c_modules/tea6415c.h"
#include "i2c_modules/tea6420.h"
#include "i2c_modules/tda9840.h"
#include "i2c_modules/fi1216mk2.h"

#include "mxb_v4l2.h"

/* global defines */
#define DEB1(x) if (debug>=1) x
#define DEB2(x) if (debug>=2) x

/* module parameters
 
 example: modprobe mxb_v4l2 debug=2 ignore=1,0
 
 means: insert module with verbosity 2, ignore the first
 saa7146 in the system, but try to detect mxb on any other. 
*/

#ifdef SAA7146_AXA
/* tells us to provide a simple stupid mixer device. this device
   can adjust the volume by changing the gain on the two tea6420
   chipsets. this does not provide a huge range, though, and is 
   pretty useless */
int mixer = 0;
#endif

/* initial frequence the tuner will be tuned to. 
   in verden (lower saxony, germany) 4148 is a
   channel called "phoenix" */
int freq = 4148;

/* debug verbosity */
int debug = 0;

/* ignore card */
int ignore[MAX_SAA7146_BOARDS] = { 0, };

/* minor device-node to be used */
int unit_video[MAX_SAA7146_BOARDS] = { 0, 1, 2, 3, };
int unit_vbi[MAX_SAA7146_BOARDS] = { 224, 225, 226, 227, };

/* MAX_SAA7146_BOARDS gets defined in saa7146.h */

#ifdef MODULE

MODULE_PARM(freq,"i");
MODULE_PARM_DESC(freq, "initial frequency the tuner will be tuned to while setup");

MODULE_PARM(debug,"i");
MODULE_PARM_DESC(debug, "debug level - 0: off; 1: normal; 2: verbose");

MODULE_PARM(ignore, "1-"__MODULE_STRING(MAX_SAA7146_BOARDS)"i");
MODULE_PARM_DESC(ignore, "exclude device from being recognized");

MODULE_PARM(unit_video, "1-"__MODULE_STRING(MAX_SAA7146_BOARDS)"i");
MODULE_PARM_DESC(unit_video, "minor device-node(s) to be used");

MODULE_PARM(unit_vbi, "1-"__MODULE_STRING(MAX_SAA7146_BOARDS)"i");
MODULE_PARM_DESC(unit_vbi, "minor device-node(s) to be used (for vbi)");

#ifdef SAA7146_AXA
MODULE_PARM(mixer,"i");
MODULE_PARM_DESC(mixer, "tells us to provide a simple stupid mixer device");
#endif

#endif

/* global variable */
int mxb_v4l2_num = 0;

#define MXB_INPUTS	6
struct v4l2_input mxb_inputs[MXB_INPUTS] = {
	{ 0, "Tuner",	V4L2_INPUT_TYPE_TUNER,	V4L2_INPUT_CAP_AUDIO, 0, },
	{ 1, "CAM Y/C",	V4L2_INPUT_TYPE_CAMERA,	V4L2_INPUT_CAP_AUDIO, 1, },
	{ 2, "CAM",	V4L2_INPUT_TYPE_CAMERA,	V4L2_INPUT_CAP_AUDIO, 1, },
	{ 3, "AUX3 Y/C",V4L2_INPUT_TYPE_CAMERA,	V4L2_INPUT_CAP_AUDIO, 4, },
	{ 4, "AUX3",	V4L2_INPUT_TYPE_CAMERA,	V4L2_INPUT_CAP_AUDIO, 4, },
	{ 5, "AUX1",	V4L2_INPUT_TYPE_CAMERA,	V4L2_INPUT_CAP_AUDIO, 2, }
};

enum { TUNER, CAM_YC, CAM, AUX3_YC, AUX3, AUX1 };

/* this array holds the information, which port of the saa7146 each
   input actually uses. the mxb uses port 0 for every input */
struct {
	int hps_source;
	int hps_sync;
} input_port_selection[MXB_INPUTS] = { 	
	{ SAA7146_HPS_SOURCE_PORT_A, SAA7146_HPS_SYNC_PORT_A },
	{ SAA7146_HPS_SOURCE_PORT_A, SAA7146_HPS_SYNC_PORT_A },
	{ SAA7146_HPS_SOURCE_PORT_A, SAA7146_HPS_SYNC_PORT_A },
	{ SAA7146_HPS_SOURCE_PORT_A, SAA7146_HPS_SYNC_PORT_A },
	{ SAA7146_HPS_SOURCE_PORT_A, SAA7146_HPS_SYNC_PORT_A },
	{ SAA7146_HPS_SOURCE_PORT_A, SAA7146_HPS_SYNC_PORT_A }
};

/* this array holds the information of the audio source (mxb_audios),
   which has to be switched corresponding to the video source (mxb_channels) */
int video_audio_connect[MXB_AUDIOS] =
	{ 0, 1, 1, 4, 4, 2 };

/* these are the necessary input-output-pins for bringing one audio source
(see above) to the CD-output */
struct audio_multiplex TEA6420_cd[9][2] =
		{ 
		{{1,1,0},{1,1,0}},	/* Tuner */
		{{1,1,0},{4,1,0}},	/* Camera */
/*		{{1,1,0},{5,1,0}}, */	/* MPEG */
		{{5,1,0},{6,1,0}},	/* AUX 1 */
		{{4,1,0},{6,1,0}},	/* AUX 2 */
		{{3,1,0},{6,1,0}},	/* AUX 3 */
		{{1,1,0},{3,1,0}},	/* Radio */
		{{1,1,0},{2,1,0}},	/* CD-Rom */
		{{6,1,0},{6,1,0}}	/* Mute */
		};

/* these are the necessary input-output-pins for bringing one audio source
(see above) to the line-output */
struct audio_multiplex TEA6420_line[9][2] =
		{
		{{2,3,0},{1,2,0}},
		{{2,3,0},{4,2,0}},
/*		{{2,3,0},{5,2,0}}, */ /* MPEG */
		{{5,3,0},{6,2,0}},
		{{4,3,0},{6,2,0}},
		{{3,3,0},{6,2,0}},
		{{2,3,0},{3,2,0}},
		{{2,3,0},{2,2,0}},
		{{6,3,0},{6,2,0}} /* Mute */
		};

#define MAXCONTROLS	1
struct v4l2_queryctrl mxb_controls[] = {
 { V4L2_CID_AUDIO_MUTE, "Mute", 0, 1, 1, 0, V4L2_CTRL_TYPE_BOOLEAN, 0, V4L2_CTRL_CAT_AUDIO, "audio"},
};

struct saa7146_v4l2_extension_ioctls ioctls[] = {
	{ VIDIOC_G_FREQ,	SAA7146_EXCLUSIVE },
	{ VIDIOC_S_FREQ, 	SAA7146_EXCLUSIVE },
	{ VIDIOC_ENUMINPUT, 	SAA7146_EXCLUSIVE },
	{ VIDIOC_G_INPUT,	SAA7146_EXCLUSIVE },
	{ VIDIOC_S_INPUT,	SAA7146_EXCLUSIVE },
	{ VIDIOC_G_AUDIO, 	SAA7146_EXCLUSIVE },
	{ VIDIOC_S_AUDIO, 	SAA7146_EXCLUSIVE },
	{ VIDIOC_G_TUNER, 	SAA7146_EXCLUSIVE },
	{ VIDIOC_S_TUNER, 	SAA7146_EXCLUSIVE },
	{ VIDIOC_QUERYCTRL, 	SAA7146_BEFORE },
	{ VIDIOC_G_CTRL,	SAA7146_BEFORE },
	{ VIDIOC_S_CTRL,	SAA7146_BEFORE },
	{ VIDIOC_S_STD,		SAA7146_EXCLUSIVE },
	{ VIDIOC_ENUMSTD,	SAA7146_EXCLUSIVE },
	{ MXB_S_AUDIO_CD, 	SAA7146_EXCLUSIVE },	/* custom control */	
	{ MXB_S_AUDIO_LINE, 	SAA7146_EXCLUSIVE },	/* custom control */	
	{ 0,			0 }
};

struct mxb_v4l2
{
	struct i2c_client*	saa7111a;
	struct i2c_client*	tda9840;
	struct i2c_client*	tea6415c;
	struct i2c_client*	fi1216mk2;
	struct i2c_client*	tea6420_1;
	struct i2c_client*	tea6420_2;

	int	cur_mode;	/* current audio mode (mono, stereo, ...) */
	int	cur_input;	/* current input */
	int	cur_freq;	/* current frequency the tuner is tuned to */
	int	cur_mute;	/* current mute status */
	int	cur_gain;	/* current gain for simple volume adjustment */
};

/* this function gets called very early in the registration-process of
   the extension. it has been reported that some devices need to enable
   the i2c-bus explicitly for example -- this can be done here. please
   note, that you cannot be sure that the device is really the hardware
   you expect, so you should do as little as possible in here, in order
   to avoid confusing the hardware.
*/
int mxb_v4l2_preinit(struct capture_device* dev)
{
	DEB2(printk("mxb_v4l2.o: mxb_v4l2_pre_init called.\n"));

	/* first check if this device is to be ignored
	   by using serial identifier */
	if( 1 == ignore[ dev->serial ]) {
		printk("mxb_v4l2.o: note: device ignored due to insmod parameter.\n");	
		return -ENODEV;
	}

	return 0;
}

int mxb_v4l2_probe(struct capture_device* dev)
{
	struct mxb_v4l2* mxb = 0;

	int i = 0;	

	DEB2(printk("mxb_v4l2.o: mxb_v4l2_probe called.\n"));

	mxb = (struct mxb_v4l2*)kmalloc(sizeof(struct mxb_v4l2), GFP_KERNEL);
	if( NULL == mxb ) {
		printk("mxb_v4l2.o: mxb_v4l2_probe: not enough kernel memory.\n");
		return -ENOMEM;
	}
	memset(mxb, 0x0, sizeof(struct mxb_v4l2));	

	/* loop through all i2c-devices on the bus and look who is there */
	for(i = 0; i < dev->i2c_adapter->client_count; i++) {
		if( I2C_TEA6420_1 == dev->i2c_adapter->clients[i]->addr )
			mxb->tea6420_1 = dev->i2c_adapter->clients[i];
		if( I2C_TEA6420_2 == dev->i2c_adapter->clients[i]->addr ) 
			mxb->tea6420_2 = dev->i2c_adapter->clients[i];
		if( I2C_TEA6415C_2 == dev->i2c_adapter->clients[i]->addr ) 
			mxb->tea6415c = dev->i2c_adapter->clients[i];
		if( I2C_TDA9840 == dev->i2c_adapter->clients[i]->addr ) 
			mxb->tda9840 = dev->i2c_adapter->clients[i];
		if( I2C_SAA7111A == dev->i2c_adapter->clients[i]->addr ) 
			mxb->saa7111a = dev->i2c_adapter->clients[i];
		if( 0x60 == dev->i2c_adapter->clients[i]->addr ) 
			mxb->fi1216mk2 = dev->i2c_adapter->clients[i];
	}

	/* check if all devices are present */
	if(    0 == mxb->tea6420_1	|| 0 == mxb->tea6420_2	|| 0 == mxb->tea6415c
	    || 0 == mxb->tda9840	|| 0 == mxb->saa7111a	|| 0 == mxb->fi1216mk2 ) {
		DEB2(printk("mxb_v4l2.o: mxb_v4l2_probe failed for this device.\n"));	
		kfree(mxb);
		return -ENODEV;
	}

	/* all devices are present, probe was successful */	
	DEB2(printk("mxb_v4l2.o: mxb_v4l2_probe succeeded for this device.\n"));	

	/* we store the pointer in our private data field */
	(struct mxb_v4l2*)dev->ext_priv = mxb;

	return 0;
}

/* bring hardware to a sane state. this has to be done, just in case someone
   wants to capture from this device before it has been properly initialized.
   the capture engine would badly fail, because no valid signal arrives on the
   saa7146, thus leading to timeouts and stuff. */
int mxb_v4l2_init_done(struct capture_device* dev)
{
	struct mxb_v4l2* mxb = (struct mxb_v4l2*)dev->ext_priv;
	
	struct {
		int	length;
		char	data[9];
	} saa7740_init[] = {
		{ 3, { 0x80, 0x00, 0x00 } },{ 3, { 0x80, 0x89, 0x00 } },
		{ 3, { 0x80, 0xb0, 0x0a } },{ 3, { 0x00, 0x00, 0x00 } },
		{ 3, { 0x49, 0x00, 0x00 } },{ 3, { 0x4a, 0x00, 0x00 } },
		{ 3, { 0x4b, 0x00, 0x00 } },{ 3, { 0x4c, 0x00, 0x00 } },
		{ 3, { 0x4d, 0x00, 0x00 } },{ 3, { 0x4e, 0x00, 0x00 } },
		{ 3, { 0x4f, 0x00, 0x00 } },{ 3, { 0x50, 0x00, 0x00 } },
		{ 3, { 0x51, 0x00, 0x00 } },{ 3, { 0x52, 0x00, 0x00 } },
		{ 3, { 0x53, 0x00, 0x00 } },{ 3, { 0x54, 0x00, 0x00 } },
		{ 3, { 0x55, 0x00, 0x00 } },{ 3, { 0x56, 0x00, 0x00 } },
		{ 3, { 0x57, 0x00, 0x00 } },{ 3, { 0x58, 0x00, 0x00 } },
		{ 3, { 0x59, 0x00, 0x00 } },{ 3, { 0x5a, 0x00, 0x00 } },
		{ 3, { 0x5b, 0x00, 0x00 } },{ 3, { 0x5c, 0x00, 0x00 } },
		{ 3, { 0x5d, 0x00, 0x00 } },{ 3, { 0x5e, 0x00, 0x00 } },
		{ 3, { 0x5f, 0x00, 0x00 } },{ 3, { 0x60, 0x00, 0x00 } },
		{ 3, { 0x61, 0x00, 0x00 } },{ 3, { 0x62, 0x00, 0x00 } },
		{ 3, { 0x63, 0x00, 0x00 } },{ 3, { 0x64, 0x00, 0x00 } },
		{ 3, { 0x65, 0x00, 0x00 } },{ 3, { 0x66, 0x00, 0x00 } },
		{ 3, { 0x67, 0x00, 0x00 } },{ 3, { 0x68, 0x00, 0x00 } },
		{ 3, { 0x69, 0x00, 0x00 } },{ 3, { 0x6a, 0x00, 0x00 } },
		{ 3, { 0x6b, 0x00, 0x00 } },{ 3, { 0x6c, 0x00, 0x00 } },
		{ 3, { 0x6d, 0x00, 0x00 } },{ 3, { 0x6e, 0x00, 0x00 } },
		{ 3, { 0x6f, 0x00, 0x00 } },{ 3, { 0x70, 0x00, 0x00 } },
		{ 3, { 0x71, 0x00, 0x00 } },{ 3, { 0x72, 0x00, 0x00 } },
		{ 3, { 0x73, 0x00, 0x00 } },{ 3, { 0x74, 0x00, 0x00 } },
		{ 3, { 0x75, 0x00, 0x00 } },{ 3, { 0x76, 0x00, 0x00 } },
		{ 3, { 0x77, 0x00, 0x00 } },{ 3, { 0x41, 0x00, 0x42 } },
		{ 3, { 0x42, 0x10, 0x42 } },{ 3, { 0x43, 0x20, 0x42 } },
		{ 3, { 0x44, 0x30, 0x42 } },{ 3, { 0x45, 0x00, 0x01 } },
		{ 3, { 0x46, 0x00, 0x01 } },{ 3, { 0x47, 0x00, 0x01 } },
		{ 3, { 0x48, 0x00, 0x01 } },
		{ 9, { 0x01, 0x03, 0xc5, 0x5c, 0x7a, 0x85, 0x01, 0x00, 0x54 } },
		{ 9, { 0x21, 0x03, 0xc5, 0x5c, 0x7a, 0x85, 0x01, 0x00, 0x54 } },
		{ 9, { 0x09, 0x0b, 0xb4, 0x6b, 0x74, 0x85, 0x95, 0x00, 0x34 } },
		{ 9, { 0x29, 0x0b, 0xb4, 0x6b, 0x74, 0x85, 0x95, 0x00, 0x34 } },
		{ 9, { 0x11, 0x17, 0x43, 0x62, 0x68, 0x89, 0xd1, 0xff, 0xb0 } },
		{ 9, { 0x31, 0x17, 0x43, 0x62, 0x68, 0x89, 0xd1, 0xff, 0xb0 } },
		{ 9, { 0x19, 0x20, 0x62, 0x51, 0x5a, 0x95, 0x19, 0x01, 0x50 } },
		{ 9, { 0x39, 0x20, 0x62, 0x51, 0x5a, 0x95, 0x19, 0x01, 0x50 } },
		{ 9, { 0x05, 0x3e, 0xd2, 0x69, 0x4e, 0x9a, 0x51, 0x00, 0xf0 } },
		{ 9, { 0x25, 0x3e, 0xd2, 0x69, 0x4e, 0x9a, 0x51, 0x00, 0xf0 } },
		{ 9, { 0x0d, 0x3d, 0xa1, 0x40, 0x7d, 0x9f, 0x29, 0xfe, 0x14 } },
		{ 9, { 0x2d, 0x3d, 0xa1, 0x40, 0x7d, 0x9f, 0x29, 0xfe, 0x14 } },
		{ 9, { 0x15, 0x73, 0xa1, 0x50, 0x5d, 0xa6, 0xf5, 0xfe, 0x38 } },
		{ 9, { 0x35, 0x73, 0xa1, 0x50, 0x5d, 0xa6, 0xf5, 0xfe, 0x38 } },
		{ 9, { 0x1d, 0xed, 0xd0, 0x68, 0x29, 0xb4, 0xe1, 0x00, 0xb8 } },
		{ 9, { 0x3d, 0xed, 0xd0, 0x68, 0x29, 0xb4, 0xe1, 0x00, 0xb8 } },
		{ 3, { 0x80, 0xb3, 0x0a } },
		{-1, { 0} }
	};

	struct i2c_msg msg;

	int i = 0, err = 0;
	struct	tea6415c_video_multiplex vm;	

	DEB2(printk("mxb_v4l2.o: mxb_v4l2_init_done called.\n"));
	
	/* mute audio on tea6420s */
	mxb->tea6420_1->driver->command(mxb->tea6420_1,TEA6420_SWITCH, &TEA6420_line[7][0]);
	mxb->tea6420_2->driver->command(mxb->tea6420_2,TEA6420_SWITCH, &TEA6420_line[7][1]);
	mxb->tea6420_1->driver->command(mxb->tea6420_1,TEA6420_SWITCH, &TEA6420_cd[7][0]);
	mxb->tea6420_2->driver->command(mxb->tea6420_2,TEA6420_SWITCH, &TEA6420_cd[7][1]);

	/* switch to tuner-channel on tea6415c*/
	vm.out = 17;
	vm.in  = 3;
	mxb->tea6415c->driver->command(mxb->tea6415c,TEA6415C_SWITCH, &vm);

	/* select tuner-output on multicable on tea6415c*/
	vm.in  = 3;
	vm.out = 13;
	mxb->tea6415c->driver->command(mxb->tea6415c,TEA6415C_SWITCH, &vm);

	/* select tuner-output on saa7111a */
	i = 0;
	mxb->saa7111a->driver->command(mxb->saa7111a,SAA711X_SET_MODE, &i);
						
	/* tune in some frequency on tuner */
	mxb->fi1216mk2->driver->command(mxb->fi1216mk2,TUNER_SET_TVFREQ, &freq);

	/* the rest for mxb */
	mxb->cur_input = 0;
	mxb->cur_freq = freq;
	mxb->cur_mute = 1;
	mxb->cur_gain = 0;

	mxb->cur_mode = V4L2_TUNER_MODE_STEREO;
	mxb->tda9840->driver->command(mxb->tda9840, TDA9840_SWITCH, &mxb->cur_mode);
			
	/* check if the saa7740 (aka 'sound arena module') is present
	   on the mxb. if so, we must initialize it. due to lack of 
	   informations about the saa7740, the values were reverse
	   engineered. */
	msg.addr = 0x1b;
	msg.flags = 0;
	msg.len = saa7740_init[0].length;
	msg.buf = &saa7740_init[0].data[0];

	if( 1 == (err = i2c_transfer(dev->i2c_adapter, &msg, 1))) {
		for(i = 1;;i++) {
			msg.len = saa7740_init[i].length;		
			if( -1 == msg.len ) {
				break;
			}
			msg.buf = &saa7740_init[i].data[0];
			if( 1 != (err = i2c_transfer(dev->i2c_adapter, &msg, 1))) {
				printk("mxb_v4l2.o: failed to initialize 'sound arena module'.\n");
				goto err;
			}
		}
		printk("mxb_v4l2.o: 'sound arena module' detected.\n");
	}
err:	
	/* the rest for saa7146: you should definitely set some basic values
	   for the input-port handling of the saa7146. */

	/* ext->saa has been filled by the core driver */
	   
	/* some stuff is done via variables */
	saa7146_set_hps_source_and_sync(dev, input_port_selection[mxb->cur_input].hps_source, input_port_selection[mxb->cur_input].hps_sync);

	/* some stuff is done via direct write to the registers */

	/* this is ugly, but because of the fact that this is completely
	   hardware dependend, it should be done directly... */
      	saa7146_write(dev, DD1_STREAM_B,	0x00000000);
	saa7146_write(dev, DD1_INIT,		0x02000200);
	saa7146_write(dev, MC2, (MASK_09 | MASK_25 | MASK_10 | MASK_26));

	return 0;
}

/* interrupt bh-handler. this gets called when irq_mask is != 0.
   it must clear the interrupt-bits in irq_mask it has handled */
/*
void mxb_v4l2_irq_bh(struct capture_device* dev, u32* irq_mask)
{
	struct mxb_v4l2* mxb = (struct mxb_v4l2*)dev->ext_priv;
}
*/

#ifdef SAA7146_AXA
struct {
	struct mxb_v4l2* 	mxb;
	int			mixer_minor;
} mxb_mixer_associations[MAX_SAA7146_BOARDS];

int mval = 0;

int mixer_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
	struct mxb_v4l2* mxb = (struct mxb_v4l2*)file->private_data;
	int ret, val = 0;
/*
	DEB2(printk("mxb_v4l2.o: mixer_ioctl() called.\n"));
*/
	if (cmd == SOUND_MIXER_INFO) {
		mixer_info info;
		strncpy(info.id, "tv card", sizeof(info.id));
		strncpy(info.name, "Multimedia eXtension board", sizeof(info.name));
		info.modify_counter = 42 /* FIXME */;
		if (copy_to_user((void *)arg, &info, sizeof(info)))
			return -EFAULT;
		return 0;
	}
	if (cmd == OSS_GETVERSION) {
		return put_user(SOUND_VERSION, (int *)arg);
	}

	if (_SIOC_DIR(cmd) & _SIOC_WRITE) {
		if (get_user(val, (int *)arg)) {
			return -EFAULT;
		}
	}

	switch (cmd) {
	case MIXER_READ(SOUND_MIXER_RECMASK):
	case MIXER_READ(SOUND_MIXER_CAPS):
	case MIXER_READ(SOUND_MIXER_RECSRC):
	case MIXER_READ(SOUND_MIXER_STEREODEVS):
	case MIXER_WRITE(SOUND_MIXER_RECSRC):
		ret = 0;
		break;
	case MIXER_READ(SOUND_MIXER_DEVMASK):
		ret = SOUND_MASK_VOLUME;
		break;
	case MIXER_WRITE(SOUND_MIXER_VOLUME): {
		struct audio_multiplex am;
		mxb->cur_gain = (((val&0xff)*6)/100);
		memcpy(&am, &TEA6420_line[video_audio_connect[mxb->cur_input]][0],sizeof(struct audio_multiplex));
		am.gain = mxb->cur_gain;
		mxb->tea6420_1->driver->command(mxb->tea6420_1,TEA6420_SWITCH, &am);
		memcpy(&am, &TEA6420_line[video_audio_connect[mxb->cur_input]][1],sizeof(struct audio_multiplex));
		am.gain = mxb->cur_gain;
		mxb->tea6420_2->driver->command(mxb->tea6420_2,TEA6420_SWITCH, &am);
/*
		if( 0 != (result = mxb->tda9840->driver->command(mxb->tda9840, TDA9840_STEREO_ADJUST, &mxb->cur_gain))) {
			printk("mxb_v4l2.o: MIXER_WRITE(SOUND_MIXER_VOLUME) error. result:%d.\n",result);
		}
*/
		DEB2(printk("mxb_v4l2.o: MIXER_WRITE(SOUND_MIXER_VOLUME): 0x%04x => %d\n",val,mxb->cur_gain));
		/* fall through */
	}
	case MIXER_READ(SOUND_MIXER_VOLUME):
		ret = (mxb->cur_gain*100)/6;
		DEB2(printk("mxb_v4l2.o: MIXER_READ(SOUND_MIXER_VOLUME): 0x%04x <= %d\n",ret,mxb->cur_gain));
		break;
	default:
		return -EINVAL;
	}
	if (put_user(ret, (int *)arg)) {
		return -EFAULT;
	}
	
	return 0;
}

static int mixer_open(struct inode *inode, struct file *file)
{
	int minor = MINOR(inode->i_rdev);
	struct mxb_v4l2* mxb = NULL;
	int i = 0;
	
	DEB2(printk("mxb_v4l2.o: mixer_open() called.\n"));

	for(i = 0; i < MAX_SAA7146_BOARDS; i++) {
		if( minor == mxb_mixer_associations[i].mixer_minor ) {
			break;
		}
	}
	if( MAX_SAA7146_BOARDS == i ) {
		DEB2(printk("mxb_v4l2.o: mixer_open(): device does not exist.\n"));
		return -ENODEV;
	}
	mxb = mxb_mixer_associations[i].mxb;

	file->private_data = mxb;
#ifdef MODULE
	MOD_INC_USE_COUNT;
#endif

	return 0;
}

int mixer_release(struct inode *inode, struct file *file)
{
/*
	struct mxb_v4l2* mxb = (struct mxb_v4l2*)file->private_data;
*/
	DEB2(printk("mxb_v4l2.o: mixer_release() called.\n"));

	file->private_data = NULL;
#ifdef MODULE
	MOD_DEC_USE_COUNT;
#endif
	return 0;
}

struct file_operations mixer_fops = {
	owner:		THIS_MODULE,
	ioctl:		mixer_ioctl,
	open:		mixer_open,
	release:	mixer_release,
};
#endif

/* this function only gets called when the probing was successful */
int mxb_v4l2_attach(struct capture_device* dev)
{
	struct mxb_v4l2* mxb = (struct mxb_v4l2*)dev->ext_priv;
	
	DEB2(printk("mxb_v4l2.o: mxb_v4l2_attach called.\n"));

	/* checking for i2c-devices can be omitted here, because we
	   already did this in "mxb_vl42_probe" */

	mxb_v4l2_num++;
	
	i2c_inc_use_client(mxb->tea6420_1);
	i2c_inc_use_client(mxb->tea6420_2);
	i2c_inc_use_client(mxb->tea6415c);
	i2c_inc_use_client(mxb->tda9840);
	i2c_inc_use_client(mxb->saa7111a);
	i2c_inc_use_client(mxb->fi1216mk2);

#ifdef SAA7146_AXA
	if( mixer != 0) {
		for(i = 0; i < MAX_SAA7146_BOARDS; i++) {
			if( NULL == mxb_mixer_associations[i].mxb ) {
				break;
			}
		}
		if( MAX_SAA7146_BOARDS == i ) {
			DEB2(printk("mxb_v4l2.o: cannot register mixer.\n"));
		} else {
			if((mxb_mixer_associations[i].mixer_minor = register_sound_mixer(&mixer_fops,-1)) < 0) {
				DEB2(printk("mxb_v4l2.o: register_sound_mixer() failed. continuing anyway.\n"));
			} else {
				mxb_mixer_associations[i].mxb = mxb;
			}
		}
	}
#endif	
	return 0;
}

int mxb_v4l2_detach(struct capture_device* dev)
{
	struct mxb_v4l2* mxb = (struct mxb_v4l2*)dev->ext_priv;

	DEB2(printk("mxb_v4l2.o: mxb_v4l2_detach called.\n"));

	i2c_dec_use_client(mxb->tea6420_1);
	i2c_dec_use_client(mxb->tea6420_2);
	i2c_dec_use_client(mxb->tea6415c);
	i2c_dec_use_client(mxb->tda9840);
	i2c_dec_use_client(mxb->saa7111a);
	i2c_dec_use_client(mxb->fi1216mk2);

	mxb_v4l2_num--;

#ifdef SAA7146_AXA
	if( 0 != mixer ) {
		for(i = 0; i < MAX_SAA7146_BOARDS; i++) {
			if( mxb == mxb_mixer_associations[i].mxb ) {
				break;
			}
		}
		if( MAX_SAA7146_BOARDS == i ) {
			DEB2(printk("mxb_v4l2.o: cannot unregister mixer.\n"));
		} else {
			unregister_sound_mixer(mxb_mixer_associations[i].mixer_minor);
			mxb_mixer_associations[i].mixer_minor = -1;
			mxb_mixer_associations[i].mxb = NULL;
		}
	}
#endif
	kfree(mxb);

	return 0;
}

void mxb_v4l2_open(struct capture_device* dev)
{
#ifdef MODULE
	MOD_INC_USE_COUNT;
#endif
}

void mxb_v4l2_close(struct capture_device* dev)
{
	struct mxb_v4l2* mxb = (struct mxb_v4l2*)dev->ext_priv;

	/* be sure to switch off audio */
	if( 0 == mxb->cur_mute ) {

		mxb->tea6420_1->driver->command(mxb->tea6420_1,TEA6420_SWITCH, &TEA6420_line[7][0]);
		mxb->tea6420_2->driver->command(mxb->tea6420_2,TEA6420_SWITCH, &TEA6420_line[7][1]);

		mxb->cur_mute = 1;
	}

#ifdef MODULE
	MOD_DEC_USE_COUNT;
#endif
}

void mxb_v4l2_vbi_open(struct capture_device* dev)
{
#ifdef MODULE
	MOD_INC_USE_COUNT;
#endif
}

int mxb_v4l2_vbi_bypass(struct capture_device* dev)
{
	struct mxb_v4l2* mxb = (struct mxb_v4l2*)dev->ext_priv;
	
	int i = 1;

	/* switch bypass in saa7111a */
	if ( 0 != mxb->saa7111a->driver->command(mxb->saa7111a,SAA711X_VBI_BYPASS, &i)) {
		printk("mxb_v4l2.o: VBI_BYPASS: could not address saa7111a.\n");
		return -1;
	}			

	return 0;
}

void mxb_v4l2_vbi_close(struct capture_device* dev)
{
#ifdef MODULE
	MOD_DEC_USE_COUNT;
#endif
}

/* hack: this should go into saa711x */
int saa7111_set_gpio(struct capture_device *dev, int bl)
{
	struct mxb_v4l2* mxb = (struct mxb_v4l2*)dev->ext_priv;

	s32 byte = 0x0;
	int result = 0;
	
	DEB2(printk("mxb_v4l2.o: saa7111_set_gpio() called.\n"));

	/* get the old register contents */
	if ( -1 == (byte = i2c_smbus_read_byte_data(mxb->saa7111a, 0x11))) {
		printk("mxb_v4l2.o: saa7111_set_gpio.o: could not read from saa711x\n");
		return -EFAULT;
	}
	
	if( 0 == bl ) {
		byte &= 0x7f;
	} else {
		byte |= 0x80;
	}

	/* write register contents back */
	if ( 0 != (result = i2c_smbus_write_byte_data(mxb->saa7111a, 0x11, byte))) {
		printk("mxb_v4l2.o: saa7111_set_gpio.o: could not write to saa711x\n");
		return -EFAULT;
	}

	return 0;
}



int mxb_v4l2_ioctl(void *id, unsigned int cmd, void *arg) 
{
	struct device_open *o = (struct device_open *)id;
	struct capture_device *dev = o->dev;
	struct mxb_v4l2* mxb = (struct mxb_v4l2*)dev->ext_priv;

	switch(cmd) {
	case MXB_S_AUDIO_CD:
	{
		int i = *(int*)arg;
				
		if( i < 0 || i >= MXB_AUDIOS ) {
			DEB2(printk("mxb_v4l2.o: illegal argument to MXB_S_AUDIO_CD: i:%d.\n",i));
			return -EINVAL;
		}
		
		DEB2(printk("mxb_v4l2.o: MXB_S_AUDIO_CD: i:%d.\n",i));

		mxb->tea6420_1->driver->command(mxb->tea6420_1,TEA6420_SWITCH, &TEA6420_cd[i][0]);
		mxb->tea6420_2->driver->command(mxb->tea6420_2,TEA6420_SWITCH, &TEA6420_cd[i][1]);

		return 0;
	}
	case MXB_S_AUDIO_LINE:
	{
		int i = *(int*)arg;
				
		if( i < 0 || i >= MXB_AUDIOS ) {
			DEB2(printk("mxb_v4l2.o: illegal argument to MXB_S_AUDIO_LINE: i:%d.\n",i));
			return -EINVAL;
		}
		
		DEB2(printk("mxb_v4l2.o: MXB_S_AUDIO_LINE: i:%d.\n",i));
		mxb->tea6420_1->driver->command(mxb->tea6420_1,TEA6420_SWITCH, &TEA6420_line[i][0]);
		mxb->tea6420_2->driver->command(mxb->tea6420_2,TEA6420_SWITCH, &TEA6420_line[i][1]);

		return 0;
	}
	case VIDIOC_G_TUNER:
	{
		struct v4l2_tuner *t = arg;
		int byte = 0;
//		int new_mode = 0;
//		int result = 0;
		
		if( 0 != t->input ) {
			DEB2(printk("mxb_v4l2.o: VIDIOC_G_TUNER: channel %d does not have a tuner attached.\n", t->input));
			return -EINVAL;
		}

		strcpy(t->name, "fi1216mk2");

		v4l2_video_std_construct(&t->std, V4L2_STD_PAL, V4L2_TRANSM_STD_B|V4L2_TRANSM_STD_G);
		/* fixme: _CAP_NORM needed? */
		t->capability = V4L2_TUNER_CAP_NORM | V4L2_TUNER_CAP_STEREO | V4L2_TUNER_CAP_LANG1 | V4L2_TUNER_CAP_LANG2 | V4L2_TUNER_CAP_SAP;
		t->rangelow = 772;	/* 48.25 MHZ / 62.5 kHz = 772, see fi1216mk2-specs, page 2 */
		t->rangehigh = 13684;	/* 855.25 MHz / 62.5 kHz = 13684 */
		t->signal = 0;
		t->afc = 0;		

		byte = mxb->tda9840->driver->command(mxb->tda9840,TDA9840_DETECT, NULL);
		t->audmode = mxb->cur_mode;
		if( byte < 0 ) {
			t->rxsubchans  = 0;
		} else {
			switch(byte) {
				case TDA9840_MONO_DETECT: {
					t->rxsubchans 	= V4L2_TUNER_SUB_MONO;
//					t->audmode	= V4L2_TUNER_MODE_MONO;
//					new_mode = TDA9840_SET_MONO;
					DEB2(printk("mxb_v4l2.o: VIDIOC_G_TUNER: V4L2_TUNER_MODE_MONO.\n"));
					break;
				}
				case TDA9840_DUAL_DETECT: {
					t->rxsubchans 	= V4L2_TUNER_SUB_LANG1 | V4L2_TUNER_SUB_LANG2;
//					t->audmode 	= V4L2_TUNER_MODE_LANG1;
//					new_mode = TDA9840_SET_LANG1;
					DEB2(printk("mxb_v4l2.o: VIDIOC_G_TUNER: V4L2_TUNER_MODE_LANG1.\n"));
					break;
				}
				case TDA9840_STEREO_DETECT: {
					t->rxsubchans 	= V4L2_TUNER_SUB_STEREO | V4L2_TUNER_SUB_MONO;
//					t->audmode 	= V4L2_TUNER_MODE_STEREO;
//					new_mode = TDA9840_SET_STEREO;
					DEB2(printk("mxb_v4l2.o: VIDIOC_G_TUNER: V4L2_TUNER_MODE_STEREO.\n"));
					break;
				}
				default: { /* TDA9840_INCORRECT_DETECT */
					t->audmode 	= V4L2_TUNER_MODE_MONO;
//					t->rxsubchans 	= V4L2_TUNER_MODE_MONO;
//					new_mode = TDA9840_SET_MONO;
					DEB2(printk("mxb_v4l2.o: VIDIOC_G_TUNER: TDA9840_INCORRECT_DETECT => V4L2_TUNER_MODE_MONO\n"));
					break;
				}
			}
		}

		return 0;
	}
	case VIDIOC_S_TUNER:
	{
		struct v4l2_tuner *t = arg;
		int result = 0;
		int byte = 0;
		
		if( 0 != t->input ) {
			DEB2(printk("mxb_v4l2.o: VIDIOC_S_TUNER: channel %d does not have a tuner attached.\n",t->input));
			return -EINVAL;
		}
	
		switch(t->audmode) {
			case V4L2_TUNER_MODE_STEREO: {
				mxb->cur_mode = V4L2_TUNER_MODE_STEREO;
				byte = TDA9840_SET_STEREO;
				DEB2(printk("mxb_v4l2.o: VIDIOC_S_TUNER: V4L2_TUNER_MODE_STEREO\n"));
				break;
			}
			case V4L2_TUNER_MODE_LANG1: {
				mxb->cur_mode = V4L2_TUNER_MODE_LANG1;
				byte = TDA9840_SET_LANG1;
				DEB2(printk("mxb_v4l2.o: VIDIOC_S_TUNER: V4L2_TUNER_MODE_LANG1\n"));
				break;
			}
			case V4L2_TUNER_MODE_LANG2: {
				mxb->cur_mode = V4L2_TUNER_MODE_LANG2;
				byte = TDA9840_SET_LANG2;
				DEB2(printk("mxb_v4l2.o: VIDIOC_S_TUNER: V4L2_TUNER_MODE_LANG2\n"));
				break;
			}
			default: { /* case V4L2_TUNER_MODE_MONO: {*/
				mxb->cur_mode = V4L2_TUNER_MODE_MONO;
				byte = TDA9840_SET_MONO;
				DEB2(printk("mxb_v4l2.o: VIDIOC_S_TUNER: TDA9840_SET_MONO\n"));
				break;
			}
		}

		if( 0 != (result = mxb->tda9840->driver->command(mxb->tda9840, TDA9840_SWITCH, &byte))) {
			printk("mxb_v4l2.o: VIDIOC_S_TUNER error. result:%d, byte:%d\n",result,byte);
		}
				
		return 0;
	}
	case VIDIOC_G_FREQ:
	{
		u32 *freq = (int*)arg;

		if(0 != mxb->cur_input) {
			DEB2(printk("mxb_v4l2.o: VIDIOC_G_FREQ: channel %d does not have a tuner!\n",mxb->cur_input));
			return -EINVAL;
		}

		*freq = mxb->cur_freq;
		DEB2(printk("mxb_v4l2.o: VIDIOC_G_FREQ: freq:0x%08x.\n",*freq));

		return 0;
	}
	case VIDIOC_S_FREQ:
	{
		u32 freq = *(int*)arg;
		int t_locked = 0;
		int v_byte = 0;
		
		if(0 != mxb->cur_input) {
			DEB2(printk("mxb_v4l2.o: VIDIOC_S_FREQ: channel %d does not have a tuner!\n",mxb->cur_input));
			return -EINVAL;
		}

		DEB2(printk("mxb_v4l2.o: VIDIOC_S_FREQ: freq:0x%08x.\n",freq));

		mxb->cur_freq = freq;

		/* tune in desired frequency */			
		mxb->fi1216mk2->driver->command(mxb->fi1216mk2,TUNER_SET_TVFREQ, &freq);

		/* check if pll of tuner & saa7111a is locked */
//		mxb->fi1216mk2->driver->command(mxb->fi1216mk2,TUNER_IS_LOCKED, &t_locked);
		mxb->saa7111a->driver->command(mxb->saa7111a,SAA711X_GET_STATUS_BYTE, &v_byte);

		/* not locked -- anything to do here ? */
		if( 0 == t_locked || 0 != (v_byte & MASK_06)) {
		}

		/* changing the frequency should invalidate the vbi-counter 
		   (=> alevt) */
		dev->vbi_counter = 0;

		return 0;
	}
	case VIDIOC_G_INPUT:
	{
		int* input = (int *)arg;
		*input = mxb->cur_input;

		DEB2(printk("mxb_v4l2.o: VIDIOC_G_INPUT %d.\n",*input));
		return 0;		
	}	
	case VIDIOC_S_INPUT:
	{
		int input = *(int *)arg;
		struct	tea6415c_video_multiplex vm;	
		int i = 0;

		DEB2(printk("mxb_v4l2.o: VIDIOC_S_INPUT %d.\n",input));

		if (input < 0 || input >= MXB_INPUTS) {
			return -EINVAL;
		}
	
		if ( 0 != dev->streaming ) {
			DEB2(printk("mxb_v4l2.o: VIDIOC_S_INPUT illegal while streaming.\n"));
			return -EPERM;
		}

		mxb->cur_input = input;
		
		saa7146_set_hps_source_and_sync(dev, input_port_selection[input].hps_source, input_port_selection[input].hps_sync);

		/* prepare switching of tea6415c and saa7111a;
		   have a look at the 'background'-file for further informations  */
		switch( input ) {
			
			case TUNER:
			{
				i = 0;
				vm.in  = 3;
				vm.out = 17;
								
			if ( 0 != mxb->tea6415c->driver->command(mxb->tea6415c,TEA6415C_SWITCH, &vm)) {
					printk("mxb_v4l2.o: VIDIOC_S_INPUT: could not address tea6415c #1\n");
					return -EFAULT;
				}
				/* connect tuner-output always to multicable */
				vm.in  = 3;
				vm.out = 13;
				break;				
			}
			case AUX3_YC:
			{
				/* nothing to be done here. aux3_yc is
				   directly connected to the saa711a */
				i = 5;
				break;
			}
			case AUX3:
			{
				/* nothing to be done here. aux3 is
				   directly connected to the saa711a */
				i = 1;
				break;
			}
			case AUX1:
			{
				i = 0;
				vm.in  = 1;
				vm.out = 17;
				break;
			}
			case CAM:
			{
				i = 0;
				vm.in  = 10;
				vm.out = 17;
				break;
			}
			case CAM_YC:
			{
				i = 4;
				vm.in  = 5;
				vm.out = 17;

				if ( 0 != mxb->tea6415c->driver->command(mxb->tea6415c,TEA6415C_SWITCH, &vm)) {
					printk("mxb_v4l2.o: VIDIOC_S_INPUT: could not address tea6415c #2\n");
					return -EFAULT;
				}
				vm.in  = 6;
				vm.out = 18;
				break;
			}
		}

		/* switch video in tea6415c only if necessary */
		switch( input ) {
			case TUNER:
			case AUX1:
			case CAM:
			case CAM_YC:
			{
				if ( 0 != mxb->tea6415c->driver->command(mxb->tea6415c,TEA6415C_SWITCH, &vm)) {
					printk("mxb_v4l2.o: VIDIOC_S_INPUT: could not address tea6415c #3\n");
					return -EFAULT;
				}
				break;
			}
			default:
			{
				break;
			}
		}
				
		/* switch video in saa7111a */
		if ( 0 != mxb->saa7111a->driver->command(mxb->saa7111a,SAA711X_SET_MODE, &i)) {
			printk("mxb_v4l2.o: VIDIOCSCHAN: could not address saa7111a #1.\n");
		}			

		/* switch the audio-source only if necessary */
		if( 0 == mxb->cur_mute ) {
			mxb->tea6420_1->driver->command(mxb->tea6420_1,TEA6420_SWITCH, &TEA6420_line[video_audio_connect[input]][0]);
			mxb->tea6420_2->driver->command(mxb->tea6420_2,TEA6420_SWITCH, &TEA6420_line[video_audio_connect[input]][1]);
		}

		return 0;
	}
	case VIDIOC_G_AUDIO:
	{
		struct v4l2_audio *a = arg;

		if( a->audio < 0 || a->audio > MXB_INPUTS ) {
	 		DEB2(printk("mxb_v4l2.o: VIDIOC_G_AUDIO %d out of range.\n",a->audio));
			return -EINVAL;
		}
		
 		DEB2(printk("mxb_v4l2.o: VIDIOC_G_AUDIO %d.\n",a->audio));
		memcpy(a, &mxb_audios[video_audio_connect[mxb->cur_input]], sizeof(struct v4l2_audio));
		
		return 0;
	}
	case VIDIOC_S_AUDIO:
	{
		struct v4l2_audio *a = arg;
		DEB2(printk("mxb_v4l2.o: VIDIOC_S_AUDIO %d.\n",a->audio));
		return 0;
	}
	case VIDIOC_ENUMINPUT:
	{
		struct v4l2_input *i = arg;
		
		DEB2(printk("mxb_v4l2.o: VIDIOC_ENUMINPUT %d.\n",i->index));

		/* sanity check to satisfy xawtv */
		if( i->index < 0 || i->index >= MXB_INPUTS) {
			return -EINVAL;
		}
		
		memcpy(i, &mxb_inputs[i->index], sizeof(struct v4l2_input));

		return 0;
	}
	
	/* the saa7146 provides some controls (brightness, contrast, saturation)
	   which gets registered *after* this function. because of this we have
	   to return with a value != 0 even if the function succeded.. */
	case VIDIOC_QUERYCTRL:
	{
		struct v4l2_queryctrl	*qc = arg;
		int 		i;

		for (i = MAXCONTROLS - 1; i >= 0; i--) {
			if (mxb_controls[i].id == qc->id) {
				mxb_controls[i].category = qc->category;
				memcpy(mxb_controls[i].group, qc->group, sizeof(qc->group));
				*qc = mxb_controls[i];
				DEB2(printk("mxb_v4l2.o: VIDIOC_QUERYCTRL %d.\n",qc->id));
				return 0;
			}
		}

/*		DEB2(printk("mxb_v4l2.o: VIDIOC_QUERYCTRL: no such control %d.\n",qc->id));
*/
		return -EAGAIN;
	}
	case VIDIOC_G_CTRL:
	{
		struct v4l2_control	*vc = arg;
		int i;

		for (i = MAXCONTROLS - 1; i >= 0; i--) {
			if (mxb_controls[i].id == vc->id) {
				break;
			}
		}
		
		if( i < 0 ) {
/*			DEB2(printk("mxb_v4l2.o: VIDIOC_G_CTRL: unknown control.\n"));
*/			return -EAGAIN;
		}
			
		switch (vc->id ) {
			case V4L2_CID_AUDIO_MUTE: {
				vc->value = mxb->cur_mute;
				DEB2(printk("mxb_v4l2.o: VIDIOC_G_CTRL V4L2_CID_AUDIO_MUTE:%d.\n",vc->value));
				return 0;
			}
		}
		
		DEB2(printk("mxb_v4l2.o: VIDIOC_G_CTRL V4L2_CID_AUDIO_MUTE:%d.\n",vc->value));
		return 0;
	}

	case VIDIOC_S_CTRL:
	{
		struct	v4l2_control	*vc = arg;
		int i = 0;
		
		for (i = MAXCONTROLS - 1; i >= 0; i--) {
			if (mxb_controls[i].id == vc->id) {
				break;
			}
		}
		
		if( i < 0 ) {
/*			DEB2(printk("mxb_v4l2.o: VIDIOC_S_CTRL: unknown control.\n"));
*/			return -EAGAIN;
		}
		
		switch (vc->id ) {
			case V4L2_CID_AUDIO_MUTE: {
				mxb->cur_mute = vc->value;
				if( 0 == vc->value ) {
					/* switch the audio-source */
					mxb->tea6420_1->driver->command(mxb->tea6420_1,TEA6420_SWITCH, &TEA6420_line[video_audio_connect[mxb->cur_input]][0]);
					mxb->tea6420_2->driver->command(mxb->tea6420_2,TEA6420_SWITCH, &TEA6420_line[video_audio_connect[mxb->cur_input]][1]);
				} else {
					mxb->tea6420_1->driver->command(mxb->tea6420_1,TEA6420_SWITCH, &TEA6420_line[7][0]);
					mxb->tea6420_2->driver->command(mxb->tea6420_2,TEA6420_SWITCH, &TEA6420_line[7][1]);
				}
				DEB2(printk("mxb_v4l2.o: VIDIOC_S_CTRL, V4L2_CID_AUDIO_MUTE: %d.\n",vc->value));
				break;
			}
		}
		return 0;
	}
	case VIDIOC_ENUMSTD:
	{
		struct v4l2_enumstd *e = arg;
		DEB2(printk("mxb_v4l2.o: VIDIOC_ENUMSTD: %d\n",e->index));
		if( e->index < 3) {
			return saa7146_enumstd(dev,e);
		}
		if( 3 == e->index ) {
			v4l2_video_std_construct(&e->std, V4L2_STD_PAL, V4L2_TRANSM_STD_I);
			e->inputs = (u32)-1;	/* we assume: all inputs support PAL/NTSC/SECAM */
			e->outputs = 0;		/* we don't have any outputs */
			return 0;
		}
		
		return -EINVAL;
	}
	case VIDIOC_S_STD:
	{
		struct v4l2_standard *std = arg;
		
		DEB2(printk("mxb_v4l2.o: VIDIOC_S_STD: %s\n",std->name));
		if( 0 != saa7146_s_std(o,std)) {
			DEB2(printk("mxb_v4l2.o: VIDIOC_S_STD: something went wrong in saa7146_s_std\n"));
			return -EINVAL;
		}
		if( dev->standard.transmission == V4L2_TRANSM_STD_I ) {
			DEB2(printk("mxb_v4l2.o: recognized V4L2_TRANSM_STD_I\n"));
			/* set the 7146 gpio register -- I don't know what this does exactly */
      			saa7146_write(dev, GPIO_CTRL,	0x00404050);
			/* unset the 7111 gpio register -- I don't know what this does exactly */
			saa7111_set_gpio(dev,0);
			return 0;
		} else {
			/* set the 7146 gpio register -- I don't know what this does exactly */
      			saa7146_write(dev, GPIO_CTRL,	0x00404050);
			/* set the 7111 gpio register -- I don't know what this does exactly */
			saa7111_set_gpio(dev,1);
			return 0;
		}
		return -EINVAL;
	}
	default:
/*
		DEB2(printk("mxb_v4l2.o: does not handle this ioctl.\n"));
*/
		return -ENOIOCTLCMD;
	}
	return 0;
}

struct saa7146_v4l2_extension extension = {
	MXB_IDENTIFIER,
	MXB_INPUTS,
	MXB_AUDIOS,
	V4L2_FLAG_TUNER,
	&unit_video[0],
	&unit_vbi[0],
	&ioctls[0],
	0,
	NULL, /* mxb_v4l2_irq_bh, */
	mxb_v4l2_preinit,
	mxb_v4l2_probe,
	mxb_v4l2_init_done,
	mxb_v4l2_attach,
	mxb_v4l2_detach,

	mxb_v4l2_open,
	mxb_v4l2_close,

	mxb_v4l2_vbi_open,
	mxb_v4l2_vbi_bypass,
	mxb_v4l2_vbi_close,

	mxb_v4l2_ioctl,
};

int __init mxb_v4l2_init(void)
{
	DEB2(printk("mxb_v4l2.o: mxb_v4l2_init called.\n"));
	
	if( 0 != saa7146_v4l2_register_extension(&extension)) {
		DEB1(printk("mxb_v4l2.o: failed to register extension.\n"));
		return -ENODEV;
	}

	/* mxb_v4l2_num gets increased if the probe function is called successfully */

	if( 0 == mxb_v4l2_num ) {
		printk(KERN_ERR "mxb_v4l2.o: no mxb(s) found.\n");
		return -ENODEV;
	}

	printk(KERN_ERR "mxb_v4l2.o: %d mxb(s) found.\n", mxb_v4l2_num);
	return 0;
}

#ifdef MODULE
MODULE_AUTHOR("Michael Hunold <michael@mihu.de>");
MODULE_DESCRIPTION("v4l2-driver for the Multimedia eXtension board");

EXPORT_NO_SYMBOLS;

int init_module(void) 
{
	return mxb_v4l2_init();
}

void cleanup_module(void) 
{
	DEB2(printk("mxb_v4l2.o: cleanup_module called.\n"));
	saa7146_v4l2_unregister_extension(&extension);
}

#endif

#ifdef MODULE
MODULE_LICENSE("GPL");
#endif
