#include <linux/module.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <linux/i2c.h>

#include "saa711x.h"

#ifdef MODULE
MODULE_PARM(debug,"i");
MODULE_PARM(mode,"i");
#endif

/* insmod parameter: some programs (e.g. vic) do not allow to
   specify the used video-mode, so you have to tell this to the
   modules by hand, 0 = PAL, 1 = NTSC, 2 = SECAM */
static int mode = 0;

/* insmod parameter: ... should be quite clear */
static int debug = 0;
#define dprintk	if (debug) printk

/* unique ID allocation */
static int saa711x_id = 0;

static struct i2c_driver driver;

/* this functions sets partial bits of the register 'subaddress' from the
   i2c-client 'client'. Valid values of 'arg' may range between
   'range_begin' and 'range_end'; the corresponding bits to be set
   range from 'bit_begin' to 'bit_end' */
int	saa711x_set_register(struct i2c_client *client, u8 arg, int subaddress, int begin, int length)
{
	s32 byte = 0x0;
	int i = 0, result = 0;

	dprintk("saa711x.o: ==> set_register: arg:0x%02x, subaddress:0x%02x, begin:%d, lenght:%d\n",arg, subaddress, begin, length); 
	
	/* get the old register contents */
	if ( -1 == (byte = i2c_smbus_read_byte_data(client, subaddress))) {
		dprintk("saa711x.o: could not read from saa711x\n");
		return -EFAULT;
	}

	/* clear out corresponding bits */
	for(i = begin; i < (begin+length); i++)
		byte &= ~(1 << i);


	/* set these bits new */
	byte |= (arg << begin);

	/* write register contents back */
	if ( 0 != (result = i2c_smbus_write_byte_data(client, subaddress, byte))) {
		dprintk("saa711x.o: could write to access saa711x\n");
		return -EFAULT;
	}
	
	/* everythings fine */
	return 0;
}

static int saa711x_command(struct i2c_client *client, unsigned int cmd, void *arg)
{
	u8 byte;

	dprintk("saa711x.o: ==> saa711x_command\n");
	
	switch (cmd)
	{
		case SAA711X_VBI_BYPASS:
		{
			int i = *(int*)arg;
                        int result = 0;

			dprintk("saa711x.o: SAA711X_VBI_BYPASS, %d.\n",i);

			if ( -1 == (result = i2c_smbus_read_byte_data(client, SAA711X_OUTPUT_CONTROL_3))) {
				printk("saa711x.o: could not read from saa711x.\n");
				return -EFAULT;
			}
                        
                        byte  = result;
                        byte &= 0xf0;
                        
                        if( i != 0 ) {
                        	byte |= 0x0a;
			}

			if ( 0 != (result = i2c_smbus_write_byte_data(client, SAA711X_OUTPUT_CONTROL_3, byte))) {
				printk("saa711x.o: could not write to saa711x.\n");
				return -EFAULT;
			}

			return 0;
		}


		case SAA711X_SUBADDRESS_WRITE:
		{
			struct saa711x_subaddress_access *sw = arg;
			int result = 0;

			dprintk("saa711x.o: SAA711X_SUBADDRESS_WRITE.\n");

			if ( 0 != (result = i2c_smbus_write_byte_data(client, sw->subaddress, sw->byte))) {
				dprintk("saa711x.o: could not write to saa711x.\n");
				return -EFAULT;
			}

			return 0;
		}
		
		case SAA711X_GET_STATUS:
		{
			int	*iarg = arg;
			s32	result = 0;
			
			dprintk("saa711x.o: SAA711X_GET_STATUS.\n");

			if ( -1 == (result = i2c_smbus_read_byte_data(client, SAA711X_STATUS_BYTE))) {
				dprintk("saa711x.o: could not read from saa711x.\n");
				return -EFAULT;
			}
			
			byte = result;
			
			/* good means, that colour signal has been detected
			   and that horizontal frequency is locked */
			if ((byte & 0x41) == 0x1)
				*iarg = SAA711X_STATUS_GOOD;
			else
				*iarg = SAA711X_STATUS_BAD;

			dprintk("saa711x.o: SAA711X_GET_STATUS, byte:0x%02x\n",byte);

			return 0;
		}

		case SAA711X_GET_STATUS_BYTE:
		{
			s32 result = 0;
			u8* byte = (u8*)arg;

			dprintk("saa711x.o: SAA711X_GET_STATUS_BYTE.\n");
			
			if ( -1 == (result = i2c_smbus_read_byte_data(client, SAA711X_STATUS_BYTE))) {
				dprintk("saa711x.o: could not read from saa711x.\n");
				return -EFAULT;
			}
			
			dprintk("saa711x.o: SAA711X_GET_STATUS_BYTE, byte:0x%02x\n",result);
			
                        *byte = result;
                        
			return 0;
		}
		
		case SAA711X_SET_NORM:
		{
			int *iarg = arg;
			int result = 0;

			dprintk("saa711x.o: SAA711X_SET_NORM.\n");
			
			switch (*iarg) {

			case SAA711X_VIDEO_MODE_NTSC: {
				result  = saa711x_set_register(client, 0xf6, SAA711X_HORIZONTAL_SYNC_START,0,8);
				result += saa711x_set_register(client, 0xe6, SAA711X_HORIZONTAL_SYNC_STOP,0,8);
				result += saa711x_set_register(client, 0x02, SAA711X_LUMINANCE_CONTROL,0,2);
				result += saa711x_set_register(client, 0x00, SAA711X_CHROMA_CONTROL,4,4);
	
				return ( 0 != result ? -EFAULT : 0 );
			}

			case SAA711X_VIDEO_MODE_SECAM: {
				result  = saa711x_set_register(client, 0xf0, SAA711X_HORIZONTAL_SYNC_START,0,8);
				result += saa711x_set_register(client, 0x30, SAA711X_HORIZONTAL_SYNC_STOP,0,8);
				result += saa711x_set_register(client, 0x00, SAA711X_LUMINANCE_CONTROL,0,2);
				result += saa711x_set_register(client, 0x05, SAA711X_CHROMA_CONTROL,4,4);

				return ( 0 != result ? -EFAULT : 0 );
			}
			case SAA711X_VIDEO_MODE_PAL: {
				result  = saa711x_set_register(client, 0xf0, SAA711X_HORIZONTAL_SYNC_START,0,8);
				result += saa711x_set_register(client, 0x30, SAA711X_HORIZONTAL_SYNC_STOP,0,8);
				result += saa711x_set_register(client, 0x02, SAA711X_LUMINANCE_CONTROL,0,2);
				result += saa711x_set_register(client, 0x00, SAA711X_CHROMA_CONTROL,4,4);

				return ( 0 != result ? -EFAULT : 0 );
			}
			
			default:
				return -EINVAL;

			}
		}

		case SAA711X_SET_MODE: {
			dprintk("saa711x.o: SAA711X_SET_MODE\n");
			return(saa711x_set_register(client, *((int*)arg), SAA711X_ANALOG_INPUT_CONTROL_1,0,3));
			}
		case SAA711X_SET_UPDATE_GAIN: {
			dprintk("saa711x.o: SAA711X_SET_GAIN\n");
			return(saa711x_set_register(client, *((int*)arg), SAA711X_ANALOG_INPUT_CONTROL_1,3,3));
			}
		case SAA711X_SET_FUSE: {
			dprintk("saa711x.o: SAA711X_SET_FUSE\n");
			return(saa711x_set_register(client, *((int*)arg), SAA711X_ANALOG_INPUT_CONTROL_1,6,2));
			}
		default:
			return -EINVAL;
	}

	return 0;
}

int saa711x_detect(struct i2c_adapter *adapter, int address, unsigned short flags, int  kind, char* name, struct i2c_client** new_client)
{
	struct	i2c_client *client;
	int err = 0;

	dprintk("saa711x.o: ==> saa711x_detect\n");

	/* let's see whether this adapter can support what we need */
	if ( 0 == i2c_check_functionality(adapter, I2C_FUNC_SMBUS_READ_BYTE_DATA|I2C_FUNC_SMBUS_WRITE_BYTE_DATA)) {
		return 0;
	}

	/* allocate memory for client structure */
	client = kmalloc(sizeof(struct i2c_client), GFP_KERNEL);
        if (0 == client) {
		return -ENOMEM;
	}
	
	/* fill client structure */
	sprintf(client->name,"%s (0x%02x)", name, address);

	client->id = saa711x_id++;
	client->flags = 0;
	client->addr = address;
	client->adapter = adapter;
	client->driver = &driver;
	client->data = NULL;

	/* tell the i2c layer a new client has arrived */
	if (0 != (err = i2c_attach_client(client))) {
		kfree(client);
		return err;
	}
	
	printk("saa711x.o: detected @ 0x%02x on adapter %s\n",2*address,&client->adapter->name[0]);
	
	*new_client = client;
	return 0;
}

static int saa711x_detach(struct i2c_client *client)
{
	int err = 0;

	dprintk("saa711x.o: ==> saa711x_detach\n");

	if ( 0 != (err = i2c_detach_client(client))) {
		printk("saa711x.o: Client deregistration failed, client not detached.\n");
		return err;
	}
	
	kfree(client);

	return 0;
}

void saa711x_inc_use(struct i2c_client *client)
{
#ifdef MODULE
	MOD_INC_USE_COUNT;
#endif
}

void saa711x_dec_use(struct i2c_client *client)
{
#ifdef MODULE
	MOD_DEC_USE_COUNT;
#endif
}

#ifdef MODULE
static
#else
extern
#endif
	int __init saa711x_init(void)
{
	int res = 0;

	printk("saa711x.o version %s (%s)\n",VERSION,DATE);

	if( 0 != (res = i2c_add_driver(&driver))) {
		printk("saa711x.o: Driver registration failed, module not inserted.\n");
		return res;
	}
	
	return 0;
}

int __init saa711x_cleanup(void)
{	
	int res = 0;

	if ( 0 != (res = i2c_del_driver(&driver))) {
		printk("saa711x.o: Driver deregistration failed, module not removed.\n");
		return res;
	}
	
	return 0;
}
