#define DATE "20000701"
#define VERSION "1.2"

#include <linux/module.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <linux/i2c.h>

#ifdef SENSORS
	#include <linux/i2c-proc.h>
#endif

#include "tea6415c.h"

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,13)
#define THIS_MODULE NULL
#endif

#ifdef MODULE
MODULE_PARM(debug,"i");
#endif

static int debug = 0;	/* insmod parameter */
#define dprintk	if (debug) printk

#define TEA6415C_NUM_INPUTS	8
#define TEA6415C_NUM_OUTPUTS	6

/* each client has this additional data */
struct tea6415c_data {
	int	sysctl_id;
	int	outputs[TEA6415C_NUM_OUTPUTS];
};

/* addresses to scan, found only at 0x03 and/or 0x43 (7-bit) */
static unsigned short normal_i2c[] = {I2C_TEA6415C_1, I2C_TEA6415C_2, I2C_CLIENT_END};
static unsigned short normal_i2c_range[] = {I2C_CLIENT_END};

/* magic definition of all other variables and things */
I2C_CLIENT_INSMOD;

static struct i2c_driver driver;

/* unique ID allocation */
static int tea6415c_id = 0;

#ifdef SENSORS
/* for easier access */
int poutputs[] = {
			TEA6415C_OUTPUT1,
			TEA6415C_OUTPUT2,
			TEA6415C_OUTPUT3,
			TEA6415C_OUTPUT4,
			TEA6415C_OUTPUT5,
			TEA6415C_OUTPUT6,
			-1
		};
		
int pinputs[] = {
			TEA6415C_INPUT1,
			TEA6415C_INPUT2,
			TEA6415C_INPUT3,
			TEA6415C_INPUT4,
			TEA6415C_INPUT5,
			TEA6415C_INPUT6,
			TEA6415C_INPUT7,
			TEA6415C_INPUT8,
			-1
		};

#define TEA6415C_SYSCTL_OUTPUT1 1000
#define TEA6415C_SYSCTL_OUTPUT2 1001
#define TEA6415C_SYSCTL_OUTPUT3 1002
#define TEA6415C_SYSCTL_OUTPUT4 1003
#define TEA6415C_SYSCTL_OUTPUT5 1004
#define TEA6415C_SYSCTL_OUTPUT6 1005

static int tea6415c_switch(struct i2c_client *client, int i, int o);

void tea6415c_output(struct i2c_client *client, int operation, int ctl_name, int *nrels_mag, long *results, int output)
{
	struct tea6415c_data *data = client->data;
	
	if (operation == SENSORS_PROC_REAL_INFO)
		*nrels_mag = 0;
	else if (operation == SENSORS_PROC_REAL_READ) {
		results[0] = data->outputs[output];
		*nrels_mag = 1;
	} else if (operation == SENSORS_PROC_REAL_WRITE) {
		if (*nrels_mag == 1) {
			/* sanity check */
			if( results[0] >= 1 && results[0] <= TEA6415C_NUM_INPUTS ) {
				/* perform connection, data gets update in switch-function */
				tea6415c_switch(client, pinputs[results[0]-1], poutputs[output-1]);
			}
		}
	}
}

void tea6415c_output1(struct i2c_client *client, int operation, int ctl_name, int *nrels_mag, long *results)
{
	return tea6415c_output(client, operation, ctl_name, nrels_mag, results, 1);
}
void tea6415c_output2(struct i2c_client *client, int operation, int ctl_name, int *nrels_mag, long *results)
{
	return tea6415c_output(client, operation, ctl_name, nrels_mag, results, 2);
}
void tea6415c_output3(struct i2c_client *client, int operation, int ctl_name, int *nrels_mag, long *results)
{
	return tea6415c_output(client, operation, ctl_name, nrels_mag, results, 3);
}
void tea6415c_output4(struct i2c_client *client, int operation, int ctl_name, int *nrels_mag, long *results)
{
	return tea6415c_output(client, operation, ctl_name, nrels_mag, results, 4);
}
void tea6415c_output5(struct i2c_client *client, int operation, int ctl_name, int *nrels_mag, long *results)
{
	return tea6415c_output(client, operation, ctl_name, nrels_mag, results, 5);
}
void tea6415c_output6(struct i2c_client *client, int operation, int ctl_name, int *nrels_mag, long *results)
{
	return tea6415c_output(client, operation, ctl_name, nrels_mag, results, 6);
}

/* these files are created for each detected tea6415c */
static ctl_table tea6415c_dir_table_template[] = {
	{TEA6415C_SYSCTL_OUTPUT1, "output1", NULL, 0, 0644, NULL, &sensors_proc_real,
	 &sensors_sysctl_real, NULL, &tea6415c_output1},
	{TEA6415C_SYSCTL_OUTPUT2, "output2", NULL, 0, 0644, NULL, &sensors_proc_real,
	 &sensors_sysctl_real, NULL, &tea6415c_output2},
	{TEA6415C_SYSCTL_OUTPUT3, "output3", NULL, 0, 0644, NULL, &sensors_proc_real,
	 &sensors_sysctl_real, NULL, &tea6415c_output3},
	{TEA6415C_SYSCTL_OUTPUT4, "output4", NULL, 0, 0644, NULL, &sensors_proc_real,
	 &sensors_sysctl_real, NULL, &tea6415c_output4},
	{TEA6415C_SYSCTL_OUTPUT5, "output5", NULL, 0, 0644, NULL, &sensors_proc_real,
	 &sensors_sysctl_real, NULL, &tea6415c_output5},
	{TEA6415C_SYSCTL_OUTPUT6, "output6", NULL, 0, 0644, NULL, &sensors_proc_real,
	 &sensors_sysctl_real, NULL, &tea6415c_output6},
	{0}
};
#endif

/* this function is called by i2c_probe */
int tea6415c_detect(struct i2c_adapter *adapter, int address, unsigned short flags, int kind)
{
	struct	i2c_client *client = 0;
	struct	tea6415c_data *data = 0;
	int err = 0;

	/* let's see whether this adapter can support what we need */
	if ( 0 == i2c_check_functionality(adapter, I2C_FUNC_SMBUS_WRITE_BYTE)) {
		return 0;
	}

	/* allocate memory for client structure */
	client = kmalloc(sizeof(struct i2c_client), GFP_KERNEL);
        if (0 == client) {
		return -ENOMEM;
	}

	/* allocate memory for additional data */
	data = kmalloc(sizeof(struct tea6415c_data), GFP_KERNEL);
        if (0 == data) {
		kfree(client);
		return -ENOMEM;
	}
		
	/* fill client structure */
	sprintf(client->name,"tea6415c (0x%02x)", address);
	client->id = tea6415c_id++;
	client->flags = 0;
	client->addr = address;
	client->adapter = adapter;
	client->driver = &driver;
	client->data = data;

	/* tell the i2c layer a new client has arrived */
	if (0 != (err = i2c_attach_client(client))) {
		kfree(client);
		return err;
	}

	/* client-specific initialization: initialize the output-array
	   with a value that indicates "invalid" */
	memset(&data->outputs[0], -1, sizeof(int)*TEA6415C_NUM_OUTPUTS);

#ifdef SENSORS
	/* register a new directory entry with module sensors */
	if ((err = sensors_register_entry(client, "tea6415c", tea6415c_dir_table_template, THIS_MODULE)) < 0) {
		kfree(data);
		kfree(client);
		return err;
	}
	data->sysctl_id = err;
#endif
	printk("tea6415c.o: detected @ 0x%02x on adapter %s\n",2*address,&client->adapter->name[0]);

	return 0;
}

static int tea6415c_attach(struct i2c_adapter *adapter)
{
	return i2c_probe(adapter,&addr_data,&tea6415c_detect);
}

static int tea6415c_detach(struct i2c_client *client)
{
	int err = 0;

#ifdef SENSORS
	sensors_deregister_entry(((struct tea6415c_data*)(client->data))->sysctl_id);
#endif

	if ( 0 != (err = i2c_detach_client(client))) {
		printk("tea6415c.o: Client deregistration failed, client not detached.\n");
		return err;
	}
	
	kfree(client->data);
	kfree(client);

	return 0;
}

/* makes a connection between the input-pin 'i' and the output-pin 'o'
   for the tea6415c-client 'client' */
static int tea6415c_switch(struct i2c_client *client, int i, int o)
{
	u8 	byte = 0;
	int 	in,out;
	
	dprintk("tea6415c.o: tea6415c_switch: adr:0x%02x, i:%d, o:%d\n", client->addr, i, o);
		
	/* check if the pins are valid */
	if ( 0 == ((  1 == i ||  3 == i ||  5 == i ||  6 == i ||  8 == i || 10 == i || 20 == i || 11 == i ) &&
		    (18 == o || 17 == o || 16 == o || 15 == o || 14 == o || 13 == o )))
		return -1;

	/* to understand this, have a look at the tea6415c-specs (p.5) */
	switch(o) {
		case 18:
			byte = 0x00;
			break;
		case 14:
			byte = 0x20;
			break;
		case 16:
			byte = 0x10;
			break;
		case 17:
			byte = 0x08;
			break;
		case 15:
			byte = 0x18;
			break;
		case 13:
			byte = 0x28;
			break;
	};
		
	switch(i) {
		case 5:
			byte |= 0x00;
			break;
		case 8:
			byte |= 0x04;
			break;
		case 3:
			byte |= 0x02;
			break;
		case 20:
			byte |= 0x06;
			break;
		case 6:
			byte |= 0x01;
			break;
		case 10:
			byte |= 0x05;
			break;
		case 1:
			byte |= 0x03;
			break;
		case 11:
			byte |= 0x07;
			break;
	};

	if ( 0 != i2c_smbus_write_byte(client,byte)) {
		dprintk("tea6415c.o: tea6415c_switch: could not write to tea6415c\n");
		return -1;
	}

#ifdef SENSORS
	/* update data */
	for(in = 0; in < TEA6415C_NUM_INPUTS; in++) {
		if( pinputs[in] == i )
			break;
	}
	for(out = 0; out < TEA6415C_NUM_OUTPUTS; out++) {
		if( poutputs[out] == o )
			break;
	}
	((struct tea6415c_data*)client->data)->outputs[out] = in;
#endif
	
	return 0;
}

void tea6415c_inc_use(struct i2c_client *client)
{
#ifdef MODULE
	MOD_INC_USE_COUNT;
#endif
}

void tea6415c_dec_use(struct i2c_client *client)
{
#ifdef MODULE
	MOD_DEC_USE_COUNT;
#endif
}

static int
tea6415c_command(struct i2c_client *client, unsigned int cmd, void* arg)
{
	struct tea6415c_video_multiplex *v = (struct tea6415c_video_multiplex*)arg;
	int result = 0;

	switch (cmd) {
		case TEA6415C_SWITCH: {
			result = tea6415c_switch(client,v->in,v->out);
			break;
		}
		default: {
			return -ENOIOCTLCMD;
		}
	}

	if ( 0 != result )
		return result;
	
	return 0;
}

static struct i2c_driver driver = {
	"tea6415c driver",
	I2C_DRIVERID_TEA6415C,
	I2C_DF_NOTIFY,
	tea6415c_attach,
	tea6415c_detach,
	tea6415c_command,
	tea6415c_inc_use,
	tea6415c_dec_use
	};

EXPORT_NO_SYMBOLS;

#ifdef MODULE
static
#else
extern
#endif
	int __init tea6415c_init(void)
{
	int res = 0;

	printk("tea6415c.o version %s (%s)\n",VERSION,DATE);

	if( 0 != (res = i2c_add_driver(&driver))) {
		printk("tea6415c.o: Driver registration failed, module not inserted.\n");
		return res;
	}
	
	return 0;
}

int __init tea6415c_cleanup(void)
{	
	int res = 0;

	if ( 0 != (res = i2c_del_driver(&driver))) {
		printk("tea6415c.o: Driver deregistration failed, module not removed.\n");
		return res;
	}
	
	return 0;
}

#ifdef MODULE

MODULE_AUTHOR("Michael Hunold <michael@mihu.de>");
MODULE_DESCRIPTION("tea6415c driver");

int init_module(void)
{
	return tea6415c_init();
}

void cleanup_module(void)
{
	tea6415c_cleanup();
}

#endif

#ifdef MODULE
MODULE_LICENSE("GPL");
#endif

