#define DATE "20000701"
#define VERSION "1.1"

#include <linux/module.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <linux/i2c.h>

#include "tea6420.h"

#ifdef MODULE
MODULE_PARM(debug,"i");
#endif

static int debug = 0;	/* insmod parameter */
#define dprintk	if (debug) printk

/* addresses to scan, found only at 0x4c and/or 0x4d (7-Bit) */
static unsigned short normal_i2c[] = {I2C_TEA6420_1, I2C_TEA6420_2, I2C_CLIENT_END};
static unsigned short normal_i2c_range[] = {I2C_CLIENT_END};

/* magic definition of all other variables and things */
I2C_CLIENT_INSMOD;

static struct i2c_driver driver;

/* unique ID allocation */
static int tea6420_id = 0;

/* make a connection between the input 'i' and the output 'o'
   with gain 'g' for the tea6420-client 'client' (note: i = 6 means 'mute') */
static int tea6420_switch(struct i2c_client *client, int i, int o, int g)
{
	u8 	byte = 0;
	
	int 	result = 0;
	
	dprintk("tea6420.o: tea6420_switch: adr:0x%02x, i:%d, o:%d, g:%d\n",client->addr,i,o,g);

	/* check if the paramters are valid */
	if ( i < 1 || i > 6 || o < 1 || o > 4 || g < 0 || g > 6 || g%2 != 0 )
		return -1;

	byte  = ((o-1)<<5);
	byte |=  (i-1);

	/* to understand this, have a look at the tea6420-specs (p.5) */
	switch(g) {
		case 0:
			byte |= (3<<3);
			break;
		case 2:
			byte |= (2<<3);
			break;
		case 4:
			byte |= (1<<3);
			break;
		case 6:
			break;
	}

	/* fixme?: 1 != ... => 0 != */
	if ( 0 != i2c_smbus_write_byte(client,byte)) {
		dprintk(KERN_ERR "tea6420.o: could not switch, result:%d\n",result);
		return -EFAULT;
	}
	
	return 0;
}

/* this function is called by i2c_probe */
int tea6420_detect(struct i2c_adapter *adapter, int address, unsigned short flags, int kind)
{
	struct	i2c_client *client;
	int err = 0, i = 0;

	/* let's see whether this adapter can support what we need */
	if ( 0 == i2c_check_functionality(adapter, I2C_FUNC_SMBUS_WRITE_BYTE)) {
		return 0;
	}

	/* allocate memory for client structure */
	client = kmalloc(sizeof(struct i2c_client), GFP_KERNEL);
        if (0 == client) {
		return -ENOMEM;
	}
	
	/* fill client structure */
	sprintf(client->name,"tea6420 (0x%02x)", address);
	client->id = tea6420_id++;
	client->flags = 0;
	client->addr = address;
	client->adapter = adapter;
	client->driver = &driver;
	client->data = NULL;

	/* tell the i2c layer a new client has arrived */
	if (0 != (err = i2c_attach_client(client))) {
		kfree(client);
		return err;
	}

	/* set initial values: set "mute"-input to all outputs at gain 0 */
	err = 0;
	for(i = 1; i < 5; i++) {
		err += tea6420_switch(client, 6, i, 0);
	}
	if( 0 != err) {
		printk("tea6420.o: could not initialize chipset. continuing anyway.\n");
	}
	
	printk("tea6420.o: detected @ 0x%02x on adapter %s\n",2*address,&client->adapter->name[0]);

	return 0;
}

static int tea6420_attach(struct i2c_adapter *adapter)
{
	return i2c_probe(adapter,&addr_data,&tea6420_detect);
}

static int tea6420_detach(struct i2c_client *client)
{
	int err = 0;

	if ( 0 != (err = i2c_detach_client(client))) {
		printk("tea6420.o: Client deregistration failed, client not detached.\n");
		return err;
	}
	
	kfree(client);

	return 0;
}

static int
tea6420_command(struct i2c_client *client, unsigned int cmd, void* arg)
{
	struct audio_multiplex *a = (struct audio_multiplex*)arg;
	int result = 0;

	switch (cmd) {
		case TEA6420_SWITCH: {
			result = tea6420_switch(client,a->in,a->out,a->gain);
			break;
		}
		default: {
			return -ENOIOCTLCMD;
		}
	}

	if ( 0 != result )
		return result;
	
	return 0;
}

void tea6420_inc_use(struct i2c_client *client)
{
#ifdef MODULE
	MOD_INC_USE_COUNT;
#endif
}

void tea6420_dec_use(struct i2c_client *client)
{
#ifdef MODULE
	MOD_DEC_USE_COUNT;
#endif
}

static struct i2c_driver driver = {
	"tea6420 driver",
	I2C_DRIVERID_TEA6420,
	I2C_DF_NOTIFY,
	tea6420_attach,
	tea6420_detach,
	tea6420_command,
	tea6420_inc_use,
	tea6420_dec_use
};

EXPORT_NO_SYMBOLS;

#ifdef MODULE
static
#else
extern
#endif
	int __init tea6420_init(void)
{
	int res = 0;

	printk("tea6420.o version %s (%s)\n",VERSION,DATE);

	if( 0 != (res = i2c_add_driver(&driver))) {
		printk("tea6420.o: Driver registration failed, module not inserted.\n");
		return res;
	}
	
	return 0;
}

int __init tea6420_cleanup(void)
{	
	int res = 0;

	if ( 0 != (res = i2c_del_driver(&driver))) {
		printk("tea6420.o: Driver deregistration failed, module not removed.\n");
		return res;
	}
	
	return 0;
}

#ifdef MODULE

MODULE_AUTHOR("Michael Hunold <michael@mihu.de>");
MODULE_DESCRIPTION("tea6420 driver");

int init_module(void)
{
	return tea6420_init();
}

void cleanup_module(void)
{
	tea6420_cleanup();
}

#endif

#ifdef MODULE
MODULE_LICENSE("GPL");
#endif
