/*
    skel_v4l2.c - skeletion v4l2 driver for developing purposes
    
    Copyright (C) 2000 Michael Hunold <michael@mihu.de>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <linux/module.h>	/* for module-version */
#include <linux/slab.h>		/* for kmalloc/kfree */
#include <linux/init.h>		/* for "__init" */
#include <linux/delay.h>	/* for mdelay */
#include <linux/poll.h>
#include <linux/i2c.h>

#include "videodevX/videodev.h"

#include "core/saa7146_v4l2.h"

/* global defines */
#define DEB1(x) if (debug>=1) x
#define DEB2(x) if (debug>=2) x

/* module parameters */
int debug = 2;

/* ignore card switch, don't ignore any card by default */
int ignore[MAX_SAA7146_BOARDS] = { 0, 0, 0, 0, };

/* minor device-nodes to be used for video */
int unit_video[MAX_SAA7146_BOARDS] = { 0, 1, 2, 3, };
/* minor device-nodes to be used for vbi */
int unit_vbi[MAX_SAA7146_BOARDS] = { 224, 225, 226, 227, };

/* MAX_SAA7146_BOARDS gets defined in saa7146.h */

#ifdef MODULE
MODULE_PARM(debug,"i");
MODULE_PARM_DESC(debug, "debug level - 0: off; 1: normal; 2: verbose");

MODULE_PARM(ignore, "1-"__MODULE_STRING(MAX_SAA7146_BOARDS)"i");
MODULE_PARM_DESC(ignore, "exclude device from being recognized");

MODULE_PARM(unit_video, "1-"__MODULE_STRING(MAX_SAA7146_BOARDS)"i");
MODULE_PARM_DESC(unit_video, "minor device-node(s) to be used for video");

MODULE_PARM(unit_vbi, "1-"__MODULE_STRING(MAX_SAA7146_BOARDS)"i");
MODULE_PARM_DESC(unit_vbi, "minor device-node(s) to be used for vbi");
#endif

/* global variables */
int skel_v4l2_num = 0;

/* we simply use the two video-interfaces (aka port a and port b) */
#define SKEL_INPUTS	6
struct v4l2_input skel_inputs[SKEL_INPUTS] = {
	{ 0, "Port A",	V4L2_INPUT_TYPE_CAMERA,	V4L2_INPUT_CAP_AUDIO, 1, },
	{ 1, "Port B",	V4L2_INPUT_TYPE_CAMERA,	V4L2_INPUT_CAP_AUDIO, 1, },
	{ 2, "Y Port B - C Port A, Sync Port A",	V4L2_INPUT_TYPE_CAMERA,	V4L2_INPUT_CAP_AUDIO, 1, },
	{ 3, "Y Port B - C Port A, Sync Port B",	V4L2_INPUT_TYPE_CAMERA,	V4L2_INPUT_CAP_AUDIO, 1, },
	{ 4, "Y Port A - C Port B, Sync Port A",	V4L2_INPUT_TYPE_CAMERA,	V4L2_INPUT_CAP_AUDIO, 1, },
	{ 5, "Y Port A - C Port B, Sync Port B",	V4L2_INPUT_TYPE_CAMERA,	V4L2_INPUT_CAP_AUDIO, 1, },
};

#define SKEL_AUDIOS	0

/* you can demand that your extension get's called for any v4l2-ioctl. here,
   we want to be called exclusively when the ioctls VIDIOC_S_INPUT and
   VIDIOC_ENUMINPUT get called by an user application */
struct saa7146_v4l2_extension_ioctls ioctls[] = {
	{ VIDIOC_S_INPUT,	SAA7146_EXCLUSIVE },
	{ VIDIOC_ENUMINPUT, 	SAA7146_EXCLUSIVE },
	{ 0,			0 }
};

struct skel_v4l2
{
	struct capture_device*	saa7146;

	int cur_input;	/* current input */
};

/* this function gets called very early in the registration-process of
   the extension. it has been reported that some devices need to enable
   the i2c-bus explicitly for example -- this can be done here. please
   note, that you cannot be sure that the device is really the hardware
   you expect, so you should do as little as possible in here, in order
   to avoid confusing the hardware. */
int	skel_v4l2_preinit(struct capture_device* saa)
{
	DEB2(printk("skel_v4l2.o: skel_v4l2_pre_init called.\n"));

	/* first check if this device is to be ignored
	   by using serial identifier */
	if( 1 == ignore[ saa->serial ]) {
		DEB1(printk("skel_v4l2.o: note: device ignored due to insmod parameter.\n"));	
		return -ENODEV;
	}

	return 0;
}

int	skel_v4l2_probe(struct capture_device* saa)
{
	struct skel_v4l2* skel = 0;	

	DEB2(printk("skel_v4l2.o: skel_v4l2_probe called.\n"));

	skel = (struct skel_v4l2*)kmalloc(sizeof(struct skel_v4l2), GFP_KERNEL);
	if( NULL == skel ) {
		printk("skel_v4l2.o: skel_v4l2_attach: not enough kernel memory.\n");
		return -ENOMEM;
	}
	memset(skel, 0x0, sizeof(struct skel_v4l2));	

	/* normally, we would loop through all i2c-devices on the i2c-bus and
	   look who is there + check if all devices are present */
/*
	for(i = 0; i < saa->i2c_adapter->client_count; i++) {
	}
*/

	DEB2(printk("skel_v4l2.o: skel_v4l2_probe succeeded for this device.\n"));	

	/* we store the pointer in our private data field */
	(struct skel_v4l2*)saa->ext_priv = skel;

	/* store handle to saa7146a */
	skel->saa7146 = saa;

	return 0;
}

/* bring hardware to a sane state. this has to be done, just in case someone
   wants to capture from this device before it has been properly initialized.
   the capture engine would badly fail, because no valid signal arrives on the
   saa7146, thus leading to timeouts and stuff. */
int skel_v4l2_init_done(struct capture_device* saa)
{
	struct skel_v4l2* skel = (struct skel_v4l2*)saa->ext_priv;

	DEB2(printk("skel_v4l2.o: skel_v4l2_init_done called.\n"));

	/* todo: initialize the helper ics to useful values */

	/* the rest for saa7146: you should definitely set some basic values
	   for the input-port handling of the saa7146. */
	   
	/* some stuff is done via variables */
	saa7146_set_hps_source_and_sync(skel->saa7146, SAA7146_HPS_SOURCE_PORT_A, SAA7146_HPS_SYNC_PORT_A);

	/* some stuff is done via direct write to the registers */

	/* this is ugly, but because of the fact that this is completely
	   hardware dependend, it should be done directly... */
      	saa7146_write(skel->saa7146, DD1_STREAM_B,	0x00000000);
	saa7146_write(skel->saa7146, DD1_INIT,		0x02000200);
	saa7146_write(skel->saa7146, MC2, (MASK_09 | MASK_25 | MASK_10 | MASK_26));

	return 0;
}

/* this function only gets called when the probing was successful */
int skel_v4l2_attach(struct capture_device* saa)
{
	struct skel_v4l2* skel = (struct skel_v4l2*)saa->ext_priv;
	
	DEB2(printk("skel_v4l2.o: skel_v4l2_attach called.\n"));

	/* checking for i2c-devices can be omitted here, because we
	   already did this in "skel_vl42_probe" */
	skel_v4l2_num++;

	/* increase the usage count of any i2c-client by calling
	   i2c_inc_use_client(...) */

	/* the rest */
	skel->cur_input = 0;
	
	return 0;
}

int skel_v4l2_detach(struct capture_device* saa)
{
	struct skel_v4l2* skel = (struct skel_v4l2*)saa->ext_priv;
	
	DEB2(printk("skel_v4l2.o: skel_v4l2_detach called.\n"));

	/* decrease the usage count of any i2c-client by calling
	   i2c_dec_use_client(...) */

	skel_v4l2_num--;

	kfree(skel);

	return 0;
}


void skel_v4l2_open(struct capture_device* saa)
{
#ifdef MODULE
	MOD_INC_USE_COUNT;
#endif
}

void skel_v4l2_close(struct capture_device* saa)
{
#ifdef MODULE
	MOD_DEC_USE_COUNT;
#endif
}

int skel_v4l2_ioctl(struct capture_device *dev, unsigned int cmd, void *arg) 
{
	struct skel_v4l2* skel = (struct skel_v4l2*)dev->ext_priv;

	switch(cmd)
	{
	case VIDIOC_S_INPUT:
	{
		int	input = *(int *)arg;
		int 	source = 0, sync = 0;
		
		if (input < 0 || input >= SKEL_INPUTS) {
			DEB2(printk("skel_v4l2.o: v4l2_ioctl: VIDIOC_S_INPUT: invalid input %d.\n",input));
			return -EINVAL;
		}

		DEB2(printk("skel_v4l2.o: v4l2_ioctl: VIDIOC_S_INPUT %d.\n",input));

		switch(input) {
			case 0: /* Port A */
			{
				source = SAA7146_HPS_SOURCE_PORT_A;
				sync = SAA7146_HPS_SYNC_PORT_A;
				break;
			}
			case 1: /* Port B */
			{
				source = SAA7146_HPS_SOURCE_PORT_B;
				sync = SAA7146_HPS_SYNC_PORT_B;
				break;
			}
			case 2: /* Y Port B - C Port A, Sync Port A */
			{
				source = SAA7146_HPS_SOURCE_YPB_CPA;
				sync = SAA7146_HPS_SYNC_PORT_A;
				break;
			}
			case 3: /* Y Port B - C Port A, Sync Port B */
			{
				source = SAA7146_HPS_SOURCE_YPB_CPA;
				sync = SAA7146_HPS_SYNC_PORT_B;
				break;
			}
			case 4: /* Y Port A - C Port B, Sync Port A */
			{
				source = SAA7146_HPS_SOURCE_YPA_CPB;
				sync = SAA7146_HPS_SYNC_PORT_A;
				break;
			}
			case 5: /* Y Port A - C Port B, Sync Port B */
			{
				source = SAA7146_HPS_SOURCE_YPA_CPB;
				sync = SAA7146_HPS_SYNC_PORT_B;
				break;
			}
		}
		skel->cur_input = input;

		/* fixme: switch input here, switch audio, too! */
		saa7146_set_hps_source_and_sync(dev, source, sync);

		return 0;
	}
	case VIDIOC_ENUMINPUT:
	{
		struct v4l2_input *i = arg;
		
		/* sanity check to satisfy xawtv */
		if( i->index < 0 || i->index >= SKEL_INPUTS) {
			DEB2(printk("skel_v4l2.o: v4l2_ioctl: VIDIOC_ENUMINPUT: invalid input %d.\n",i->index));
			return -EINVAL;
		}
		
		memcpy(i, &skel_inputs[i->index], sizeof(struct v4l2_input));

		DEB2(printk("skel_v4l2.o: v4l2_ioctl: VIDIOC_ENUMINPUT %d.\n",i->index));
		return 0;
	}
	default:
		DEB2(printk("skel_v4l2.o: v4l2_ioctl does not handle this ioctl.\n"));
/*
*/
		return -ENOIOCTLCMD;
	}
	return 0;
}

struct saa7146_v4l2_extension extension = {
	"skeleleton saa7146 driver",
	SKEL_INPUTS,
	SKEL_AUDIOS,
	0,
	&unit_video[0],
	&unit_vbi[0],
	&ioctls[0],
	0,
	NULL,
	skel_v4l2_preinit,
	skel_v4l2_probe,
	skel_v4l2_init_done,
	skel_v4l2_attach,
	skel_v4l2_detach,
	skel_v4l2_open,
	skel_v4l2_close,
	skel_v4l2_open,
	NULL,
	skel_v4l2_close,
	skel_v4l2_ioctl,
};

int __init skel_v4l2_init(void)
{
	DEB2(printk("skel_v4l2.o: skel_v4l2_init called.\n"));
	
	if( 0 != saa7146_v4l2_register_extension(&extension)) {
		DEB1(printk("skel_v4l2.o: failed to register extension.\n"));
		return -ENODEV;
	}

	/* skel_v4l2_num gets increased if the probe function is called successfully */

	if( 0 == skel_v4l2_num ) {
		printk(KERN_ERR "skel_v4l2.o: no saa7146(s) found.\n");
		return -ENODEV;
	}

	printk(KERN_ERR "skel_v4l2.o: %d saa7146(s) found.\n", skel_v4l2_num);
	return 0;
}

#ifdef MODULE
MODULE_AUTHOR("Michael Hunold <michael@mihu.de>");
MODULE_DESCRIPTION("skeletion v4l2-driver");

EXPORT_NO_SYMBOLS;

int init_module(void) 
{
	return skel_v4l2_init();
}

void cleanup_module(void) 
{
	DEB2(printk("skel_v4l2.o: cleanup_module called.\n"));
	saa7146_v4l2_unregister_extension(&extension);
}

#endif

