/*   iic_read.c
**  read a sequence of bytes from the iic bus using the device interface 
**
**  Copyright (C) 2001 Hartmut Hackmann <hartmut.hackmann@t-online.de>
**
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation; either version 2 of the License, or
**  (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
**
** Author : H. Hackmann
** Date 18.02.2000  Hac
**         Initial Version
**      12.03.2001  Hac
**         added "force" option
*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>

#include <sys/ioctl.h>
#include <linux/types.h>
#include <linux/i2c.h>

#define MAX_MESSAGE 40                /* maximum message length */

char *prgname,                        /* needed for messages */
     *device;

char          default_device[] = "/dev/i2c-0";
unsigned char message[MAX_MESSAGE];
unsigned char slave;
int           read_bytes = 1;

/********************     Usage Message    ************************/

void usage(void) {
    fprintf(stderr, "usage : %s [options] <slave> <message>\n"
            "  writes <message> to the iic slave <slave>. \n" 
            "  <slave> is the iic slave address in 8bit notation, \n"
            "  <message> is a sequence of bytes each as a separate arg \n"
            "  afterwards (-n) bytes (default: 1) are read and printed to stdout.\n"
            "  options: -b n : use number base n instead of 16 (hex)\n"
            "           -d s : use device s instead of %s\n"
#if defined(I2C_SLAVE_FORCE) 
            "           -f   : force, ignores assignment to another driver\n"
#endif
            "           -h   : print this message, then exit\n"
            "           -n n : the number of bytes to read\n" 
            , prgname, default_device);
    exit(1);
}

/********************   fatal error, abort   ************************/

void fatal(char *msg) { 
    fprintf(stderr, "%s: error - %s, exiting\n", prgname, msg);
    exit(2);
}

/******************************************************************/

unsigned char get_byte(char *str, int base) {
    int byte;
    byte = strtol(str, NULL, base);
    if ((byte < 0) || (byte > 255))
        fatal( "argument out of range");
    return byte;
}

/******************************************************************/
/* 
 * Function: uinttostr
 *    Converts an unsigned integer to a string
 * Parameters:
 *    num  : The unsigned number to convert
 *    base : The base of the number system to use, may be between 2 and 36
 *    width: The length of the string to generate including the padding 0.
 *           if it is 0, the string has the required length.
 *    pad:   If non 0, the string will have leading 0s instead of blanks.
 *    str:   The pointer to the character array to store the string.
 * Results:
 *    The resulting string is stored in the string <str>. The function
 *    returns a pointer to the 0-character ending the resulting string.
 *    If the result does not fit into the desired length, a sequence 
 *    of '-' will be returned
 */

char *uinttostr(unsigned num, unsigned base, int width, int pad, char *str) {
    
    int      i, digit;
    unsigned base_shifted; 
    char     *current_pos;

    /* first check whether the string can hold the number 
     * calculate the highest number that fits into the string
     */
    width --;
    current_pos = str;
    base_shifted = 1;
    if(width > -1) {
        for(i = 1; i < width; i++)
            base_shifted = base_shifted * base;
        if(base_shifted < num / base) {
            /* string is too short */
            for(i = 0; i < width; i++)
                *current_pos++ ='-';
            *current_pos = '\0';
            return current_pos;
        }
    }
    else {
        while( (num / base_shifted) >= base)
            base_shifted = base_shifted * base;
    }
    /* base_shifted now gives the "worth" of the highest digit,
     * convert from left to right
     */
    while(base_shifted >= 1) {
        digit = num / base_shifted;
        num = num - digit * base_shifted;
        base_shifted = base_shifted / base;
        if((digit == 0) && !pad)
            *current_pos++ = ' ';
        else {
            pad = 1;
            if(digit > 9)
                *current_pos++ = 'a' + digit - 10;
            else
                *current_pos++ = '0' + digit;
        }
    }
    *current_pos = '\0';
    return current_pos;
}

/******************************************************************/

int main(int argc, char *argv[]) {
    extern char *optarg;
    extern int  optind, optopt;
    char        c,
                bytestr[9];
    int         base = 16;
    int         outwidth;
    int         handle, length, i;
    int         mode = I2C_SLAVE_FORCE;
    
    prgname = argv[0];
    device = default_device;
 
    /* process command line options */

#if defined(I2C_SLAVE_FORCE) 
    while ((c = getopt(argc, argv, "b:d:fhn:")) != -1) {
#else
    while ((c = getopt(argc, argv, "b:d:hn:")) != -1) {
#endif
        switch(c) {
            case 'b' : base = strtol(optarg, NULL, 10);
                       if((base < 2) || (base > 35))
                           usage();
                       break;
            case 'd' : device = optarg; 
                       break;
#if defined(I2C_SLAVE_FORCE) 
            case 'f' : mode = I2C_SLAVE_FORCE;  
                       break;
#endif
            case 'n' : read_bytes = strtol(optarg, NULL, 10);
                       if(( read_bytes < 1) || (read_bytes > MAX_MESSAGE)) 
                           fatal("read message too long");
                       break;
            default  : usage();
        }
    }

    /* get the slave address */
    if(argv[optind] == NULL)
       usage();
    slave = get_byte(argv[optind++], base);

    /* all remaining args are message bytes */
    length = 0;
    while(argv[optind] != NULL) {
        if(length >= MAX_MESSAGE - 1 )
            fatal("write message too long");
        message[length] = get_byte(argv[optind++], base);
        length++;
    }

    /* open the I2c interface */
    handle = open(device,O_RDWR);
    if(handle < 0 )
        fatal(strerror(errno));
    
    /* set the proper I2C slave */
    if (ioctl(handle, mode, slave >> 1) < 0) {
	fatal(strerror(errno));
        }
	
    /* write the message */
    if(write(handle, message, length) != length)
        fatal(strerror(errno));

    /* read the desired bytes */
    if(read(handle, message, read_bytes) != read_bytes)
        fatal(strerror(errno));

    /* find how many characters are needed for the output string */ 
    outwidth = uinttostr(255, base, 0, 0, bytestr) - bytestr + 1; 

    /* print the result */
    for(i = 0; i < read_bytes; i++) {
        uinttostr((unsigned int) message[i], base, outwidth, 1, bytestr);
        fprintf(stdout,"%s ", bytestr);
    }
    fprintf(stdout, "\n");
    
    /* done, cleanup */
    close(handle);
    return(0);
}

